classdef MESHQ < matlab.mixin.Copyable
    
    properties (Access='public')
        
        name
        
        vertices
        quads
        edges
        
        uv                      % 2D UV coords
        quads_uv                % UV faces
        
        nv                      % #vertices
        nf                      % #faces
        ne                      % #edges
        nv_uv                   % #texture vertices
        ne_uv                   % #texture edges
        
        AM,valence              % adjacency matrix vertices-vertices
        v2e                     % vertices to edges indexing
        AM_VF                   % adjacency matrix vertices-faces
        AM_uv                   % adjacency matrix uv vertices-vertices
        
        va                      % vertex areas, nv x 1
        vn                      % vertices normal vectors, nv x 3
        fva                     % face vector areas (Alexa/Wardetzky Laplacian), nf x 3
        fn                      % face normal directions = normalize(fva), nf x 3
        fa                      % face areas = norm(fva), nf x 1
        
        el                      % edge lengths, ne x 1
        mel                     % mean edge length
        
        E1, E2, E3, E4          % quad edges, nf x 3 each
        B1, B2, B3, B4          % quad edge midpoints, nf x 3 each
        
        Ep1, Ep2, Ep3, Ep4      % edges of max projection planar quads, nf x 3 each
        
        lambda                  % lambda from AW Laplacian
        Mf_g                    % Geometric M1 mass matrix per face, \tilde{M}_f, nf x {4 x 4}
        Mf                      % M1 mass matrix per face, nf x {4 x 4}
        
        d0                      % DEC d0, ne x nv (orientation = sign(v_end - v_start))
        M1                      % M1 mass matrix, ne x ne
        M0                      % M0 mass matrix, nv x nv
        
        W                       % Scale invariant Laplacian = d0^T * M1 * d0
        L                       % Laplacian = M0^{-1} * W
        
    end
    
    methods
        
        function [ mesh ] = MESHQ( varargin )
            
            if nargin == 1 % given mesh by name only and assuming off
                mesh.name = varargin{1};
                ext = 'off';
                fname = [mesh.name '.' ext];
                [mesh.vertices, mesh.quads] = MESH_IO.roff(fname, 4);
            else
                if ischar(varargin{1}) % given name and extension
                    mesh.name = varargin{1};
                    ext = varargin{2};
                    fname = [mesh.name '.' ext];
                    if strcmp(ext, 'off')
                        [mesh.vertices, mesh.quads] = MESH_IO.roff(fname, 4);
                    elseif strcmp(ext, 'obj')
                        [mesh.vertices, mesh.quads, mesh.uv, mesh.quads_uv] = ...
                            MESH_IO.robj(fname, 4);
                    else 
                        error('wrong extension');
                    end
                    
                else % inserted mesh_vertices and mesh_quads matrices
                    mesh.vertices = varargin{1};
                    mesh.quads = varargin{2};
                    if nargin > 3 % inserted uv as well
                        mesh.uv = varargin{3};
                        mesh.quads_uv = varargin{4};
                    end
                end
            end            
            
            mesh.nv = size(mesh.vertices,1);
            mesh.nf = size(mesh.quads,1);
            
            % Adjacency and edge list
            [mesh.AM, mesh.edges, mesh.v2e, mesh.AM_VF] = adj_mat( mesh );
            mesh.ne = size(mesh.edges,1);
            mesh.valence = sum(mesh.AM,2);
            
            % Adjacency uv
            if ~isempty(mesh.uv)
                mesh.nv_uv = size(mesh.uv,1);
                [mesh.AM_uv, mesh.ne_uv] = adj_mat_uv( mesh );
            end
            
            % Geometric edges and edge centers
            [mesh.E1,mesh.E2,mesh.E3,mesh.E4] = face_edges( mesh );
            [mesh.B1,mesh.B2,mesh.B3,mesh.B4] = face_edge_midpoints( mesh );
            
            % Areas and projected planar quads
            [mesh.fva, mesh.fn, mesh.fa] = face_areas_and_normals( mesh );
            [mesh.Ep1,mesh.Ep2,mesh.Ep3,mesh.Ep4] = face_edges_projected( mesh );
             
            % Vertices normals as weigthed sum of adjacent quads normals
            mesh.vn = vertices_normals( mesh );
            
            % Mass matrices
            mesh.lambda = 2;
            % [mesh.Mf_g, mesh.Mf, mesh.M1] = edge_mass_matrix( mesh );                        
            [mesh.va, mesh.M0] = vertex_mass_matrix( mesh );
            
            % Operators
            %[mesh.d0, mesh.W, mesh.L] = differential_operators ( mesh );
            
        end
        
        function [ AM, Ed, v2e, AM_VF] = adj_mat( mesh )
            % AM, nv x nv, adjacency
            % Ed, ne x 2, edge list
            % v2e, nv x nv, edge id from vertices
            % AM_VF, nv x nf, vertices faces adjcency matrix
            
            T = double( mesh.quads );
            
            I = [T(:,2);T(:,3);T(:,4);T(:,1)];
            J = [T(:,3);T(:,4);T(:,1);T(:,2)];
            S = ones(size(I(:),1),1);
            AM = sparse(I,J,S,mesh.nv,mesh.nv);
            
            [ii,jj,~] = find(triu(AM + AM'));
            Ed = [ii,jj];
            n = size(Ed,1);
            v2e = sparse([ii;jj],[jj;ii],[1:n,1:n]');
            
            % Create the Vertices-Faces adjacency matrix
            ii = [mesh.quads(:,1); mesh.quads(:,2); mesh.quads(:,3); mesh.quads(:,4)]; % vertices indices
            jj = [1:mesh.nf, 1:mesh.nf, 1:mesh.nf, 1:mesh.nf].'; % faces indices X 4
            vv = ones(size(jj)); % initiating to 1 - it's an adjacency matrix
            AM_VF = sparse(double(ii), jj, vv, mesh.nv, mesh.nf);
        end
        
        function [AM_uv, ne_uv] = adj_mat_uv ( mesh )
            %AM_uv nv_uv x nv_uv, adjacency
            
            T = double( mesh.quads_uv );
            
            II = [T(:,2);T(:,3);T(:,4);T(:,1)];
            JJ = [T(:,3);T(:,4);T(:,1);T(:,2)];
            S = ones(size(II(:),1),1);
            AM_uv = sparse(II,JJ,S,mesh.nv_uv,mesh.nv_uv);
            
            [ii,jj,~] = find(triu(AM_uv + AM_uv'));
            Ed = [ii,jj];
            ne_uv = size(Ed,1);
            
        end

        function [ E1, E2, E3, E4 ] = face_edges( mesh )
            X = mesh.vertices;
            Q = mesh.quads;
            
            E1 = X(Q(:,1),:) - X(Q(:,2),:);
            E2 = X(Q(:,2),:) - X(Q(:,3),:);
            E3 = X(Q(:,3),:) - X(Q(:,4),:);
            E4 = X(Q(:,4),:) - X(Q(:,1),:);
            
            E = [E1; E2; E3; E4];
            
            mesh.el = MESHQ.normv( E );
            mesh.mel = mean(mesh.el);
        end
        
        function [ B1, B2, B3, B4 ] = face_edge_midpoints( mesh )
            X = mesh.vertices;
            Q = mesh.quads;
            
            B1 = (X(Q(:,1),:) + X(Q(:,2),:))/2;
            B2 = (X(Q(:,2),:) + X(Q(:,3),:))/2;
            B3 = (X(Q(:,3),:) + X(Q(:,4),:))/2;
            B4 = (X(Q(:,4),:) + X(Q(:,1),:))/2;            
        end
        
        function [fva, fn, fa] = face_areas_and_normals( m )
            % oriented face normal, by Alexa + Wardetzky (AW) Eq (8)
            
            fva = zeros(m.nf,3);             
            for i=1:m.nf
                Ef = [m.E1(i,:); m.E2(i,:); m.E3(i,:); m.E4(i,:)];
                Bf = [m.B1(i,:); m.B2(i,:); m.B3(i,:); m.B4(i,:)];
                A = Ef'*Bf;
                fva(i,:) = [-A(2,3), A(1,3), -A(1,2)];
            end
            fva = -fva; % normals will point outward for CCW
            fn = MESHQ.normalize_vf(fva);
            fa = MESHQ.normv(fva);
        end
        
        function [Ep1,Ep2,Ep3,Ep4] = face_edges_projected( mesh )
            % maximal projections of quads, \bar{f} in WA paper
            
            X = mesh.vertices; Q = mesh.quads;
            X1 = X(Q(:,1),:); X2 = X(Q(:,2),:); X3 = X(Q(:,3),:); X4 = X(Q(:,4),:);
            X1b = X1 - dot(X1,mesh.fn,2).*mesh.fn;
            X2b = X2 - dot(X2,mesh.fn,2).*mesh.fn;
            X3b = X3 - dot(X3,mesh.fn,2).*mesh.fn;
            X4b = X4 - dot(X4,mesh.fn,2).*mesh.fn;
            
            Ep1 = X1b - X2b;
            Ep2 = X2b - X3b;
            Ep3 = X3b - X4b;
            Ep4 = X4b - X1b;
        end
        
        function [vn] = vertices_normals( mesh )
            
            % vertices normals as weighted sum of faces normals
            vn = mesh.AM_VF*mesh.fva;
            
            % normalize
            vn = MESHQ.normalize_vf(vn);
        end
        
        function [Mf_g, Mf, M1] = edge_mass_matrix( m )   
            % Mf, Mf_g in nf x 16, flattened per face matrices from eq (9)
            % M1 in ne x ne, entries collected from corresponding Mf entries
            
            Mf_g = zeros(m.nf,16);                                     
            Mf = zeros(m.nf,16); 
            ii = zeros(16*m.nf,1); jj = ii; ss = ii;
            loc = 1; 
            Q = m.quads;
            
            e = zeros(m.nf,4);
            e(:,1) = diag(m.v2e(Q(:,1),Q(:,2)));
            e(:,2) = diag(m.v2e(Q(:,2),Q(:,3)));
            e(:,3) = diag(m.v2e(Q(:,3),Q(:,4)));
            e(:,4) = diag(m.v2e(Q(:,4),Q(:,1)));
            
            s = zeros(m.nf,4);
            s(:,1) = double(sign(Q(:,2)-Q(:,1)));
            s(:,2) = double(sign(Q(:,3)-Q(:,2)));
            s(:,3) = double(sign(Q(:,4)-Q(:,3)));
            s(:,4) = double(sign(Q(:,1)-Q(:,4)));
                      
            for i=1:m.nf
                Bf = [m.B1(i,:); m.B2(i,:); m.B3(i,:); m.B4(i,:)];
                mfg = Bf*Bf'/m.fa(i)*sqrt(2);

                Epf = [m.Ep1(i,:); m.Ep2(i,:); m.Ep3(i,:); m.Ep4(i,:)];
                
                % get Epf null space
                [~,~,V] = svd(Epf');
                C = V(:,3:4); % since polygon is planar, kernel is size 2
                % C = null(Epf'); null sometimes misses vectors in kernel
                
                % make the matrix full rank
                mf = mfg + C*C'*m.lambda;

                % Construct indices for building M1
                for k=1:4
                    for l=1:4
                        ii(loc) = e(i,k);
                        jj(loc) = e(i,l);
                        ss(loc) = mf(k,l)*s(i,k)*s(i,l);
                        loc = loc+1;
                    end
                end
                
                Mf_g(i,:) = mfg(:);
                Mf(i,:) = mf(:);                
            end

            M1 = sparse(ii,jj,ss,m.ne,m.ne);
              
        end

        function [va, M0] = vertex_mass_matrix( mesh )
            % va, in nv x 1
            % M0, in nv x nv, diagonal
            
            I = double( mesh.quads ); I = I(:);
            J = repmat(1:mesh.nf,4,1)'; J = J(:);
            S = repmat(mesh.fa/4,4,1);
            
            f2v = sparse(I,J,S,mesh.nv,mesh.nf);
            
            va = f2v*ones(mesh.nf,1);
            M0 = spdiags(va,0,mesh.nv,mesh.nv);
        end
        
        function [d0, W, L] = differential_operators ( mesh ) 
            % d0, ne x nv
            % W, nv x nv, scale invariant Laplacian
            % L, nv x nv, Laplacian
            
            ii = [1:mesh.ne, 1:mesh.ne]';
            jj = [mesh.edges(:,1); mesh.edges(:,2)];
            ss = [-ones(mesh.ne,1); ones(mesh.ne,1)];
            d0 = sparse(ii,jj,ss,mesh.ne,mesh.nv);
            
            [~, ~, mesh.M1] = mesh.edge_mass_matrix;
            W = d0'*mesh.M1*d0;
            L = spdiags(1./mesh.va,0,mesh.nv,mesh.nv)*W;
        end
        
        function [subdividedMesh, S_0, S_0_trivial] = subdivideMesh(mesh) % Subdivide the mesh
            % subdivideMesh  Subdivide the mesh with Catmull Clark rules for 0-forms
            
            % SUBDIVIDE GEOMETRY - Catmull-Clarck subdivision
            % subdivision CC
            [VertexPoints, EdgePoints, ~, ~, t2e, FacePoints, I,J,V]=GetCCMatrixEigen(mesh.vertices,double(mesh.quads));
            % build S_0 
            S_0 = sparse(I+1,J+1,V, mesh.nv+mesh.ne+mesh.nf, mesh.nv);
            % arrange new vertices
            subdivision_vertices =[VertexPoints;EdgePoints;FacePoints];
            % arrange new quads
            subdivision_quads = [mesh.quads(:,1) mesh.nv+abs(t2e(:,1)) mesh.nv+mesh.ne+(1:mesh.nf)' mesh.nv+abs(t2e(:,4));...
                mesh.quads(:,2) mesh.nv+abs(t2e(:,2)) mesh.nv+mesh.ne+(1:mesh.nf)' mesh.nv+abs(t2e(:,1));...
                mesh.quads(:,3) mesh.nv+abs(t2e(:,3)) mesh.nv+mesh.ne+(1:mesh.nf)' mesh.nv+abs(t2e(:,2));...
                mesh.quads(:,4) mesh.nv+abs(t2e(:,4)) mesh.nv+mesh.ne+(1:mesh.nf)' mesh.nv+abs(t2e(:,3))];
            
             % return the subdivided mesh
            subdividedMesh = MESHQ(subdivision_vertices, subdivision_quads);
            fprintf('input nv = %d, output nv = %d\n', mesh.nv, subdividedMesh.nv);
            
            % SUBDIVIDE UV COORDINATES - Trivial subdivision
            if ~isempty(mesh.nv_uv)
                % subdivision QUADS - TRIVIAL
                [VertexPoints, EdgePoints, ~, ~, t2e, FacePoints, I, J, V]=GetTrivialMatrixEigen([mesh.uv, zeros(mesh.nv_uv,1)],double(mesh.quads_uv));
                % build S_0 Trivial
                S_0_trivial = sparse(I+1,J+1,V, mesh.nv_uv + mesh.ne_uv + mesh.nf, mesh.nv_uv);
                % arrange new UV vertices
                subdivision_vertices_uv = [VertexPoints;EdgePoints;FacePoints];
                % arrange new UV quads
                subdivision_quads_uv = [mesh.quads_uv(:,1) mesh.nv_uv+abs(t2e(:,1)) mesh.nv_uv+mesh.ne_uv+(1:mesh.nf)' mesh.nv_uv+abs(t2e(:,4));...
                    mesh.quads_uv(:,2) mesh.nv_uv+abs(t2e(:,2)) mesh.nv_uv+mesh.ne_uv+(1:mesh.nf)' mesh.nv_uv+abs(t2e(:,1));...
                    mesh.quads_uv(:,3) mesh.nv_uv+abs(t2e(:,3)) mesh.nv_uv+mesh.ne_uv+(1:mesh.nf)' mesh.nv_uv+abs(t2e(:,2));...
                    mesh.quads_uv(:,4) mesh.nv_uv+abs(t2e(:,4)) mesh.nv_uv+mesh.ne_uv+(1:mesh.nf)' mesh.nv_uv+abs(t2e(:,3))];

                % update uv coordinates
                subdividedMesh.uv = subdivision_vertices_uv(:,1:2);
                subdividedMesh.nv_uv = size(subdivision_vertices_uv,1);
                subdividedMesh.quads_uv = subdivision_quads_uv;
                [subdividedMesh.AM_uv, subdividedMesh.ne_uv] = subdividedMesh.adj_mat_uv;
                
            end
                

        end
        
        function [basis_hierarchical, t_vec] = lbo_basis_hierarchical_SDEC_S0(mesh, num_eigen_vec, subdivision_levels, flag_full_basis) % compute SDEC lbo eigen basis hierarchical
            % lbo_basis_hierarchical_SDEC  Compute the num_eigen_vec sub basis of the LBO
            
            % timings
            t_vec = zeros(subdivision_levels + 1, 1);
            eigs_timings = 0;
            
            % check required dimensions
            assert(nnz(num_eigen_vec) == subdivision_levels+1, 'Err number of subdivision levels is not equal to size of eigen vecs required');
            
            % initialize
            basis_hierarchical = cell(subdivision_levels+1, 1);
            
            
            %% get SEC operators
            % copy for subdivision
            shape_subdiv = copy(mesh);
            
            % initialize accumulation matrices
            S_0_cell = cell(1,subdivision_levels);            
            shape_subdiv_cell = cell(1,subdivision_levels);
            
            % subdivide
            % tic;
            for l=1:subdivision_levels
                
                t = tic;
                
                % subdivide and get SEC S_0 matrices
                disp(['subdividing to level ', num2str(l), '/', num2str(subdivision_levels)]);
                [shape_subdiv, S_0] = shape_subdiv.subdivideMesh;
                
                % update S_0 cell array and shape_subdiv cell array
                S_0_cell{l} = S_0;
                shape_subdiv_cell{l} = shape_subdiv;
                
                t_vec(l) = t_vec(l) + toc(t);
                
            end
            % toc;
            
            %% Get M0 and W of the hierarchy
            % initiate cell arrays to contain all M0 and W in the
            % hierarchy.
            disp('Preparing SEC M0 and W for subdivision levels');
            % tic;
            M0_cell = cell(1, subdivision_levels+1);
            W_cell  = cell(1, subdivision_levels+1);
            % M0 SEC and W DEC of the finest level
            t = tic;
            if subdivision_levels > 0 
                M0_cell{subdivision_levels+1} = shape_subdiv_cell{end}.M0;
                [~, W_cell{subdivision_levels+1}, ~] = shape_subdiv_cell{end}.differential_operators;
            else
                M0_cell{subdivision_levels+1} = mesh.M0;
                [~, W_cell{subdivision_levels+1}, ~] = mesh.differential_operators;
            end
            t_vec(end) = t_vec(end) + toc(t);
            
            % run over levels and build M0 and W. From fine to coarse M0 =
            % S' * ... * S' * M0_DEC_fine * S *...*S
            for i = subdivision_levels:-1:1
                
                t = tic;
                
                % get current level S_0 and S_1
                S_0 = S_0_cell{i};
                
                % update M0 to this level
                M0_cell{i} = S_0'*M0_cell{i+1}*S_0;
                
                % update W to this level
                W_cell{i} = S_0'*W_cell{i+1}*S_0;
                
                t_vec(i) = t_vec(i) + toc(t);
                
            end
            % toc;
            
            
            %% compute basis for each hierarchy level - starting from coarset level
            for i = 1:subdivision_levels+1
                
                t = tic;
                
                fprintf('Compute SEC basis for sbd level %d/%d', i-1, subdivision_levels);
                
                % get current shape
                if flag_full_basis == 0 % basis prolongation
                    if i == 1 % coarset shape
                        shape = mesh;
                        num_eigen = num_eigen_vec(i);
                        eigs_addition = 0;
                        sigma = -1e-5;
                        prev_starting_eig_ind = 1;
                        curr_starting_eig_ind = prev_starting_eig_ind + num_eigen;
                    else % subdivided shape
                        shape = shape_subdiv_cell{i-1};
                        num_eigen = num_eigen_vec(i);
                        eigs_addition = 15; % computing extra to get exatcly the demanded num of eigs
                        eigs_back = 0; % taking back 5 eigens for linear fit, to be sure we don't skip eigens
                        % aprroximate linear line lambda(i) = m*i
                        lin_fit = polyfit(prev_starting_eig_ind:1:curr_starting_eig_ind-1, basis_hierarchical{i-1}.eigenvalues', 1);
                        lin_val = polyval(lin_fit, curr_starting_eig_ind-eigs_back:1:curr_starting_eig_ind + num_eigen -1);
                        sigma = (lin_val(end) + lin_val(1))/2; % take the middle (one value from the end, else could miss eigens)
                    end
                else
                    num_eigen = num_eigen_vec(i);
                    eigs_addition = 0;
                    sigma = -1e-5;
                    fprintf(' %d eigens \n', num_eigen);
                    if i == 1 % coarset shape
                        shape = mesh;
                    else % subdivided shape
                        shape = shape_subdiv_cell{i-1};

                    end
                end
                
                % get current operators and matrices
                M0_matrix = M0_cell{i};
                M0_matrix = (M0_matrix + M0_matrix')/2; %verify symmetry
                
                % comptute SDEC LBO (symmetric) operator
                W_SEC = W_cell{i};
                W_SEC = (W_SEC+W_SEC')/2; %verify symmetry
                
                % compute SDEC eigenfunctions
                [V_SEC, D_SEC, eigs_flag] = eigs(W_SEC, M0_matrix, num_eigen + eigs_addition, sigma, 'IsSymmetricDefinite', 1);
                
                % verify eigenvalues convergence
                assert(eigs_flag == 0, 'Error: eigenvalues did not converge');
                
                % verify A*V = B*V*D, calculate frobenius norm
                solution_norm = norm(W_SEC*V_SEC - M0_matrix*V_SEC*D_SEC, 'fro');
                assert(solution_norm < 1e-4, 'Error: A*V != B*V*D');
                
                % verify orthogonality with respect to mass matrix
                orthogonality_norm = norm(V_SEC'*M0_matrix*V_SEC - speye(num_eigen+eigs_addition) ,'fro');
                assert(orthogonality_norm < 1e-10, 'Error: not orthonormal with respect to mass');
                
                % sort ascend
                [vals, inds] = sort(diag(D_SEC));
                V_SEC = V_SEC(:, inds);
                
                % check if there are complex eigen values
                assert(all(imag(vals) == 0), 'Error: complex eigenvalues');
                
                % return the lbo eigen functions/values, W, mass matrix,
                % subdivision operator for 0-forms (vertices)
                % return in struct
                if flag_full_basis == 0 % prolonged basis
                    if i == 1 % low pass of the coarset level
                        fprintf(', computed %d eigens starting from 0 \n', num_eigen);
                        basis_hierarchical{i}.eigenfunctions = V_SEC;
                        basis_hierarchical{i}.eigenvalues = vals;
                    else % take just bandpass
                        % eliminating double eigs using inner products
                        prev_basis_prolonged = basis_hierarchical{i-1}.S_0 * basis_hierarchical{i-1}.eigenfunctions;
                        prev_basis_new_basis_inner_prod = prev_basis_prolonged' * M0_matrix * V_SEC;
                        % identify the overlapping eigens by inner product
                        % larger than 0.4 (there could be rotations)
                        inds_double_eigs = find(max(abs(prev_basis_new_basis_inner_prod)) > 0.4);
                        assert(~isempty(inds_double_eigs) , 'Error: computed too little eigs of previous level, could be a gap');
                        assert(numel(inds_double_eigs) <= eigs_addition, 'Error: computed too much eigs of previous level, could not provide the requested number of eigs');
                        inds_to_keep = setdiff(1:num_eigen+eigs_addition, inds_double_eigs); % remove indices of double eigs
                        if numel(inds_to_keep) > num_eigen % verify we take only the wanted num_eigen
                            inds_to_keep = inds_to_keep(1:num_eigen);
                        end
                        
                        basis_hierarchical{i}.eigenfunctions = V_SEC(:, inds_to_keep);
                        basis_hierarchical{i}.eigenvalues = vals(inds_to_keep);
                        prev_starting_eig_ind = curr_starting_eig_ind;
                        curr_starting_eig_ind = curr_starting_eig_ind + num_eigen;
                        fprintf(', computed %d eigens, using %d, overlap %d \n',num_eigen + eigs_addition, num_eigen, numel(inds_double_eigs));
                        
                        % update previous basis prolongation
                        basis_hierarchical{i-1}.eigenfunctions_prolonged = prev_basis_prolonged;
                    end
                else
                    basis_hierarchical{i}.eigenfunctions = V_SEC;
                    basis_hierarchical{i}.eigenvalues = vals;
                end
                
                
                % construct the basis struct for this level
                basis_hierarchical{i}.shape = shape;
                basis_hierarchical{i}.W = W_SEC;
                basis_hierarchical{i}.A = M0_matrix;
                if i < subdivision_levels+1
                    basis_hierarchical{i}.S_0 = S_0_cell{i};
                    basis_hierarchical{i}.S_to_fine = S_0_cell{i};
                    if i > 1
                        for jj = 1:i-1
                            basis_hierarchical{jj}.S_to_fine = S_0_cell{i} * basis_hierarchical{jj}.S_to_fine;
                        end
                    end
                else
                    basis_hierarchical{i}.S_to_fine = speye(shape.nv);
                end
                basis_hierarchical{i}.num_eigs = numel(basis_hierarchical{i}.eigenvalues);
                
                t_vec(i) = t_vec(i) + toc(t);
           
            end
                        
        end
        
        function basis_hierarchical = lbo_basis_hierarchical_from_coarse(mesh, num_eigen_vec, subdivision_levels) % compute SDEC lbo eigen basis hierarchical
            % lbo_basis_hierarchical_SDEC  Compute the num_eigen_vec sub basis of the LBO
            
            % check required dimensions
            assert(numel(num_eigen_vec) == subdivision_levels+1, 'Err number of subdivision levels is not equal to size of eigen vecs required');
            
            % initialize
            basis_hierarchical = cell(subdivision_levels+1, 1);
            
            %% get SEC operators
            % copy for subdivision
            shape_subdiv = copy(mesh);
            
            % initialize accumulation matrices
            S_0_cell = cell(1,subdivision_levels);
            shape_subdiv_cell = cell(1,subdivision_levels);
            
            % subdivide
            % tic;
            for l=1:subdivision_levels
                
                % subdivide and get SEC S_0 matrices
                disp(['subdividing to level ', num2str(l), '/', num2str(subdivision_levels)]);
                [shape_subdiv, S_0] = shape_subdiv.subdivideMesh;
                
                % update S_0 cell array and shape_subdiv cell array
                S_0_cell{l} = S_0;
                shape_subdiv_cell{l} = shape_subdiv;
                
            end
            % toc;
            
            %% Get M0 and W of the hierarchy
            % initiate cell arrays to contain all M0 and W in the
            % hierarchy.
            disp('Preparing SEC M0 and W for subdivision levels');
            % tic;
            M0_cell = cell(1, subdivision_levels+1);
            W_cell  = cell(1, subdivision_levels+1);
            % M0 SEC and W DEC of the finest level
            if subdivision_levels > 0
                M0_cell{subdivision_levels+1} = shape_subdiv_cell{end}.M0;
                [~, W_cell{subdivision_levels+1}, ~] = shape_subdiv_cell{end}.differential_operators;
            else
                M0_cell{subdivision_levels+1} = mesh.M0;
                [~, W_cell{subdivision_levels+1}, ~] = mesh.differential_operators;
            end
            % run over levels and build M0 and W. From fine to coarse M0 =
            % S' * ... * S' * M0_DEC_fine * S *...*S
            for i = subdivision_levels:-1:1
                
                % get current level S_0 and S_1
                S_0 = S_0_cell{i};
                
                % update M0 to this level
                M0_cell{i} = S_0'*M0_cell{i+1}*S_0;
                
                % update W to this level
                W_cell{i} = S_0'*W_cell{i+1}*S_0;
                
            end
            % toc;
            
            
            %% compute basis for each hierarchy level - starting from coarset level
            %tic;
            for i = 1:subdivision_levels+1
                
                fprintf('Compute SEC basis for sbd level %d/%d \n', i-1, subdivision_levels);
                
                % get current operators and matrices
                M0_matrix = M0_cell{i};
                M0_matrix = (M0_matrix + M0_matrix')/2; %verify symmetry
                
                % comptute SDEC LBO (symmetric) operator
                W_SEC = W_cell{i};
                W_SEC = (W_SEC+W_SEC')/2; %verify symmetry
                
                % get current shape
                if i == 1 % coarset shape
                    shape = mesh;
                    num_eigen = sum(num_eigen_vec);
                    eigs_addition = 0;
                    sigma = -1e-5;
                    
                    % compute SDEC eigenfunctions
                    [V_SEC, D_SEC, eigs_flag] = eigs(W_SEC, M0_matrix, num_eigen + eigs_addition, sigma, 'IsSymmetricDefinite', 1);
                    
                    % verify eigenvalues convergence
                    assert(eigs_flag == 0, 'Error: eigenvalues did not converge');
                    
                    % verify A*V = B*V*D, calculate frobenius norm
                    solution_norm = norm(W_SEC*V_SEC - M0_matrix*V_SEC*D_SEC, 'fro');
                    assert(solution_norm < 1e-4, 'Error: A*V != B*V*D');
                    
                    % verify orthogonality with respect to mass matrix
                    orthogonality_norm = norm(V_SEC'*M0_matrix*V_SEC - speye(num_eigen+eigs_addition) ,'fro');
                    assert(orthogonality_norm < 1e-10, 'Error: not orthonormal with respect to mass');
                    
                    % sort ascend
                    [vals, inds] = sort(diag(D_SEC));
                    V_SEC = V_SEC(:, inds);
                    
                    % check if there are complex eigen values
                    assert(all(imag(vals) == 0), 'Error: complex eigenvalues');
                
                else
                    shape = shape_subdiv_cell{i-1};
                    
                end
                

                
                % return the lbo eigen functions/values, W, mass matrix,
                % subdivision operator for 0-forms (vertices)
                % return in struct
                if i == 1
                    basis_hierarchical{i}.eigenfunctions = V_SEC(:, 1:sum(num_eigen_vec(1:i)));
                else
                    basis_hierarchical{i}.eigenfunctions = basis_hierarchical{1}.S_to_fine * V_SEC(:, 1:sum(num_eigen_vec(1:i)));
                end
                basis_hierarchical{i}.eigenvalues = vals(1:sum(num_eigen_vec(1:i)));
                basis_hierarchical{i}.shape = shape;
                basis_hierarchical{i}.W = W_SEC;
                basis_hierarchical{i}.A = M0_matrix;
                if i < subdivision_levels+1
                    basis_hierarchical{i}.S_0 = S_0_cell{i};
                    basis_hierarchical{i}.S_to_fine = S_0_cell{i};
                    if i > 1
                        for jj = 1:i-1
                            basis_hierarchical{jj}.S_to_fine = S_0_cell{i} * basis_hierarchical{jj}.S_to_fine;
                        end
                    end
                else
                    basis_hierarchical{i}.S_to_fine = speye(shape.nv);
                end
                basis_hierarchical{i}.num_eigs = numel(basis_hierarchical{i}.eigenvalues);
                
            end
            %toc;
            
        end

        
    end
    
    methods (Static)
        function norm_V = normv(V)
            % input: 
            % V is N x k. N rows of vectors in dimension k
            % output: 
            % norm_V is N x 1. The i entry is the i'th vector L2 norm
            % (length).
            
            norm_V = sqrt(sum(V.^2,2)); 
        end
        
        function normalized_VF = normalize_vf(VF)
            % input:
            % VF is N x k. N rows of vectors in dimension k
            % output: 
            % normalized_VF is N x k. The i entry is the normalized i'th vec.
            
            normalized_VF = VF./MESHQ.normv(VF);
        end
    end
end



%%%% Junkyard ===============
% Alternative to loop in edge_mass_matrix, x2 slower :(
%             [II,JJ] = meshgrid(1:4,1:4); II = II(:); JJ = JJ(:);
%                       
%             for i=1:m.nf
%                 Bf = [m.B1(i,:); m.B2(i,:); m.B3(i,:); m.B4(i,:)];
%                 mfg = Bf*Bf'/m.fa(i)*sqrt(2);
% 
%                 Epf = [m.Ep1(i,:); m.Ep2(i,:); m.Ep3(i,:); m.Ep4(i,:)];
%                 C = null(Epf');
%                 mf = mfg + C*C'*m.lambda;
%                 
%                 tic
%                 sf = s(i,:)'*s(i,:);
%                 mfs = mf.*sf;
%                 
%                 % Construct indices for building M1               
%                 ii(loc:loc+15) = e(i,II)';
%                 jj(loc:loc+15) = e(i,JJ)';
%                 ss(loc:loc+15) = mfs(:)';
%                 loc = loc+16;
%                 toc
%                 
%                 Mf_g(i,:) = mfg(:);
%                 Mf(i,:) = mf(:);                
%             end
