% This script computes Hierarchical Functional Maps between Subdivision
% Surfaces, the code used to produce the results in the paper by Meged
% Shoham, Amir Vaxman, Mirela Ben-Chen. Presented in SGP 2019.

clc;
close all;
clearvars;

addpath(genpath('QUADS_class'));
addpath('TRIANGLES_class');
addpath(genpath('external'));
addpath('utils');
path_data = '..\Hierarchical_FMaps_Subd_data\';
addpath(genpath(path_data));

%% Parameters
num_energies = 100; % spectral descriptors
alpha1 = 1; % descriptors preservation C_12*F_2 = F_1
alpha2 = 1e-2; % Laplacian commutation C_12*Lambda_2 = Lambda_1*C_12
alpha3 = 1e4; % Symmetry commutation C_12*Symmetry_2 = Symmetry_1*C_12
alpha4 = 0; % Descriptors multiplication commutation C_12*X_2 = X_1*C_12 - Nognen et al. 2017, "Informative descriptor preservation via commutativity for shape matching"
num_commuting = 10; % number of descriptors per energy for multiplication commutation (downsample for speed if alpha4 > 0)

flag_landmarks = 1; % 0 - landmaraks tagged from coarse level, 1 - landmarks tagged from fine level
flag_post_process_subdivision_ICP = 1; % make the final P_12 row stochastic with more than 1 nnz entry per row
flag_write = 1; % write results or not
flag_dataset = 8; % 1: woman-man, 2: zebra-horse, 3: tiger-cat, 4: elephant_mammoth, 5:trollDM-troll, 7: trollDM-orc, 8: troll-troll


%% Dataset
switch flag_dataset
    
    case 1 % woman-man
        
        num_sbd_1 = 3;
        num_sbd_2 = 3;
        k_1_vec = [100, 50, 50, 50];
        k_2_vec = [100, 50, 50, 50];
        
        path_results = '..\Hierarchical_FMaps_Subd_results\woman_man\corr\';
        
        shape_1_name = 'man_lowpoly_Apose';
        shape_2_name = 'woman_lowpoly_Apose';
        type = 'obj';
        
        shape_2_texture_im = [];
        
        landmarks = dlmread([path_results 'landmarks_clean.txt'], ',');
        
    case 2 % zebra-horse
                
        num_sbd_1 = 4;
        num_sbd_2 = 3;
        k_1_vec = [100, 50, 50, 50, 50];
        k_2_vec = [100, 0, 100, 50, 50];
        
        path_results = '..\Hierarchical_FMaps_Subd_results\zebra_horse\corr\';
        
        shape_1_name = 'horse_no_eyes_s0';
        shape_2_name = 'Zebra_reduced_rot_s0';
        type = 'obj';
        
        shape_2_texture_im = 'Zebra_Color_Map.tif';
        
        landmarks = dlmread([path_results 'landmarks_clean.txt'], ',');
        
    case 3 % tiger-cat
                
        num_sbd_1 = 4;
        num_sbd_2 = 3;
        k_1_vec = [100, 50, 50, 50, 50];
        k_2_vec = [100, 0, 100, 50, 50];
        
        path_results = '..\Hierarchical_FMaps_Subd_results\tiger_cat\corr\';
        
        shape_1_name = 'cat_s0';
        shape_2_name = 'Tiger_s0';
        type = 'obj';
        
        shape_2_texture_im = 'Tiger_Color_Map.tif';
        
        landmarks = dlmread([path_results 'landmarks_clean.txt'], ',');
        
    case 4 % elephant-mammoth
                
        num_sbd_1 = 4;
        num_sbd_2 = 3;
        k_1_vec = [100, 50, 50, 50, 50];
        k_2_vec = [100, 0, 100, 50, 50];
        
        path_results = '..\Hierarchical_FMaps_Subd_results\elephant_mammoth\corr\';
        
        shape_1_name = 'mammoth_s0';
        shape_2_name = 'elephant_s0';
        type = 'obj';
        
        shape_2_texture_im = 'elephant_colorG.tif';
        
        landmarks = dlmread([path_results 'landmarks_clean.txt'], ',');
        
    case 5 % trollDM-troll
                
        num_sbd_1 = 3;
        num_sbd_2 = 3;
        k_1_vec = [100, 50, 50, 50];
        k_2_vec = [100, 50, 50, 50];
        
        path_results = '..\Hierarchical_FMaps_Subd_results\troll_troll\corr\';
        
        shape_1_name = 'troll_s0';
        shape_2_name = 'troll_DM_s0';
        type = 'obj';
        
        shape_2_texture_im = 'TROLL_TM.tif';
        
        landmarks = dlmread([path_results 'landmarks_clean.txt'], ',');
        
    case 7 % trollDM-orc
                
        num_sbd_1 = 3;
        num_sbd_2 = 3;
        k_1_vec = [100, 50, 50, 50];
        k_2_vec = [100, 50, 50, 50];
        
        path_results = '..\Hierarchical_FMaps_Subd_results\trollDM_orc\corr\';
        
        shape_1_name = 'orc_s0';
        shape_2_name = 'troll_DM_s0';
        type = 'obj';
        
        shape_2_texture_im = 'TROLL_TM.tif';
        
        landmarks = dlmread([path_results 'landmarks_clean.txt'], ',');
        
    case 8 % troll-orc
                
        num_sbd_1 = 3;
        num_sbd_2 = 3;
        k_1_vec = [100, 50, 50, 50];
        k_2_vec = [100, 50, 50, 50];
        
        path_results = '..\Hierarchical_FMaps_Subd_results\troll_orc\corr\';
        
        shape_1_name = 'orc_s0';
        shape_2_name = 'troll_s0';
        type = 'obj';
        
        shape_2_texture_im = 'troll_texture_transfer.png';
        
        landmarks = dlmread([path_results 'landmarks_clean.txt'], ',');
        
end


%% Prepare outputs struct and timings
mapping.source = shape_2_name;
mapping.target = shape_1_name;
mapping.sudbvision_levels_source = num_sbd_2;
mapping.sudbvision_levels_target = num_sbd_1;
mapping.P_12 = cell(num_sbd_1 + 2, 0);

% timings
[t_B, t_DS, t_C, t_P, t_P_post] = deal(0);
[t_B_levels_1, t_DS_levels_1] = deal(zeros(num_sbd_1+1,1));
[t_B_levels_2, t_DS_levels_2] = deal(zeros(num_sbd_2+1,1));
[t_C_levels, t_P_levels] = deal(zeros(max(num_sbd_2,num_sbd_1)+1,1));


%% Load shapes
fprintf('M1: %s \n', shape_1_name);
fprintf('M2: %s \n', shape_2_name);
% read
shape_1 = MESHQ(shape_1_name, type);
shape_2 = MESHQ(shape_2_name, type);
% normalize to unit area
shape_1 = MESHQ((shape_1.vertices - mean(shape_1.vertices))./sqrt(sum(shape_1.va)), shape_1.quads, shape_1.uv, shape_1.quads_uv);
shape_2 = MESHQ((shape_2.vertices - mean(shape_2.vertices))./sqrt(sum(shape_2.va)), shape_2.quads, shape_2.uv, shape_2.quads_uv);

fprintf('\n');


%% compute hierarchical bases
tic;

disp('Compute subdivision hierarchy for shape 1');
[basis_1_hierarchy, timings_basis_vec_1] = shape_1.lbo_basis_hierarchical_SDEC_S0(k_1_vec(k_1_vec>0), num_sbd_1, 0);
t_B_levels_1 = t_B_levels_1 + timings_basis_vec_1;
fprintf('\n');

disp('Compute subdivision hierarchy for shape 2');
[basis_2_hierarchy, timings_basis_vec_2] = shape_2.lbo_basis_hierarchical_SDEC_S0(k_2_vec(k_2_vec>0), num_sbd_2, 0);
t_B_levels_2 = t_B_levels_2 + timings_basis_vec_2;
fprintf('\n');

t_B = t_B + toc;


%% compute C_12 & P_12 for each level
for i = 1:numel(k_1_vec)
    
    fprintf('\nSolve for subdivision level %d/%d \n', i-1, num_sbd_1)
    
    % get common number of eigs in both shapes (for descriptors)
    num_eigs_common = min(sum(k_1_vec(1:i)),sum(k_1_vec(1:i)));
    
    % Shape 1 basis and constraints
    fprintf('Prepare shape 1\n');
    if k_1_vec(i) > 0
        
        % hierarchy index of this level
        hierarchy_1_ind = nnz(k_1_vec(1:i));
        
        tic;
        fprintf('Prolong basis and compute psuedo inverse \n');
        % get basis
        basis_1 = basis_1_hierarchy{hierarchy_1_ind};
        % prolong basis
        if i == 1
            % subd matrix (coarse level just dummy speye for consitency in functions)
            S_0_1 = speye(size(basis_1.eigenfunctions,1));
            % coarse level eigens
            basis_1_eigenfunctions_prol = basis_1.eigenfunctions;
            eigenvals_1_prol = basis_1.eigenvalues;
            % coarse level pinv
            basis_1_prol_pinv = basis_1.eigenfunctions' * basis_1.A;
            
        else
            % subd matrix
            S_0_1 = basis_1_hierarchy{hierarchy_1_ind-1}.S_0;
            % eigenvalues concating
            eigenvals_1_prol = [eigenvals_1_prol; basis_1.eigenvalues];
            % prolong bases
            basis_1_eigenfunctions_prol = [S_0_1 * basis_1_eigenfunctions_prol, basis_1.eigenfunctions];
            % prolong bases pinv
            basis_1_prol_pinv = pinv(basis_1_eigenfunctions_prol);
        end
        t = toc;
        t_B_levels_1(hierarchy_1_ind) = t_B_levels_1(hierarchy_1_ind) + t;
        t_B = t_B + t;
        
        % compute self-symmetry operator
        fprintf('Compute self-symmetry operators \n');
        tic;
        if i == 1
            % for coarse find by symmetry axis
            [symmetry_F_1, symmetry_1] = compute_symmetry(basis_1.shape, [], [], basis_1_eigenfunctions_prol, basis_1_prol_pinv, 1, []);
        else
            % for subdivision levels, find by previous symmetry operator
            [symmetry_F_1, symmetry_1] = compute_symmetry(basis_1.shape, basis_1_hierarchy{hierarchy_1_ind - 1}.shape, S_0_1, basis_1_eigenfunctions_prol, basis_1_prol_pinv, [], symmetry_1);
        end
        t = toc;
        t_DS_levels_1(hierarchy_1_ind) = t_DS_levels_1(hierarchy_1_ind) + t;
        t_DS = t_DS + t;
        
        % compute WK descriptors cofficients
        fprintf('Compute descriptors \n');
        tic;
        F_1 = WK_descriptors(basis_1_eigenfunctions_prol, eigenvals_1_prol, basis_1_prol_pinv, basis_1.A, num_energies, num_eigs_common, landmarks(1,:), flag_landmarks, basis_1_hierarchy{hierarchy_1_ind}.S_to_fine);
        t = toc;
        t_DS_levels_1(hierarchy_1_ind) = t_DS_levels_1(hierarchy_1_ind) + t;
        t_DS = t_DS + t;
        
        % compute commuting descriptors term cofficients for coarse level
        if alpha4 > 0
            if i == 1
                tic;
                fprintf('Compute commuting multiplication descriptors \n');
                commuting_F_1 = commuting_descriptors(basis_1_eigenfunctions_prol, basis_1_prol_pinv, F_1, num_commuting);
                t = toc;
                t_DS_levels_1(hierarchy_1_ind) = t_DS_levels_1(hierarchy_1_ind) + t;
                t_DS = t_DS + t;
            end
        else
            commuting_F_1 = [];
        end
    end
    
    % Shape 2 basis and constraints
    fprintf('Prepare shape 2 \n');
    if k_2_vec(i) > 0
        
        % hierarchy index of this level
        hierarchy_2_ind = nnz(k_2_vec(1:i));
        
        tic;
        fprintf('Prolong basis and compute psuedo inverse \n');
        % get basis
        basis_2 = basis_2_hierarchy{hierarchy_2_ind};
        % prolong basis
        if i == 1
            % subd matrix (coarse level just dummy speye for consitency in functions)
            S_0_2 = speye(size(basis_2.eigenfunctions,1));
            % coarse level eigens
            basis_2_eigenfunctions_prol = basis_2.eigenfunctions;
            eigenvals_2_prol = basis_2.eigenvalues;
            % coarse level pinv
            basis_2_prol_pinv = basis_2.eigenfunctions' * basis_2.A;
            
        else
            % subd matrix
            S_0_2 = basis_2_hierarchy{hierarchy_2_ind-1}.S_0;
            % eigenvalues concating
            eigenvals_2_prol = [eigenvals_2_prol; basis_2.eigenvalues];
            % prolong bases
            basis_2_eigenfunctions_prol = [S_0_2 * basis_2_eigenfunctions_prol, basis_2.eigenfunctions];
            % prolong bases pinv
            basis_2_prol_pinv = pinv(basis_2_eigenfunctions_prol);
        end
        t = toc;
        t_B_levels_2(hierarchy_2_ind) = t_B_levels_2(hierarchy_2_ind) + t;
        t_B = t_B + t;
        
        % compute self-symmetry operator
        fprintf('Compute self-symmetry operators \n');
        tic;
        if i == 1
            % for coarse find by symmetry axis
            [symmetry_F_2, symmetry_2] = compute_symmetry(basis_2.shape, [], [], basis_2_eigenfunctions_prol, basis_2_prol_pinv, 1, []);
        else
            % for subdivision levels, find by previous symmetry operator
            [symmetry_F_2, symmetry_2] = compute_symmetry(basis_2.shape, basis_2_hierarchy{hierarchy_2_ind-1}.shape, S_0_2, basis_2_eigenfunctions_prol, basis_2_prol_pinv, [], symmetry_2);
        end
        t = toc;
        t_DS_levels_2(hierarchy_2_ind) = t_DS_levels_2(hierarchy_2_ind) + t;
        t_DS = t_DS + t;
        
        % compute WK descriptors cofficients
        fprintf('Compute descriptors \n');
        tic;
        F_2 = WK_descriptors(basis_2_eigenfunctions_prol, eigenvals_2_prol, basis_2_prol_pinv, basis_2.A, num_energies, num_eigs_common, landmarks(2,:), flag_landmarks, basis_2_hierarchy{hierarchy_2_ind}.S_to_fine);
        t = toc;
        t_DS_levels_2(hierarchy_2_ind) = t_DS_levels_2(hierarchy_2_ind) + t;
        t_DS = t_DS + t;
        
        % compute commuting descriptors term cofficients for coarse level
        if alpha4 > 0
            if i == 1
                tic;
                fprintf('Compute commuting multiplication descriptors \n');
                commuting_F_2 = commuting_descriptors(basis_2_eigenfunctions_prol, basis_2_prol_pinv, F_2, num_commuting);
                t = toc;
                t_DS_levels_2(hierarchy_2_ind) = t_DS_levels_2(hierarchy_2_ind) + t;
                t_DS = t_DS + t;
            end
        else
            commuting_F_2 = [];
        end
    end
    
    % compute fmap
    fprintf('Solve FMAP \n');
    F_1_mat = cell2mat(F_1);
    F_2_mat = cell2mat(F_2);
    if i == 1 % coarset level
        tic;
        C_12 = compute_fmap(F_1_mat, F_2_mat, basis_1.eigenvalues, basis_2.eigenvalues, alpha1, alpha2, alpha3, alpha4, symmetry_F_1, symmetry_F_2, commuting_F_1, commuting_F_2);
        t = toc;
        t_C_levels(i) = t;
        t_C = t_C + t;
    else % subdivided level
        % solve fmap hierarchaly
        k_1v = [sum(k_1_vec(1:i-1)), k_1_vec(i)];
        k_2v = [sum(k_2_vec(1:i-1)), k_2_vec(i)];
        tic;
        C_12 = compute_fmap_hierarchy(C_12, k_1v, k_2v, F_1_mat, F_2_mat, eigenvals_1_prol, eigenvals_2_prol, alpha1, alpha2, alpha3, symmetry_F_1, symmetry_F_2);
        t = toc;
        t_C_levels(i) = t;
        t_C = t_C + t;
    end
    
    % ICP refinement
    fprintf('Refine FMAP \n')
    tic;
    if i == 1 % coarset level
        [C_12, knn_matches] = icp_refine_coarse(C_12 ,basis_1.eigenfunctions, basis_2.eigenfunctions);
    else % subdivded level
        [C_12, knn_matches] = icp_refine_subd(C_12, knn_matches, basis_1_eigenfunctions_prol, basis_2_eigenfunctions_prol, S_0_1, S_0_2);
    end
    t = toc;
    t_P_levels(i) = t;
    t_P = t_P + t;
    
    % create the point2point matrix - row stochastic sum(P_12,2) = 1 for
    % all entries, sum(P_12,1) = 1 if the vertex is reached, 0 otherwise
    P_12 = sparse(1:basis_1.shape.nv, knn_matches(:,1), 1, basis_1.shape.nv, basis_2.shape.nv);
    
    % store map
    mapping.P_12{i} = P_12;
    
    % visualize transfer coordinates
    % transfer_coordinates(hierarchy_1_ind, hierarchy_2_ind, basis_1_hierarchy, basis_2_hierarchy, P_12);
    
    fprintf('\n');
end


%% Compute post-process subdivision ICP
% in the finest level add a final part of knnsearch
if flag_post_process_subdivision_ICP == 1
    tic;
    [P_12_post_process, P_12_p2p, shape_2_post_shape] = post_process_subdivision_knn(knn_matches(:,1), C_12, basis_1_eigenfunctions_prol, basis_2_eigenfunctions_prol, basis_2.shape);
    t = toc;
    %t_P_levels(i) = t_P_levels(i) + t;
    t_P_post = t_P_post + t;
    fprintf('\n');
    
end
mapping.P_12{num_sbd_1 + 2} = P_12_post_process;
mapping.P_12_p2p = P_12_p2p;

% visualize transfer coordinates
% transfer_coordinates(hierarchy_1_ind, hierarchy_2_ind, basis_1_hierarchy, basis_2_hierarchy, P_12_post_process);


%% Plot timings
fprintf('Timings B      (subdivision + basis):\n %.1f sec \n', t_B);
fprintf('Timings D+S    (descriptors):\n %.1f sec \n', t_DS);
fprintf('Timings C      (functional map):\n %.1f sec \n', t_C);
fprintf('Timings P      (p2p reconstruction):\n %.1f sec \n', t_P);
fprintf('Timings P_post (p2p reconstruction):\n %.1f sec \n', t_P_post);
fprintf('Timings T      (total):\n %.1f min \n', (t_B + t_DS + t_C + t_P + t_P_post)/60);
disp('Timings B levels shape 1'); disp(t_B_levels_1);
disp('Timings B levels shape 2'); disp(t_B_levels_2);
disp('Timings D+S levels shape 1'); disp(t_DS_levels_1);
disp('Timings D+S levels shape 2'); disp(t_DS_levels_2);
disp('Timings C levels'); disp(t_C_levels);
disp('Timings P levels'); disp(t_P_levels);

% write timings to text
dlmwrite([path_results, shape_1_name,'_',shape_2_name,'_TIMINGS.txt'], [t_B; t_DS; t_C; t_P; t_P_post; (t_B + t_DS + t_C + t_P + t_P_post)/60; t_B_levels_1; t_B_levels_2; t_DS_levels_1; t_DS_levels_2; t_C_levels; t_P_levels]);

%% save mapping results
if flag_write == 1
    save([path_results 'mappings.mat'], 'mapping');
    if ~isempty(shape_2_texture_im)
        transfer_texture(basis_1_hierarchy{end}.shape, shape_2_post_shape, P_12_p2p, [path_results '\texture\'], [shape_1_name], shape_2_texture_im)
    end
end

disp('Finished');
