classdef myMesh< matlab.mixin.Copyable
    % myMesh  Summary of myMesh
    % class the includes shared vertex (vertices,faces matrices) and
    % adjacency matrix and several methods
    %
    %   This class implements mesh object. The object includes the
    %   properties such as shared vertex representation and adjacency
    %   matrix. There are several methods implemented such as:...
    %   TODO: make a full documentation
    %
    %   The properties are:
    %       1. vertices: [Vx3] <double>: each row is a vertex [x,y,z]
    %       2. faces: [Fx3] <integer>: each row is the number of the
    %          composing vertices (indexed from 1)
    %       3. adjacencyMatrix: sparse [VxV] <boolean>: each element is
    %          either true or false where there is an edge between the
    %          vertices or not.
    %       4. numVertices: the number of vertices in the mesh
    %       5. numFaces: the number of faces in the mesh
    %       6. numEdges: the number of edges in the mesh
    %
    %   The methods are:
    %       1. obj = mesh(filename) - constructor, reading off file and
    %                                 updating properties 1-6
    %       2. write - write the mesh to off file
    %       3. visualize - visualize the mesh
    
    properties
        vertices               % Vertices matrix [Vx3]
        faces                  % Faces matrix [Fx3]
        adjacencyMatrix        % Vertices Adjacency Matrix [VxV]
        adjacencyVerticesFaces % Vertices-Faces Adjacency Matrix [VxF]
        numVertices            % The number of veritces V
        numFaces               % The number of faces F
        numEdges               % The number of edges E
        g                      % The graph - for edges
    end
    
    methods
        function obj = myMesh(varargin) % constructor
            % myMesh  Constructor
            % Usage 1 - read from file:
            %           M = myMesh('.../mesh.off');
            % Usage 2 - construct inside matlab:
            %           M = myMesh(vertices,faces);
            
            if nargin > 1
                % Input is vertices and faces inside matlab
                vertices = varargin{1};
                faces = varargin{2};
                % Insert the veritces and faces matrices
                obj.numVertices = size(vertices,1);
                obj.vertices = vertices;
                obj.numFaces = size(faces,1);
                obj.faces = faces; % assuming that counting starts form idx 1
            else
                % Read the off file
                filename = varargin{1};
                fileID = fopen(filename, 'r');
                data = textscan(fileID, '%f %f %f %f',   ...
                    'Delimiter',           ' ', ...
                    'CommentStyle',        '#', ...
                    'HeaderLines',         1,   ...
                    'CollectOutput',       1,   ...
                    'MultipleDelimsAsOne', 1);
                fclose(fileID);
                % Create the veritces and faces matrices
                obj.numVertices = data{1}(1,1);
                obj.vertices = data{1}(2:obj.numVertices+1,1:3);
                obj.numFaces = data{1}(1,2);
                obj.faces = data{1}(obj.numVertices+2:end,2:4) + 1; %+1 to fix the 0 counting of the off file
            end
            
            % Create the vertices adjacency matrix (by first creating a graph)
            % building the edges: edge(i) = s(i)<-->t(i)
            s = [obj.faces(:,1).', obj.faces(:,2).', obj.faces(:,3).'];
            t = [obj.faces(:,2).', obj.faces(:,3).', obj.faces(:,1).'];
            st = unique(sort([s;t]).', 'rows'); % the graph object does not accept duplicate edges
            obj.g = graph(st(:,1), st(:,2)); % the graph object contains the edges
            obj.adjacencyMatrix = adjacency(obj.g); % sparse matrix
            obj.numEdges = size(st,1);
            % TODO: this does not treat a vertex that is not on a face
            
            % Create the Vertices-Faces adjacency matrix
            i = [obj.faces(:,1); obj.faces(:,2); obj.faces(:,3)]; % vertices indices
            j = [1:obj.numFaces, 1:obj.numFaces, 1:obj.numFaces].'; % faces indices X 3
            v = ones(size(j)); % initiating to 1 - it's an adjacency matrix
            obj.adjacencyVerticesFaces = sparse(i, j, v, obj.numVertices, obj.numFaces);
            
        end
        function h = visualize(obj) % visualize the mesh
            clear S
            S.Vertices = obj.vertices;
            S.Faces = obj.faces;
            S.FaceColor = 'yellow';
            S.EdgeColor = 'none';
            S.LineWidth = 0.1;
            S.SpecularColorReflectance = 0.1;
            S.SpecularExponent = 100;
%             figure();
            h = patch(S);
            xlabel('x');
            ylabel('y');
            zlabel('z');
            cameratoolbar('SetCoordSys','none');
            cameratoolbar('setmode','orbit');
            lighting phong, camlight head,
            axis equal
            
        end
        function trianglesArea = trianglesArea(obj) % compute triangless area
            % Vectorizing for each triangle the cross product formula for
            % area (parallelogram):
            % (1/2)*||(A-B)cross(A-C)|| (norm 2) where A,B,C are the
            % 3D vectors of each vertex (corresponding to some coordinate
            % system). The subtruction (e.g A-B) creates the free vectors
            % to compute the triangle area.
            % The computation is as follows:
            % First compute the cross product of free vectors in vectorized
            % way (for each face f from 1 to F=numFaces):
            %   [ f1(vector1) - f1(vector2) ] cross [ f1(vector1) - f1(vector3) ]
            %   [ f2(vector1) - f2(vector2) ] cross [ f2(vector1) - f2(vector3) ]
            %   [ f3(vector1) - f3(vector2) ] cross [ f3(vector1) - f3(vector3) ]
            %          .               .        .           .             .
            %          .               .        .           .             .
            %          .               .        .           .             .
            %   [ fF(vector1) - fF(vector2) ] cross [ fF(vector1) - fF(vector3) ]
            % The output of this is a [F,3] matrix where each row is the
            % cross vector per face. Then taking the norm of each row by
            % computing the sqrt sum of squares: (x1^2+y1^2+z1^2)^0.5. Anf
            % finally taking 1/2 to get the triangle area.
            trianglesArea = 0.5*sqrt(sum(cross(obj.vertices(obj.faces(:,1),:)-obj.vertices(obj.faces(:,2),:),...
                obj.vertices(obj.faces(:,1),:)-obj.vertices(obj.faces(:,3),:),2).^2,2));
        end
        function verticesArea = verticesArea(obj) % compute vertices area
            % The computation is just the vertices-faces adjaceny matrix
            % multiplication with the triangles area vector divided by 3.
            % The multiplication will give each vertice the area of it's 1
            % ring faces divided by 3 as wanted.
            verticesArea = obj.adjacencyVerticesFaces*obj.trianglesArea/3;
            
            % Old version
            % The vertices area are computed for vertex as the sum of
            % adjacent (1 ring) faces area divided by 3. As if the
            % accumlation 1/3 of each adjacent face's area.
            % The computation is done as follows:
            % We can calculate easily the triangles areas, and from the
            % faces matrix we know that the corresponding area of the face
            % contributes this area to all vertices the builds that face.
            % e.g: trianglesArea = [ 2,          faces = [ 1, 2, 5 ;
            %                        3,                    1, 6, 12;
            %                        5,                    3, 5, 4 ;
            %                        .
            %                        .
            %                        .
            %                        6]                    9, 1, 2 ]
            % In this example vertex 1 is in triangles 1,2,lastTriangle so
            % all the corresponding triangles areas (2,3,6) should be taken
            % in account for the computation for vertex 1. By column
            % stacking faces (one long column vector of indices) and
            % repmatting trianglesArea we can use the accumarray function
            % to compute the all the vertices areas with a single command.
            % The old algorithm:
            % trianglesArea = obj.trianglesArea;
            % verticesArea = accumarray(obj.faces(:), repmat(trianglesArea,3,1), [obj.numVertices,1])/3;
        end
        function h = functionVis(obj, func, funcType) % display discrete function on vertices or faces
            % The function displays a discrete function over the vertices
            % or faces, depends on the input funcType = {'v','f'}
            switch funcType
                case 'v' % one value per vertex - interpolating on faces
%                     figure();
                    h = patch('Faces', obj.faces,...
                        'Vertices', obj.vertices,...
                        'FaceVertexCData', func,...
                        'FaceColor', 'interp',...
                        'EdgeColor', 'none');
                    xlabel('x');
                    ylabel('y');
                    zlabel('z');
%                     colorbar;
                    colormap('jet');
%                     lighting phong, 
%                     camlight head,
                    axis equal
                    axis off
                case 'f' % one value per face
                    figure();
                    patch('Faces', obj.faces,...
                        'Vertices', obj.vertices,...
                        'FaceVertexCData', func,...
                        'FaceColor', 'flat');
                    colorbar;
                    xlabel('x');
                    ylabel('y');
                    zlabel('z');
            end
            cameratoolbar('SetCoordSys','none');
            cameratoolbar('setmode','orbit');
            axis equal;
        end
        function h = visVertices(obj, verts_inds) % visualize a function of a list of vertices
            
            % create function to vis
            f = zeros(obj.numVertices, 1);
            f(verts_inds) = 1;
            
            % visualize
            h = obj.functionVis(f, 'v');
            
        end
        function genus = genus(obj) % compute genus (euler's formula)
            % genus  Compute genus according to euler's formula
            %   According to euler's formula V - E + F = 2(1-g)
            %   So the calculation is genus = 1 -(V - E + F)/2
            genus = 1 -(obj.numVertices - obj.numEdges + obj.numFaces)/2;
        end
        function writeOFF(obj,filename) % write the mesh to off file
            % writeOFF  This function writes the mesh to off file
            
            % TODO: enable controll of floating point precision on write
            fileID = fopen(filename, 'w');
            fprintf(fileID, 'OFF\n');
            fprintf(fileID, '%d %d %d\n', obj.numVertices, obj.numFaces, obj.numEdges);
            fprintf(fileID, '%f %f %f\n', obj.vertices.');
            % fixing to OFF format counting from 0
            fprintf(fileID, '%d %d %d %d\n', [repmat(3,obj.numFaces,1), (obj.faces -1)].');
            fclose(fileID);
        end
        function edges = edges(obj, full) % get edges
            % edges  Outputing the edges
            % If full is true duplicate edges are returned (non boundary
            % edges), else only the unique is returned.
            
            if full == true
                % Building the edges: edge(i) = s(i)<-->t(i)
                s = [obj.faces(:,1).', obj.faces(:,2).', obj.faces(:,3).'];
                t = [obj.faces(:,2).', obj.faces(:,3).', obj.faces(:,1).'];
                edges = sortrows(sort([s;t]).');
                
            else
                edges = obj.g.Edges.EndNodes;
            end
        end
        function [boundaryEdges, numBoundEdg] = boundaryEdges(obj) % compute the number of boundary edges
            % boundaryEdges  Computing the number of boundary edges
            % A boundary edge is an edge that is contained only in one
            % triangle (and not 2 in a manifold mesh, or more in a
            % non-manifold mesh)
            
            % get edges
            edges = obj.edges(true);
            % keep only unique edges (don't appear more the once)
            [uniqueEdges, ~, ic] = unique(edges,'rows');
            % ic will give indices of the edges (each row is an edge)
            % each edge indice gives 1 in the accumulation, so for boundary
            % edges we expect to have just 1 and for inner edge to get 2
            % (manifold mesh);
            numSameEdge = accumarray(ic, 1, [obj.numEdges,1]);
            % return the boundary edges
            boundaryEdges = uniqueEdges(numSameEdge == 1,:);
            numBoundEdg = size(boundaryEdges,1);
        end
        function edgesLength = edgesLength(obj) % compute edges lengths
            % edgesLength  Computing edges length
            % For each edge i compute ||e_i(V1) - e_i(V2)|| corresponding
            % to edges from edges method
            
            % get unique edges
            edges = obj.edges(false);
            % compute edges length vectorized
            edgesLength = sqrt(sum((obj.vertices(edges(:,1),:)-obj.vertices(edges(:,2),:)).^2, 2));
        end
        function [I_F2V, funcOut] = interpF2V(obj, funcIn) % interp func from faces to vertices
            % interpF2V  Interpolating discrete func on faces to vertices
            %   If funcIn is empty return just the (sparse) interpolator
            %   I_F2V and funcOut as empty []
            
            % prepare sparse vertices-faces adjacecny matrix where instead of 1
            % there the corresponding are of the triangle
            i = [obj.faces(:,1); obj.faces(:,2); obj.faces(:,3)];
            j = [1:obj.numFaces, 1:obj.numFaces, 1:obj.numFaces].';
            facesAreas = obj.trianglesArea;
            v = [facesAreas; facesAreas; facesAreas];
            adjacencyVerticesFacesAreas = sparse(i, j, v, obj.numVertices, obj.numFaces);
            % prepare sparse diagnoal matrix with the inverse vertices
            % areas on the diagonal
            verticesAreasInverseDiagonal = sparse(1:obj.numVertices, 1:obj.numVertices, 1./obj.verticesArea, obj.numVertices, obj.numVertices);
            % tne following is giving I_F2V(i,j) = A_f(j)/3A_v(i) iff face
            % j belongs to 1 ring of vertex i and 0 otherwise;
            I_F2V = verticesAreasInverseDiagonal*adjacencyVerticesFacesAreas/3;
            % interpolate the discrete function on faces to discrete
            % function on vertices
            if nargin < 2 % no funcIn
                funcOut = [];
            else
                funcOut = full(I_F2V*funcIn);
            end
        end
        function [I_V2F, funcOut] = interpV2F(obj, funcIn) % interp func from vertices to faces
            % interpV2F  Interpolating discrete func on vertices to faces
            %   If funcIn is empty return just the (sparse) interpolator
            %   I_V2F eand funcOut as empty []
            
            % First prepare I_F2V (the same as in interpF2V method)
            % prepare sparse vertices-faces adjacecny matrix where instead of 1
            % there the corresponding are of the triangle
            i = [obj.faces(:,1); obj.faces(:,2); obj.faces(:,3)];
            j = [1:obj.numFaces, 1:obj.numFaces, 1:obj.numFaces].';
            facesAreas = obj.trianglesArea;
            v = [facesAreas; facesAreas; facesAreas];
            adjacencyVerticesFacesAreas = sparse(i, j, v, obj.numVertices, obj.numFaces);
            % prepare sparse diagnoal matrix with the inverse vertices
            % areas on the diagonal
            verticesAreas = obj.verticesArea;
            verticesAreasInverseDiagonal = sparse(1:obj.numVertices, 1:obj.numVertices, 1./verticesAreas, obj.numVertices, obj.numVertices);
            % tne following is giving I_F2V(i,j) = A_f(j)/3A_v(i) iff face
            % j belongs to 1 ring of vertex i and 0 otherwise;
            I_F2V = verticesAreasInverseDiagonal*adjacencyVerticesFacesAreas/3;
            % prepare Av (sparse)
            verticesAreasDiagonal = sparse(1:obj.numVertices, 1:obj.numVertices, verticesAreas, obj.numVertices, obj.numVertices);
            % prepare Af^-1
            facesAreasInverseDiagonal = sparse(1:obj.numFaces, 1:obj.numFaces, facesAreas, obj.numFaces, obj.numFaces);
            % create I_V2F
            I_V2F = facesAreasInverseDiagonal*transpose(I_F2V)*verticesAreasDiagonal;
            % interpolate the discrete function on vertices to discrete
            % function on faces
            if nargin < 2 % no funcIn
                funcOut = [];
            else
                funcOut = full(I_V2F*funcIn);
            end
            
        end
        function vecFieldVis(obj, vecField, func, funcType) % visualize vector field on the mesh
            % vecFieldVis  Visualizing vector field on the mesh, with
            % possible func on the mesh. The vector field is set on the
            % middle of each face The expected input discrete vector field
            % is size 3Fx1 and in format:
            %   [ f1_x
            %     f2_x
            %     f3_x
            %       .
            %       .
            %       .
            %     fF_x
            %     f1_y
            %     f2_y
            %       .
            %       .
            %       .
            %     fF_y
            %     f1_z
            %       .
            %       .
            %       .
            %     fF_z ]
            %
            % There is a possibility to also visualize a funciton func is a
            % function over the vertices of faces, funcType is a char
            % {'v','f'} which corresponds to vertices or faces.
            
            % First visualize the mesh with/without a function
            if nargin == 2 % object and vector field
                obj.visualize;
            else
                obj.functionVis(func, funcType);
            end
            hold on
            
            % Visualize the vector field
            % The positions of each vector on the face is it's middle,
            % calculated by the mean of the 3 corresponding vertcies
            pos = (1/3)*(obj.vertices(obj.faces(:,1),:) +...
                obj.vertices(obj.faces(:,2),:) +...
                obj.vertices(obj.faces(:,3),:));
            quiver3(pos(:,1), pos(:,2), pos(:,3), ...
                vecField(1:obj.numFaces),...
                vecField(obj.numFaces+1:2*obj.numFaces),...
                vecField(2*obj.numFaces+1:3*obj.numFaces),'k');
            hold off;
            
        end
        function E_matrix = E_matrix(obj) % compute E matrix
            % Ematrix  Compute E matrix
            % Computing using the cross products. As we can assume that
            % in the off file the triangles are composed in a right hand
            % law (v1->v2, v2->v3, v3->v1), we can build rotation by two
            % cross products.
            % The normal N is given by:
            % N = (v2-v1) cross (v3-v2).
            % Or by N = (v3-v2) cross (v1-v3)
            % Or by N = (v1-v3) cross (v2-v1)
            % e1 is given by:
            % e1 = (v3-v2).
            % To get the rotated_e1 we just need to cross multiply with the
            % Normal.
            % rotated_e1 = N_normalized cross e1.
            % The Normal is the first in the product, maintaining the
            % rotation inwards in the triangle, the inplane rotation.
            
            % Get v1,v2,v3 from all faces
            v1 = obj.vertices(obj.faces(:,1),:);
            v2 = obj.vertices(obj.faces(:,2),:);
            v3 = obj.vertices(obj.faces(:,3),:);
            
            % Compute e1, e2, e3
            e1 = v3 - v2;
            e2 = v1 - v3;
            e3 = v2 - v1;
            
            % Compute N
            N = cross(e3,e1,2);
            N = N./sqrt(sum(N.^2,2));
            
            % Compute the in plane rotations of e1,e2,e3
            e1_rot = cross(N,e1,2);
            e2_rot = cross(N,e2,2);
            e3_rot = cross(N,e3,2);
            
            % Visualize the mesh - for debug
            %obj.visualize;
            %hold on
            % Visualize the vector field
            % The positions of each vector is on the middle of the edge
            % calculated by the mean of the 2 corresponding vertices
            %pos1 = (1/2)*(v2+v3);
            %pos2 = (1/2)*(v3+v1);
            %pos3 = (1/2)*(v1+v2);
            %quiver3([pos1(:,1), pos2(:,1), pos3(:,1)],...
            %        [pos1(:,2), pos2(:,2), pos3(:,2)],...
            %        [pos1(:,3), pos2(:,3), pos3(:,3)],...
            %       [e1_rot(:,1), e2_rot(:,1), e3_rot(:,1)],...
            %        [e1_rot(:,2), e2_rot(:,2), e3_rot(:,2)],...
            %        [e1_rot(:,3), e2_rot(:,3), e3_rot(:,3)],0.1);
            %hold off;
            
            % Create the sparse matrix E
            % prepare indices and vals for sparse matrix creation
            % rows indices
            i = repmat((1:3*obj.numFaces).', 3, 1);
            j = [repmat(obj.faces(:,1), 3, 1); ...
                repmat(obj.faces(:,2), 3, 1); ...
                repmat(obj.faces(:,3), 3, 1)];
            vals = [e1_rot(:); e2_rot(:); e3_rot(:)];
            E_matrix = sparse(i, j, vals, 3*obj.numFaces, obj.numVertices);
            
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Old version to rotate e1 by any angle - implemented to pi/2
            % easily changed in the rodriguez formula.
            % The idea: look at f1 = {v1,v2,v3}, e.g {1,17,325}. Compute
            % (v3-v2)cross(v1-v2), this is the normal of the plane spread
            % by the free vectors (v3-v2) as X axis and (v1-v2) as Y axis.
            % normalize to get k as Z axis. Now we want to rotate around Z
            % axis (around vector k) with pi/2. The rotation by pi/2 with
            % rodriguez formula (based on vector k as axis of rotation)
            % will be the in plane rotation. The rodrigues formula: R = I +
            % sin(t)K + (1-cos(t))K^2, where t is the angle and K =
            % skew_symetric_cross_matrix(k) . In our case: R = I + K + K^2.
            
            % Vectorize this idea
            % get v1,v2,v3 from all faces
            %v1 = obj.vertices(obj.faces(:,1),:);
            %v2 = obj.vertices(obj.faces(:,2),:);
            %v3 = obj.vertices(obj.faces(:,3),:);
            
            % create e1 - the edge infront of v1 - made of v2 and v3
            %e1 = v3-v2;
            % create k vector (normal to the plane and axis of rotation)
            %k = cross(e1, v1-v2, 2);
            %k = k./sqrt(sum(k.^2, 2)); % normalize
            
            % prepare elements and indices of large spare K
            %Ksize = 3*obj.numFaces; % num of rows/cols, K is (3Fx3F)
            % z plus indices and values
            %i_z_plus = 2:3:Ksize;
            %j_z_plus = 1:3:Ksize;
            %z_plus_vals = k(:,3);
            % z minus indices and values
            %i_z_minus = 1:3:Ksize;
            %j_z_minus = 2:3:Ksize;
            %z_minus_vals = -z_plus_vals;
            % y plus indices and values
            %i_y_plus = 1:3:Ksize;
            %j_y_plus = 3:3:Ksize;
            %y_plus_vals = k(:,2);
            % y minus indices and values
            %i_y_minus = 3:3:Ksize;
            %j_y_minus = 1:3:Ksize;
            %y_minus_vals = -y_plus_vals;
            % x plus indices and values
            %i_x_plus = 3:3:Ksize;
            %j_x_plus = 2:3:Ksize;
            %x_plus_vals = k(:,1);
            % x minus indices and values
            %i_x_minus = 2:3:Ksize;
            %j_x_minus = 3:3:Ksize;
            %x_minus_vals = -x_plus_vals;
            % concatenate indices and elements for K construction
            %i = [i_z_plus,i_z_minus,i_y_plus,i_y_minus,i_x_plus,i_x_minus].';
            %j = [j_z_plus,j_z_minus,j_y_plus,j_y_minus,j_x_plus,j_x_minus].';
            %vals = [z_plus_vals;z_minus_vals;y_plus_vals;y_minus_vals;x_plus_vals;x_minus_vals];
            % build sparse K matrix
            %K = sparse(i, j, vals ,Ksize ,Ksize);
            % spy(K); % to see if its the way I meant (3x3 block diagonal)
            
            % build sparse R matrix
            %R = speye(Ksize,Ksize) + K + K*K;
            % spy(R); % to see if its still what I meant (3x3 block diagonal)
            
            % aplly R to e1_col, e1_col needs to be a vector in the form:
            % e1_col --> [x1,y1,z1,x2,y2,z2,...,xF,yF,zF]^T - column vector
            %e1_col = e1.';
            %e1_col = e1_col(:);
            %e1_col_rotated = full(R*e1_col);
            
            % Visualize the mesh
            %obj.visualize;
            %hold on
            % Visualize the vector field
            % The positions of each vector on the face is it's middle,
            % calculated by the mean of the 3 corresponding vertcies
            %pos = (1/2)*(v2+v3);
            %quiver3(pos(:,1), pos(:,2), pos(:,3), ...
            %        e1_col_rotated(1:3:end),...
            %        e1_col_rotated(2:3:end),...
            %        e1_col_rotated(3:3:end));
            %hold off;
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            
            
            % Now we need to put e1 for all faces 1 to F, we need to put
            % them in the correct place in sparse matrix E, the rows
            % indices for the x vals are 1:F, for y vals F+1:2*F and for z
            % vals 2*F+1:3F. the columns indices are f1(:,1) duplicated 3
            % times. the vals are set: [x_vals(:); y_vals(:); z_vals(:)].
            % If we made the upper consturction by for of 3 loops we could
            % concatenate all indices i_e1;i_e2;i_e3 (j also) and values
            % and build the sparse E by one inesrtion.
            
        end
        function G_F_matrix = G_F_matrix(obj) % compute G_F mass matrix
            % G_F_matrix  Compute G_F - faces mass matrix
            % The faces areas lies on the diagonal of the sparse matrix and
            % replicated 3 times such that G_F_matrix is 3Fx3F
            
            % Build the sparse matrix
            i = 1:obj.numFaces*3;
            j = 1:obj.numFaces*3;
            vals = repmat(obj.trianglesArea, 3, 1);
            G_F_matrix = sparse(i, j, vals);
        end
        function G_V_matrix = G_V_matrix(obj) % compute G_V mass matrix
            % G_V_matrix  Compute G_B - vertices mass matrix
            % The vertices areas lie on the diagonal of the sparsre matrix
            % G_V_matrix is VxV
            
            % Build the sparse matrix
            i = 1:obj.numVertices;
            j = 1:obj.numVertices;
            vals = obj.verticesArea;
            G_V_matrix = sparse(i, j, vals);
        end
        function grad_matrix = grad_matrix(obj) % compute grad matrix
            % grad_matrix  Compute the grad matrix (gradient)
            % Computring the grad matrix by:
            % grad_matrix = (1/2)*(G_F)^-1*E
            % E           is 3FxV
            % G_F         is 3Fx3F
            % grad_matrix is 3FxV
            
            % Get E (sparse)
            E = obj.E_matrix;
            
            % Get G_F^-1 (sparse)
            G_F = obj.G_F_matrix;
            G_F_inv = sparse(1:obj.numFaces*3, 1:obj.numFaces*3, 1./diag(G_F));
            
            % Compute the gradient matrix
            grad_matrix = (1/2) * G_F_inv * E;
        end
        function div_matrix = div_matrix(obj) % compute div matrix
            % div_matrix  Computing the div matrix (divergence)
            % Computing the div matrix by:
            % div_matrix = -G_V^-1 * transpose(grad_matrix) * G_F
            %            = -G_V^-1 * transpose((1/2)*G_F^-1*E) * G_F
            %            = -(1/2) * G_V^-1 * transpose(E)
            % G_F is 3Fx3F
            % transpose(grad_matrix) is Vx3F
            % G_V^-1     is VxV
            % div_matrix is Vx3F
            
            % Get E (sparse)
            E = obj.E_matrix;
            
            % Get G_V^-1 (sparse)
            G_V = obj.G_V_matrix;
            G_V_inv = sparse(1:obj.numVertices,1:obj.numVertices,1./diag(G_V));
            
            % Compute div matrix
            div_matrix = -(1/2) * G_V_inv * transpose(E);
        end
        function div_matrix_inetgrated = div_matrix_integrated(obj) % compute div matrix integrated
            % div_matrix  Computing the div matrix (divergence) integrated
            % Computing the div matrix by:
            % div_matrix = - transpose(grad_matrix) * G_F
            %            = - transpose((1/2)*G_F^-1*E) * G_F
            %            = -(1/2) * transpose(E)
            % G_F is 3Fx3F
            % transpose(grad_matrix) is Vx3F
            % div_matrix is Vx3F
            
            % Get E (sparse)
            E = obj.E_matrix;
            
            % Compute div matrix
            div_matrix_inetgrated = -(1/2) * transpose(E);
        end
        function W_matrix = W_matrix(obj) % compute W matrix
            % W_matrix  Compute W_matrix
            % Computing W = (1/4)*(E.transpose)*(G_F)^-1*E
            % Where E encodes the rotated edges and G_F is a diagonal
            % matrix with the areas of the faces along the diagonal
            % replicated 3 times.
            % E   is 3FxV
            % G_F is 3Fx3F
            % W   is VxV
            
            % Get E (sparse)
            E = obj.E_matrix;
            
            % Get G_F^-1 (sparse)
            G_F = obj.G_F_matrix;
            G_F_inv = sparse(1:obj.numFaces*3, 1:obj.numFaces*3, 1./diag(G_F));
            
            % Compute W
            W_matrix = (1/4) * transpose(E) * G_F_inv * E;
        end
        function L_matrix = L_matrix(obj) % compute L matrix - Laplacian
            % L_matrix  Computing the Laplacian L
            % The computation is done by:
            % L = -div o grad
            %   = (1/4) *  G_V^-1 * transpose(E) * G_F^-1 * E
            % E    is 3FxV
            % G_F  is 3Fx3F
            % G_V  is VxV
            % grad is 3FxV
            % div  is Vx3F
            % L    is VxV
            
            L_matrix = -obj.div_matrix * obj.grad_matrix;
        end
        function W_matrix = W_matrix_cot(obj) % compute W matrix with cotan formula
            % W_matrix_cot  Compute W_matrix by cotan formula
            % W   is VxV
            
            % Get v1,v2,v3 from all faces
            v1 = obj.vertices(obj.faces(:,1),:);
            v2 = obj.vertices(obj.faces(:,2),:);
            v3 = obj.vertices(obj.faces(:,3),:);
            
            % Compute cotan for each edge (i-j vertices)
            v23_cot = dot((v3-v1),(v2-v1),2)./sqrt(sum((cross((v3-v1),(v2-v1),2)).^2,2));
            v12_cot = dot((v1-v3),(v2-v3),2)./sqrt(sum((cross((v1-v3),(v2-v3),2)).^2,2));
            v13_cot = dot((v1-v2),(v3-v2),2)./sqrt(sum((cross((v1-v2),(v3-v2),2)).^2,2));
            
            % Insert cot values for adjacen i~j vertices in sparse W
            % using sparse matlab matrix, inserting the same value
            % for (i,j) and (j,i) beacuse W is symmetric. When using matlab
            % sparse matrix creation the(i,j) can be inserted several times
            % and the value gets accumolated for example:
            % a = sparse([1,1],[1,1],[5,10],2,2) -> a(1,1) = 15
            % so the accumolation of cotan for each i~j edge from both
            % faces is done properly
            i = [obj.faces(:,2); obj.faces(:,3);...
                obj.faces(:,1); obj.faces(:,2);...
                obj.faces(:,1); obj.faces(:,3)];
            j = [obj.faces(:,3); obj.faces(:,2);...
                obj.faces(:,2); obj.faces(:,1);...
                obj.faces(:,3); obj.faces(:,1)];
            v = -(1/2)*[v23_cot; v23_cot; v12_cot; v12_cot; v13_cot; v13_cot];
            % create a midterm sparse matrix of only adjacent i~j values
            W_matrix = sparse(i, j, v, obj.numVertices,obj.numVertices);
            % prepare diagonal for W
            diag = -1*sum(W_matrix,2);
            % insert the diagonal
            W_matrix = spdiags(diag, 0, W_matrix);
            
        end
        function visHeatEquation(obj, u, numSteps, dt, pauseTime) % visualize heat equation on the mesh
            % visHeatEquation  Visualizing heat equation on the mesh
            % Visualize the heat equation on the mesh with starting
            % function on the vertices.
            % The calculation is done by discretization of the heat
            % equation u_t = Laplacian(u) and running itereative algorithm.
            % There are two possibilites for the discretrization:
            % not stable: [u(t+dt) - u(t)]/dt = Lu(t)
            % stable: [u(t+dt) - u(t)]/dt = Lu(t+dt)
            % Inputs:
            %   1. u [Vx1]: function on the vertices
            %   2. numSteps [int]: number of steps to run
            %   3. dt [double]: step size
            %   4. pauseTime [double]: pause time
            
            % compute the laplacian and hitting matrix
            L = obj.L_matrix;
            % TODO: why do i need to take the opposite sign +dt*L and not
            % -dt*L
            hit_matrix = ( speye(size(L)) + dt*L );
            
            figure();
            xlabel('x');
            ylabel('y');
            zlabel('z');
            
            % iterate
            for i = 1:numSteps
                
                % visualize the function on the vertices on the same figure
                patch('Faces', obj.faces,...
                    'Vertices', obj.vertices,...
                    'FaceVertexCData', u,...
                    'FaceColor', 'interp',...
                    'EdgeColor', 'none');
                
                colorbar;
                colormap('jet');
                title(['step: ', num2str(i)]);
                
                % update function
                % stable computation: [u(t+dt) - u(t)]/dt = Lu(t+dt)
                %                 --> [I - dt*L]u(t+dt) = u(t)
                u = hit_matrix\u;
                %                 u = +dt*L*u + u;
                
                % puase to see
                pause(pauseTime);
            end
        end
        function normalsFaces = normalsFaces(obj, flagNormalize) % compute faces normals
            % normalsFaces  Compute normal per face
            % If the flagNormalize = 1, returns normalized normals else not
            % Assumption: The matrix P of vertices is right hand ordered,
            % as if N = Pij X Pjk
            
            % Compute Pij, Pjk, Pki for each face
            Pij = obj.vertices(obj.faces(:,2),:) - obj.vertices(obj.faces(:,1),:);
            Pjk = obj.vertices(obj.faces(:,3),:) - obj.vertices(obj.faces(:,2),:);
            
            % Compute the unit normal for each face
            normalsFaces = cross(Pij, Pjk, 2);
            
            if flagNormalize == 1
                Nnorm = sqrt(sum(normalsFaces.^2,2));
                normalsFaces = normalsFaces./Nnorm;
            end
        end
        function normalsVertices = normalsVertices(obj) % compute vertices normals
            % normalsVertices  Compute vertices normals
            % The normal for each vertex is computed by:
            % N_i_hat = sigma_j[A_f_j * N_f_j]
            % N_i = normalize(N_i_hat)
            % Meaning the normal is computed as the normalized sum of
            % weighted one-ring faces normals. the weight is the area of
            % each face.
            
            % Get v1,v2,v3 from all faces
            v1 = obj.vertices(obj.faces(:,1),:);
            v2 = obj.vertices(obj.faces(:,2),:);
            v3 = obj.vertices(obj.faces(:,3),:);
            
            % Compute e1, e2, e3
            e1 = v3 - v2;
            e3 = v2 - v1;
            
            % Compute faces normals ( the normals are un-normalized, the
            % size is the face are, this is why the 1/2 is).
            N_f = 0.5*cross(e3,e1,2);
            
            % Compute vertices normals by adjaceny of vertcis-faces matrix
            % The area weight is encapsulated within the normals size.
            N_i = obj.adjacencyVerticesFaces*N_f;
            
            % Normalize
            N_i = N_i./sqrt(sum(N_i.^2,2));
            
            % Visuazlize the normals on the vertices
            %             obj.visualize;
            %             hold on;
            %             quiver3(obj.vertices(:,1), obj.vertices(:,2), obj.vertices(:,3), ...
            %                     N_i(:,1), N_i(:,2), N_i(:,3));
            %             hold off;
            
            % Return vertices normals
            normalsVertices = N_i;
        end
        function obj = translateMesh(obj, transVec) % translate the mesh
            % translateMesh  Translating the mesh according to transVec;
            % transVec [1x3 double]
            % returning the changed mesh object
            
            assert(all(size(transVec) == [1,3]), 'translation vector expected in size [1,3]');
            
            % Compute the new vertices according to translation
            obj.vertices = obj.vertices + repmat(transVec, obj.numVertices, 1);
        end
        function obj_out = rotateMesh(obj, rotationMat) % rotation the mesh
            % rotateMesh  Rotating the mesh according to rotationMat;
            % rotationMat [3x3 double]
            % returning the changed mesh object
            
            assert(det(rotationMat) == 1, 'The determinant of the rotation matrix is not 1');
            
            obj_out = copy(obj);
            
            % Compute the new vertices according to rotation
            obj_out.vertices = obj_out.vertices*rotationMat';
        end
        function visHeatEquationDeformation(obj, numSteps, dt, pauseTime ) % deform by heat equation on coordinates
            % visHeatEquationDeformation  Deforming the mesh by applying
            % the heat equation on the coordinates.
            %
            % Following: Desbrun, Mathieu, et al. "Implicit fairing of
            % irregular meshes using diffusion and curvature flow."
            % Proceedings of the 26th annual conference on Computer
            % graphics and interactive techniques. ACM Press/Addison-Wesley
            % Publishing Co., 1999.? APA
            %
            % Note that the mesh changes with each iteration and though the
            % Laplacian matrix as well
            %
            % The heat equation is performed in the unconditionally stable
            % backward Euler method solving the linear equation:
            % (1-dt*L)X(n+1) = X(n)
            
            % visualize before deformation
            figure();
            xlabel('x');
            ylabel('y');
            zlabel('z');
            patch('Faces', obj.faces,...
                'Vertices', obj.vertices,...
                'FaceColor', 'green',...
                'EdgeColor', 'black');
            cameratoolbar('SetCoordSys','none');
            cameratoolbar('setmode','orbit');
            
            % itereate
            for i=1:numSteps
                
                % get laplacian matrix
                L = obj.L_matrix;
                
                % compute the new vertices by applying the laplacian on the
                % current veritces
                obj.vertices = (speye(size(L))+dt*L)\obj.vertices;
                
                % visualize the mesh
                clf;
                patch('Faces', obj.faces,...
                    'Vertices', obj.vertices,...
                    'FaceColor', 'green',...
                    'EdgeColor', 'black');
                
                % puase to see
                pause(pauseTime);
            end
        end
        function visMeshModesReconstruction(obj, numModes) % visualize mesh as sum of first modes
            % visMeshModes  Visualize the mesh with first modes of the laplacian
            % The laplacian is not normalized by weights to keep it
            % positive semi-definite and symmetric --> W matrix
            
            % Get W marix
            W = obj.W_matrix_cot;
            
            % Compute the first numModes eigen vectors
            [V,~] = eigs(W,numModes,'sm');
            
            % Compute each coordinate function as:
            % f = sum_i=1_to_numModes (<V_i,f>V_i)
            % for all three coordinates at once using matrix
            % multiplication:
            vertices_new = V*V.'*obj.vertices;
            
            % Visuzlise the mesh
            figure();
            xlabel('x');
            ylabel('y');
            zlabel('z');
            patch('Faces', obj.faces,...
                'Vertices', vertices_new,...
                'FaceColor', 'green',...
                'EdgeColor', 'black');
            
            cameratoolbar('SetCoordSys','none');
            cameratoolbar('setmode','orbit');
            
            
        end
        function Q_matrix = Q_matrix(obj) % compute Q matrix - angle deficiency
            % Q_matrix  Compute angular deficiency about each vertex
            % Q is VxV diagonal matrix, Q(i,i) = 2*pi - sigma_neighbours(theta)
            % eps is regulaizer for flat areas where there is no gaussiab
            % curvature
            eps = 0.1;
            
            % Get v1,v2,v3 coordinates from all faces
            v1 = obj.vertices(obj.faces(:,1),:);
            v2 = obj.vertices(obj.faces(:,2),:);
            v3 = obj.vertices(obj.faces(:,3),:);
            
            % Compute theta for each vertex in the face
            theta_v1 = acos(dot((v3-v1),(v2-v1),2)./(sqrt(sum((v3-v1).^2,2)).*sqrt(sum((v2-v1).^2,2))));
            theta_v2 = acos(dot((v3-v2),(v1-v2),2)./(sqrt(sum((v3-v2).^2,2)).*sqrt(sum((v1-v2).^2,2))));
            theta_v3 = acos(dot((v1-v3),(v2-v3),2)./(sqrt(sum((v1-v3).^2,2)).*sqrt(sum((v2-v3).^2,2))));

            % Create sparse matrix size VxF where each i'th row is the angles of
            % vertex V_i in faces F_j 
            i = [obj.faces(:,1); obj.faces(:,2); obj.faces(:,3)];
            j = [1:obj.numFaces,1:obj.numFaces, 1:obj.numFaces]';
            v = [theta_v1; theta_v2; theta_v3];
            angles_V_F = sparse(i,j,v, obj.numVertices, obj.numFaces);
            assert(all( sum(angles_V_F) - pi < 1e-12 )); % summing columns should be pi, this is summing angles in each triangle
            
            % Vertices Angular Deficiency
            Q = abs(2*pi - sum(angles_V_F,2));
            % Regularize 
            Q_to_regularize_inds = Q < eps;
            Q(Q_to_regularize_inds) =  eps;
            
            % Create Q sparse matrix of angular deficiency
            Q_matrix = spdiags(Q, 0, obj.numVertices, obj.numVertices);
            
        end
        
    end
    
end

