function C = compute_fmap_hierarchy(C_12_prev, k_1_vec, k_2_vec, F_1_prol, F_2_prol, ...
    eigenvals_1_prol, eigenvals_2_prol, alpha1, alpha2, alpha3, symmetry_1, symmetry_2)
%% Compute hierarchy matrices

% get number of eigs in previous and current level
k_1_prev = k_1_vec(1);
k_2_prev = k_2_vec(1);
k_1_curr = k_1_vec(2);
k_2_curr = k_2_vec(2);

% get hierarchy eigenvalues matrices
lambda_1_previous = spdiags(eigenvals_1_prol(1:k_1_prev), 0, k_1_prev, k_1_prev);
lambda_2_previous = spdiags(eigenvals_2_prol(1:k_2_prev), 0, k_2_prev, k_2_prev);
lambda_1_curr = spdiags(eigenvals_1_prol(k_1_prev+1:end), 0, k_1_curr, k_1_curr);
lambda_2_curr = spdiags(eigenvals_2_prol(k_2_prev+1:end), 0, k_2_curr, k_2_curr);

% get hierarchy functional symmetry operator matrices - take only 3 leading diagonals
symmetry_1_previous = spdiags(spdiags(symmetry_1(1:k_1_prev,1:k_1_prev), -1:1), -1:1, k_1_prev, k_1_prev);
symmetry_2_previous = spdiags(spdiags(symmetry_2(1:k_2_prev,1:k_2_prev), -1:1), -1:1, k_2_prev, k_2_prev);
symmetry_1_curr = spdiags(spdiags(symmetry_1(k_1_prev+1:end,k_1_prev+1:end), -1:1), -1:1, k_1_curr, k_1_curr);
symmetry_2_curr = spdiags(spdiags(symmetry_2(k_2_prev+1:end,k_2_prev+1:end), -1:1), -1:1, k_2_curr, k_2_curr);

% get hierarchy descriptors matrices
F_1_previous = F_1_prol(1:k_1_prev, :);
F_2_previous = F_2_prol(1:k_2_prev, :);
F_1_curr     = F_1_prol(k_1_prev+1:end, :);
F_2_curr     = F_2_prol(k_2_prev+1:end, :);


%% Solve for C_i
disp('Solve for C_i');
disp('Prepare matrices');
h1 = alpha1*kron(F_2_curr*F_2_curr', speye(k_1_prev));
h2 = alpha2*( kron(speye(k_2_curr), lambda_1_previous.^2) -2*kron(lambda_2_curr, lambda_1_previous) + kron(lambda_2_curr.^2, speye(k_1_prev)) );
h3 = alpha3*( kron(speye(k_2_curr), symmetry_1_previous'*symmetry_1_previous) -kron(symmetry_2_curr', symmetry_1_previous') -kron(symmetry_2_curr, symmetry_1_previous) + kron(symmetry_2_curr*symmetry_2_curr', speye(k_1_prev)) );
H =  h1 + h2 + h3;
grad = -alpha1*reshape((F_1_previous - C_12_prev*F_2_previous)*F_2_curr', k_1_prev*k_2_curr, 1);
fprintf('Solve linear equation of %d variables \n', numel(grad));
Vec_C_i = H \ -grad;
fmap_solution_norm = norm(H*Vec_C_i + grad);
assert(fmap_solution_norm < 1e-10, 'Error: Functional map linear solve norm is above threshold');
C_i = reshape(Vec_C_i, k_1_prev, k_2_curr);


%% Solve for C_ii and C_iii
disp('Solve for C_ii and C_iii');

% construct the linear system A*vecC = b, minimize ||A*vecC - b||^2 -> vecC
% = (A'*A)^-1*A'*b,
% TODO: compute H and grad instead of A and b and solve for x = H \ -grad
% where: H = A'*A, -grad = A'*b, USING the known strutcures (ZEROS blocks)
% and the kronecker products rules.
% A = [ a b    -->  A' = [ a' c' 0  e' 0     --> H = A'A = [ a'a + c'c +e'e   a'b
%       c 0                b' 0  d' 0  f' ]                  b'a              b'b + d'd +f'f ]
%       0 d
%       e 0
%       0 f ]
%
% b = [ F1(:) -->  -grad = A'*b = [ a' * F1(:)
%         0                         b' * F1(:) ]
%         0
%         0
%         0  ]

disp('Prepare matrices');
a_T = kron(F_2_previous, speye(k_1_curr));
b_T = kron(F_2_curr, speye(k_1_curr));
a_T_a = alpha1 * kron(F_2_previous*F_2_previous', speye(k_1_curr));
a_T_b = alpha1 * kron(F_2_previous*F_2_curr',     speye(k_1_curr));
b_T_a = alpha1 * kron(F_2_curr*F_2_previous',     speye(k_1_curr));
b_T_b = alpha1 * kron(F_2_curr*F_2_curr',         speye(k_1_curr));
c_T_c = alpha2 *(kron(lambda_2_previous.^2, speye(k_1_curr)) -2*kron(lambda_2_previous, lambda_1_curr) + kron(speye(k_2_prev), lambda_1_curr.^2));
d_T_d = alpha2 *(kron(lambda_2_curr.^2, speye(k_1_curr)) -2*kron(lambda_2_curr, lambda_1_curr) + kron(speye(k_2_curr), lambda_1_curr.^2));
e_T_e = alpha3 *(kron(symmetry_2_previous*symmetry_2_previous', speye(k_1_curr)) -kron(symmetry_2_previous', symmetry_1_curr') -kron(symmetry_2_previous, symmetry_1_curr) + kron(speye(k_2_prev), symmetry_1_curr'*symmetry_1_curr));
f_T_f = alpha3 *(kron(symmetry_2_curr*symmetry_2_curr', speye(k_1_curr)) -kron(symmetry_2_curr', symmetry_1_curr') -kron(symmetry_2_curr, symmetry_1_curr) + kron(speye(k_2_curr), symmetry_1_curr'*symmetry_1_curr));

H = [a_T_a + c_T_c + e_T_e, a_T_b; b_T_a, b_T_b + d_T_d + f_T_f];
minus_grad = alpha1*([a_T; b_T] * F_1_curr(:));
fprintf('Solve linear equation for %d variables \n', numel(minus_grad));
Vec_C_ii_C_iii = H \ minus_grad;

fmap_solution_norm = norm(H*Vec_C_ii_C_iii - minus_grad);
assert(fmap_solution_norm < 1e-10, 'Error: Functional map linear solve norm is above threshold')

% reshape
C_ii_C_iii = reshape(Vec_C_ii_C_iii, k_1_curr, k_2_prev+k_2_curr);


%% Build C out of all components
C = [C_12_prev, C_i;...
    C_ii_C_iii];


end
