function [C, knn_matches] = icp_refine_subd(C, knn_matches, b1, b2, S_0_1, S_0_2)
% compute ICP subdivision refinement
% inputs:
%   1. C [k1 x k2] - functional map to refine
%   2. knn_matches [n1_prev x 3] - previous suubdivision level knn_matches 
%   3. b1 [n1 x k1] - basis for shape 1
%   4. b2 [n2 x k2] - basis for shape 2
%   5. S_0_1 [n1_next x n1] - subdivision matrix for current subdivision
%                             level of shape 1
%   6. S_0_2 [n2_next x n2] - subdivision matrix for current subdivision
%                             level of shape 2

%% Compute subdisivio indices
% compute shape 2 even vertices 2-ring (current) inds
shape_2_even_vert_2_ring_inds = cell(1, size(S_0_2,2));
for i = 1:size(S_0_2,2)
    shape_2_even_vert_2_ring_inds{i} = find(S_0_2(:,i));
end

% prepare matches - this is a subdivided match and we the
% knn_matches from previous level.
% for each old vertex (even) which appeared in previous level (fits
% the input knn_matches to this function) we will search in its
% current (subdivided mesh) 2-ring (all the vertices it
% influenced).
% for new vertices (odd) which appeared only in this level, we will
% search in the current 2-ring of its two main creators
% (each new vertex is on an egde).

% compute shape 1 new vertices searching inds in shape 2 (like
% knn_matches is for old verts). Note that the new vertices are
% chornologically ordered S_0 rows, after the old vertices.
% For Loop secheme with triagnles, these will the new points on each
% edge, for Catmull-Clarck these will wil the new points on each edge
% and a new point in each face.
% We take the influencers on that new point:
% New point on edge: the two edge points.
% New point in face: the 4 points of the face (Catmul-Clarck quads).
% All the influencers have weight > 1/8 in the subdivision matrix S_0
shape_1_new_verts_influencers = cell(size(S_0_1,1)-size(S_0_1,2), 1);
S_0_1_T = S_0_1'; % faster to search column in matlab sparse matrices
for i = size(S_0_1,2)+1:size(S_0_1,1)
    shape_1_new_verts_influencers{i-size(S_0_1,2)} = find(S_0_1_T(:, i) > 1/8)';
end

% take unique only
matches_shape_1_new_unique = cellfun(@(x) unique(reshape(knn_matches(x,:),1, [])), shape_1_new_verts_influencers, 'UniformOutput', false);

% unite knn_matches of old even vertices with new odd vertices
knn_matches_all_verts = vertcat(num2cell(knn_matches, 2), matches_shape_1_new_unique);

% compute the indices to search for each vertex in shape1
shape_2_even_vert_2_ring_inds_to_search = cellfun(@(x) unique(cell2mat(shape_2_even_vert_2_ring_inds(x)')), knn_matches_all_verts, 'UniformOutput', false);


%% Iterate to refine
disp('Perform ICP postprocessing to the functional map');
fprintf('Mean number of points to search %.1f \n', mean(cellfun(@(x) numel(x), shape_2_even_vert_2_ring_inds_to_search)));
for i = 1:3
    
    % get rotated points
    rotated_points_2 = C*b2';
    
    % compute knn matches
    knn_matches_subd = zeros(1, size(S_0_1,1));
    parfor j = 1:size(S_0_1,1)
        inds_to_search = shape_2_even_vert_2_ring_inds_to_search{j};
        rotated_points_2_to_search = rotated_points_2(:, inds_to_search);
        ind_closet = knnsearch(rotated_points_2_to_search', b1(j, :), 'K', 1);
        knn_matches_subd(j) = inds_to_search(ind_closet);
    end
    
    % compute SVD to return orthonormal C
    W = (b2(knn_matches_subd,:) \ b1)';
    [U, ~, V] = svd(W);
    C = U*eye(size(C))*V';
    
end

% get rotated points
rotated_points_2 = C*b2';

% compute knn matches
K = 3;
knn_matches = zeros(size(S_0_1,1), K);
parfor j = 1:size(S_0_1,1)
    inds_to_search = shape_2_even_vert_2_ring_inds_to_search{j};
    rotated_points_2_to_search = rotated_points_2(:, inds_to_search);
    ind_closet = knnsearch(rotated_points_2_to_search', b1(j, :), 'K', K);
    knn_matches(j, :) = inds_to_search(ind_closet);
end

% figure; imagesc(C); colormap(bluewhitered); colorbar;


end

