classdef myMeshDEC < myMesh
    % myMeshDEC  Summary of myMeshDEC
    % class the inherets from myMesh and extending DEC utills
    
    % TODO: add a good documentation of the properties and the methods
    % Chosen orientations:
    %   1. Edges: from small index to large index, e.g: e_i = {v5, v17}
    %             then the orientation is v5 --> v17
    %   2. Faces: right handed as the input mesh faces list, e.g: f_i =
    %             {v7, v1, v3} then the orientation is right handed v7 ->
    %             v1 -> v3 -> v7
    
    properties
        P           % [Vx3] vertices positions
        V           % [Vx1] vertices list
        E           % [Ex2] edges list - sorted such that: E(k,1) < E(k,2)
        F           % [Fx3] faces list
    end
    
    methods
        function obj = myMeshDEC(varargin) % constructor
            % myMeshDEC  Constructor
            % Usage 1 - read from file:
            %           M = myMeshDEC('.../mesh.off');
            % Usage 2 - construct inside matlab:
            %           M = myMeshDEC(vertices,faces);
            
            % Call super class
            obj@myMesh(varargin{:});
            
            % Initiate properties
            obj.P = obj.vertices;
            obj.V = (1:obj.numVertices)';
            obj.E = obj.edges(false);
            obj.F = obj.faces;
            
            
        end
        function h = visEdgesOrientation(obj) % visualize edges orientation
            % visEdgesOrientation  visualizing the edges orientation
            % The edges orientation are chosen from the small vertex value
            % to the largest. For example, if edge_k = (v_5,v_18) then the
            % edge orientation is chosen as v_5 --> v_18
            % In this implementation the edges list E is sorted such that
            % E(k,1) < E(k,2)
            
            % Create the 3D orientation vectors
            orientationVecs = obj.P(obj.E(:,2),:)-obj.P(obj.E(:,1),:);
            
            % Compute the locations to visualize this orientations - the
            % smaller vertex index, such that the arrow head will be at the
            % center of the edge
            orientationPos = obj.P(obj.E(:,1),:);
            
            % Visualize the mesh
            h = obj.visualize;
            hold on;
            
            % Add the orientation vectors to the plot
            s = 0.5; % scale for the arrows to be plotted
            quiver3(orientationPos(:,1), orientationPos(:,2), orientationPos(:,3),...
                orientationVecs(:,1)*s, orientationVecs(:,2)*s, orientationVecs(:,3)*s, 'AutoScale', 'off',...
                'LineWidth',2);
            hold off;
        end
        function h = visDebug(obj) % visualize for debug
            % visDebug  This function visualized the mesh with orientation
            % of egdes, and the number of vertices, edges and faces on top
            % of it.
            
            h = obj.visEdgesOrientation;
            hold on;
            vertex_normals = obj.normalsVertices;
            for i=1:obj.numVertices
                x = obj.vertices(i,1) + vertex_normals(i,1)*0.05;
                y = obj.vertices(i,2) + vertex_normals(i,2)*0.05;
                z = obj.vertices(i,3) + vertex_normals(i,3)*0.05;
                text(x, y, z, num2str(i), 'FontSize', 20, 'Color', 'red');
            end
            for i=1:obj.numEdges
                e = obj.E(i,:);
                x = (obj.vertices(e(1),1) + obj.vertices(e(2),1))/2;
                y = (obj.vertices(e(1),2) + obj.vertices(e(2),2))/2;
                z = (obj.vertices(e(1),3) + obj.vertices(e(2),3))/2;
                text(x, y, z, num2str(i), 'FontSize', 20, 'Color', 'cyan');
            end
            faces_normals = obj.normalsFaces(1);
            for i=1:obj.numFaces
                f = obj.F(i,:);
                x = (obj.vertices(f(1),1) + obj.vertices(f(2),1) + obj.vertices(f(3),1))/3 + faces_normals(i,1)*0.1 ;
                y = (obj.vertices(f(1),2) + obj.vertices(f(2),2) + obj.vertices(f(3),2))/3 + faces_normals(i,2)*0.1 ;
                z = (obj.vertices(f(1),3) + obj.vertices(f(2),3) + obj.vertices(f(3),3))/3 + faces_normals(i,3)*0.1 ;
                text(x, y, z, num2str(i), 'FontSize', 20, 'Color', 'black');
            end
            hold off;
            
            
            
        end
        function d0_matrix = d0_matrix(obj) % create d0 matrix
            % d0_matrix  Create d0 operator matrix
            % The computation is done accroding the orienation that was
            % decided: if edge_k = (v_5,v_18) then the edge orientation is
            % chosen as v_5 --> v_18
            % d0 size is ExV
            % works on 0-forms on the veritces to 1-forms on the edges
            
            % Create the sparse matrix d0 place -1 for the smaller vertex
            % index on the edge and +1 for the larger vertex index.
            edgesInds = (1:obj.numEdges)'; % edges indices
            verticesMinusInds = obj.E(:,1);
            vertiecsPlusInds = obj.E(:,2);
            i = [edgesInds; edgesInds];
            j = [verticesMinusInds; vertiecsPlusInds];
            v = [-1*ones(obj.numEdges,1); ones(obj.numEdges,1)];
            d0_matrix = sparse(i, j, v, obj.numEdges, obj.numVertices);
        end
        function d1_matrix = d1_matrix(obj) % create d1 matrix
            % d1_matrix  Create d1 operator matrix
            % The computation is done accroding the orienation, e.g: if
            % f_i ={v5, v3, v7} then accoring to right hand rule the edges:
            % e12 = v5->v6, e23 = v3->v7, e31 = v7->v5.
            % For an edge v_i->v_j the matrix will icnlude 1 if j>i since
            % the edges orientations were chosen to be positive if j>i. -1
            % else.
            % d1 size is FxE
            % works on 1-forms on the edges to 2-forms on the faces
            
            % Get edges in each face
            facesEdges = [obj.F(:,[1,2]); obj.F(:,[2,3]); obj.F(:,[3,1])];
            
            % Compute the signs vector
            v = double(facesEdges(:,1)>facesEdges(:,2))*-1 + ...
                double(facesEdges(:,1)<facesEdges(:,2));
            
            % Get the indices of facesEdges in E list
            [~, ~, j] = unique(sort(facesEdges, 2), 'rows');
            
            % Create the sparse matrix
            i = repmat((1:obj.numFaces)',3,1);
            d1_matrix = sparse(i, j, v, obj.numFaces, obj.numEdges);
        end
        function star0_matrix = star0_matrix(obj) % create star0 matrix
            % star0_matrix  Create diagonal hodge star 0 matrix
            % The diagonal hodge star matrix takes 0-form to 2-form by
            % multiplying the with the volumes ratio: |Vi*|/|Vi|
            % |Vi*| = (1-ring area)/3
            % |Vi|  = 1 (the convention for 0-forms)
            % star0_matrix is size VxV
            
            % Compute the volume ratios:
            volume_ratios = obj.verticesArea;
            
            % Create star0 sparse matrix
            star0_matrix = sparse(obj.V, obj.V, volume_ratios);
        end
        function star0_matrix = star0_matrix_FEM(obj) % create star0 matrix
            % star0_matrix  Create first order FEM hodge star 0 matrix
            % The hodge star matrix takes 0-form to 2-form by
            % computing using first order FEM.
            % star0_matrix is size VxV
            
            % compute triangles area
            faceAreas = obj.trianglesArea;
            
            % initalize: r - rows, c - columns, v - values
            r = [];
            c = [];
            v = [];
            
            % compute all inner products
            for i = 1:3
                j = mod(i,3)+1;
                
                % Inner product of any two hat functions is area/12 on a given face
                t1 = obj.F(:,i);
                t2 = obj.F(:,j);
                r = [r; t1; t2];
                c = [c; t2; t1];
                v = [v; faceAreas/12; faceAreas/12];
                
                % Self product is area/6
                r = [r; t1];
                c = [c; t1];
                v = [v; faceAreas/6];
            end
            
            % Create star0 sparse matrix
            star0_matrix = sparse(double(r),double(c),v);
            
        end
        function star1_matrix = star1_matrix(obj) % create star1 matrix
            % star1_matrix  Create diagonal hodge star 1 matrix
            % The diagonal hodge star matrix takes 1-form to 1-form by
            % multiplying the with the volumes ratio: |e_ij*|/|e_ij|
            % |e_ij*|/|e_ij| = (1/2)*(cot(a_j) + cot(b_j))
            % --> see Keenan Crane "DIGITAL GEOMETRY PROCESSING WITH
            % DISCRETE EXTERIOR CALCULUS" discrete laplacian using DEC
            % (Chapter 6 in the PDF)
            % star1_matrix is size ExE
            
            % Get 3D vertices v1,v2,v3 from all faces
            v1 = obj.P(obj.F(:,1),:);
            v2 = obj.P(obj.F(:,2),:);
            v3 = obj.P(obj.F(:,3),:);
            
            % Compute cotan for each edge (i-j vertices)
            e23_cot = dot((v3-v1),(v2-v1),2)./sqrt(sum((cross((v3-v1),(v2-v1),2)).^2,2));
            e12_cot = dot((v1-v3),(v2-v3),2)./sqrt(sum((cross((v1-v3),(v2-v3),2)).^2,2));
            e13_cot = dot((v1-v2),(v3-v2),2)./sqrt(sum((cross((v1-v2),(v3-v2),2)).^2,2));
            
            % Get the corresponding edges, there may be repetitions for the
            % same edge since it may be on two faces. Sort the edges to
            % ease to next move
            e23 = sort(obj.F(:,[2,3]), 2);
            e12 = sort(obj.F(:,[1,2]), 2);
            e13 = sort(obj.F(:,[1,3]), 2);
            
            % Get the indices of [e23; e13; e12] in the edges list E
            [~, ~, ic] = unique([e23; e12; e13], 'rows');
            
            % Concatenate the cotan values (in the corresponding order)
            v = 0.5*[e23_cot; e12_cot; e13_cot];
            
            % Create the star1 sparse matrix
            star1_matrix = sparse(ic, ic, v, obj.numEdges, obj.numEdges);
            
        end
        function star1_whitney_matrix = star1_whitney_matrix(obj) % create star1 whitney matrix
            % star1_whitney_matrix  Create the non diagonal positive hodge
            % star 1 matrix, based on whitney 1-forms basis inner products
            % See:
            % Valentin Zobel. Spectral Analysis of the Hodge Laplacian on
            % Discrete Manifolds. Master Thesis, 2010.
            
            % get d1 matrix
            d1 = obj.d1_matrix;
            
            % get edges length
            edges_length_square = obj.edgesLength.^2;
            
            % get triangles are
            triangles_area = obj.trianglesArea;
            
            % get t2e - edges in triangles size Fx3
            [r, ~] = find(d1');
            t2e = reshape(r, 3, [])';
            % get e2t
            
            
            
            % initialize containers for indicds and value for sparse matrix
            % creating
            i_ind2 = zeros(obj.numFaces*3*3, 1);
            j_ind2 = zeros(obj.numFaces*3*3, 1);
            val2 =   zeros(obj.numFaces*3*3, 1);
            running_ind = 1;
            i_ind = [];
            j_ind = [];
            val   = [];
            
            % run over edges
            for e = 1:obj.numEdges
                
                % get the edge length square
                e_length = edges_length_square(e);
                
                % get the adjacent triangles to the edge
                [triangles_adj_to_e, ~] = find(d1(:, e) ~= 0);
                
                % run over adjacent triangles and compute the values
                for j = 1:numel(triangles_adj_to_e)
                    
                    % get triangle
                    t = triangles_adj_to_e(j);
                    
                    % get triangle area
                    t_area = triangles_area(t);
                    
                    % get other edges
                    edges_in_triangle = t2e(t, :);
                    other_edges = edges_in_triangle(~ismembc(edges_in_triangle, e));
                    e2 = other_edges(1);
                    e3 = other_edges(2);
                    e2_length = edges_length_square(e2);
                    e3_length = edges_length_square(e3);
                    
                    % compute the signs - with e orientation or opposite
                    % computing it by their sign in the triangle using d1
                    % matrix, if they agree on the sign in the triangle,
                    % i.e. both are 1 or both are -1 then they are in the
                    % same direction, otherwise they are opposite.
                    % multiplying the values will lead to +1 same, -1 diff
                    sign_e1e2 = d1(t,e) * d1(t,e2);
                    sign_e1e3 = d1(t,e) * d1(t,e3);
                    
                    % set indices and values
                    
                    new_val= 1/(24*t_area)* [1; sign_e1e2; sign_e1e3] .* ([-1/2, 3/2, 3/2; -1/2, -1/2, 3/2; -1/2, 3/2, -1/2] * [e_length; e2_length; e3_length]);
                    %                     new_val = [      1/(24*t_area)   * ( (-1/2)*e_length + (3/2)*e2_length + (3/2)*e3_length ); ...
                    %                                sign_e1e2/(24*t_area) * ( (3/2)*e3_length - (1/2)*e_length  - (1/2)*e2_length ); ...
                    %                                sign_e1e3/(24*t_area) * ( (3/2)*e2_length - (1/2)*e_length  - (1/2)*e3_length )];
                    i_ind2(running_ind:running_ind+2) = e;
                    j_ind2(running_ind:running_ind+2) = [e; e2; e3];
                    val2(running_ind:running_ind+2) = new_val;
                    
                    %                     i_ind = [i_ind; e; e ; e];
                    %                     j_ind = [j_ind; e; e2; e3];
                    %                     val = [val; new_val];
                    
                    running_ind = running_ind+3;
                    
                end
            end
            
            % create the sparse matrix
            %             star1_whitney_matrix = sparse(i_ind, j_ind, val, obj.numEdges, obj.numEdges);
            star1_whitney_matrix = sparse(i_ind2, j_ind2, val2, obj.numEdges, obj.numEdges);
            
        end
        function star2_matrix = star2_matrix(obj) % create star2 matrix
            % star2_matrix  Create diagonal hodge star 2 matrix
            % The diagonal hodge star matrix takes 2-form to 0-form by
            % multiplying the with the volumes ratio: |F_ijk*|/|F_ijk|
            % |F_ijk*| = 1 (since it a 0-form on the dual mesh)
            % |F_ijk|  = the face (triangle) area
            % star1_matrix is size FxF
            
            % Compute the volume ratios:
            volume_ratios = 1./obj.trianglesArea;
            
            % Create star0 sparse matrix
            star2_matrix = sparse(1:obj.numFaces, 1:obj.numFaces, volume_ratios);
        end
        function VecField = reconsturctVectorField(obj, C, flag_normalize_vec_field, flagDisplay) % reconstruct vector field
            % reconsturctVectorField  Reconstruct a vector field on the
            % faces. Following "Design of Tangent Vector Fields".
            % Set a vector in the middle of every face using the whitney
            % 1-form interpolation. For example on triangle with vertices
            % Pi, Pj, Pk, edges 1-form discrete values Cij, Cjk, Cki, area
            % Aijk and barcentric coordiantes ai, aj, ak:
            % u_ijk(ai, aj, ak) = [ (Cki*ak - Cij*ai)*JPjk +
            %                       (Cij*ai - Cjk*ak)*JPki +
            %                       (Cjk*aj - Cki*ai)*JPij] / 2*Aijk
            % Where C is the vectof of discrete 1-forms, value per edge
            % J is the in plan,e rotation, Pij is the vector from Pi to Pj.
            % J*Pij = NxPij, where N is the triangle normal.
            % Setting a vector per face means that the barycentric
            % coordiantes are (ai, aj, ak) = (1/3, 1/3, 1/3) leading to a
            % simpler formula:
            % % u_ijk(1/3, 1/3, 1/3) = [ (Cki - Cij)*JPjk +
            %                            (Cij - Cjk)*JPki +
            %                            (Cjk - Cki)*JPij) ] / 6*Aijk
            %
            % C is of size Ex1 containing discrete 1-form value per edge
            %
            % Assumption: The matrix P of vertices is right hand ordered,
            % as if N = Pij X Pjk
            
            % TODO: COMPLETE THE IMPLEMENTATION FOR SURFACE WITH BOUNDARY
            assert(isempty(obj.boundaryEdges), 'The implementation does not deal yet with surface with a boundary');
            
            % Compute Pij, Pjk, Pki for each face
            Pij = obj.P(obj.F(:,2),:) - obj.P(obj.F(:,1),:);
            Pjk = obj.P(obj.F(:,3),:) - obj.P(obj.F(:,2),:);
            Pki = obj.P(obj.F(:,1),:) - obj.P(obj.F(:,3),:);
            
            % Compute the unit normal for each face
            N = cross(Pij, Pjk, 2);
            Nnorm = sqrt(sum(N.^2,2));
            N = N./Nnorm;
            
            % Compute JPij, JPjk, JPki for each face
            JPij = cross(N, Pij, 2);
            JPjk = cross(N, Pjk, 2);
            JPki = cross(N, Pki, 2);
            
            % Visualize the mesh - for debug
            %obj.visualize;
            %hold on
            % Visualize the vector field
            % The positions of each vector is on the middle of the edge
            % calculated by the mean of the 2 corresponding vertices
            %pos_ij = (1/2)*(obj.P(obj.F(:,1),:) + obj.P(obj.F(:,2),:));
            %pos_jk = (1/2)*(obj.P(obj.F(:,2),:) + obj.P(obj.F(:,3),:));
            %pos_ki = (1/2)*(obj.P(obj.F(:,3),:) + obj.P(obj.F(:,1),:));
            %quiver3([pos_ij(:,1), pos_jk(:,1), pos_ki(:,1)],...
            %        [pos_ij(:,2), pos_jk(:,2), pos_ki(:,2)],...
            %        [pos_ij(:,3), pos_jk(:,3), pos_ki(:,3)],...
            %        [JPij(:,1), JPjk(:,1), JPki(:,1)],...
            %        [JPij(:,2), JPjk(:,2), JPki(:,2)],...
            %        [JPij(:,3), JPjk(:,3), JPki(:,3)],0.01);
            %hold off;
            
            % Get edges in each face
            facesEdges = [obj.F(:,[1,2]); obj.F(:,[2,3]); obj.F(:,[3,1])];
            
            % Get the indices of facesEdges in E list and the signs
            signs = double(facesEdges(:,1)>facesEdges(:,2))*-1 + ...
                double(facesEdges(:,1)<facesEdges(:,2));
            % This calculation for sign holds since the edges orientation
            % was chosen as described in the property
            [~, ~, j] = unique(sort(facesEdges, 2), 'rows');
            
            % C values per edge in the faces with correct sign
            % (orientation)
            CVals = C(j).*signs;
            
            % Get Cij, Cjk, Cki for each face
            Cij = CVals(1:obj.numFaces);
            Cjk = CVals(obj.numFaces+1:2*obj.numFaces);
            Cki = CVals(2*obj.numFaces+1:end);
            
            % Interpolate the vector at the middle of each face
            % u_ijk(1/3, 1/3, 1/3) = [ (Cki - Cij)*JPjk +
            %                          (Cij - Cjk)*JPki +
            %                          (Cjk - Cki)*JPij) ] / 6*Aijk
            VecField = ((Cki - Cij).*JPjk + ...
                (Cij - Cjk).*JPki + ...
                (Cjk - Cki).*JPij)./(6*Nnorm);
            
            % Normalize the vector field
            if flag_normalize_vec_field == 1
                VecField = VecField./sqrt(sum(VecField.^2,2));
            end
            
            % Plot the vector reconstructed vector field
            if flagDisplay == true
                obj.vecFieldVis(VecField(:));
            end
        end
        function [subdividedMesh, S_0, S_1, S_2] = subdivideMeshLoop_old(obj) % Loop subdiviosn
            % subdivideLoop  Perform Loop subdivision for the triangular
            % mesh.
            % Following notations from: "Subdivision Exterior Calculus for
            % Geometry Processing Supplemental Material: Subdivision Rules"
            % Warning: coded not efficiently.
            %
            % Output:
            %   1. subdivdedMesh - the subdivided mesh
            %   2. S_0 - Subdivision matrix for 0-forms
            %   3. S_1 - Subdivision matrix for 1-forms
            %   4. S_2 - Subdivision matrix for 2-forms
            
            % TODO: COMPLETE THE IMPLEMENTATION FOR SURFACE WITH BOUNDARIES
            assert(isempty(obj.boundaryEdges), 'The implementation does not deal yet with surface with a boundary');
            
            % Assert the input is a manifold
            assert(size(obj.edges(1),1)/2 == size(obj.edges(0),1), 'The input is non-manifold');
            
            % Subdvide the vertices (0-form Loop scheme):
            disp('S0 calculation');
            
            % Even vertices (the old vertices new location)
            n = full(sum(obj.adjacencyMatrix,2)); % number of neighbours for each vertex
            alpha = (n>3).*(3./(8*n)) + (n==3)*(3/16); % alpha
            % fast version
            S_0_even = sparse(1:obj.numVertices, 1:obj.numVertices, alpha, obj.numVertices, obj.numVertices)*obj.adjacencyMatrix;
            S_0_even = S_0_even + sparse(1:obj.numVertices, 1:obj.numVertices, (1-alpha.*n), obj.numVertices, obj.numVertices);
            % slow version
            % S_0_even = obj.adjacencyMatrix.*alpha;
            % S_0_even = S_0_even + speye(obj.numVertices).*(1-alpha.*n);
            
            % Odd vertices (new vertices indtroduced by the edges)
            % say the edge is of vertices j,k and on the 2 triangles F1, F2
            % that hold the edge there are two more vertices i,l
            % vertices on the edge: j,k
            d0 = obj.d0_matrix;
            d1 = obj.d1_matrix;
            S_0_odd_edge_verts = abs(d0)*(3/8);
            % 2 triangles of the edge: F1, F2
            [row, ~] = find(abs(d1) == 1);
            % all vertices of the 2 triangles of the edge: i,j(x2),k(x2),l
            all_verts_around_edge = reshape(obj.F(row,:)',6,[])';
            % getting vertices i,j, using: https://stackoverflow.com/questions/32494331/setdiff-row-by-row-without-using-loops-in-matlab
            mask = all(bsxfun(@ne,all_verts_around_edge,permute(obj.E,[1 3 2])),3);
            At = all_verts_around_edge';
            oppsoite_vertices = reshape(At(mask.'),[],size(all_verts_around_edge,1));
            % insert the oppsosite vertices with the right weight
            S_0_odd_opposite_verts = sparse(reshape([1:obj.numEdges;1:obj.numEdges],[],1),oppsoite_vertices(:),1/8,obj.numEdges,obj.numVertices);
            % build S_odd
            S_0_odd = S_0_odd_edge_verts + S_0_odd_opposite_verts;
            
            % S is the combination of S_even (old vertices new wieghts) and
            % S_odd (the new vertices weights)
            S_0 = [S_0_even; S_0_odd];
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % New Triangulation of subdivded mesh
            disp('New triangulation');
            
            % create the new vertices positions
            P_new = S_0*obj.P;
            
            % edges in each triangle
            [triangle_edges, ~] = find(abs(d1') == 1);
            triangle_edges = reshape(triangle_edges,3,[])';
            
            % get triangles normals (for ordering later the new triangles)
            faces_normals  = obj.normalsFaces(0);
            
            % break triangles and create new ones
            % each triangle gives 4 new ones - not dealing yet with
            % surfaces with boundaries
            F_new = zeros(obj.numFaces*4,3);
            new_triangle_ind = 1;
            for i=1:obj.numFaces
                
                % get the edges of the triangle
                t_edges = triangle_edges(i,:);
                
                % get the list of vertices in the face's three edges
                vertices_in_t_edges = obj.E(t_edges(:),:);
                
                % build the 3 new triangles that include the old vertices
                % that just changed (even) location and two new edge ones
                % (odd). There are numEdges new vertices in the subdivided
                % mesh and they are sorted according to obj.E list.
                % obj.E(1)->the first new vertex etc. Meaning that in the new
                % vertex list the first numVertices vertices are the old
                % ones (even) that changed location (will show itself in P
                % list of xyz). The new vertices will count from
                % numVertices+1 to numVertices + numEdges, since each edge
                % is giving a new vertex.
                for j=1:3
                    % get the vertex
                    v = obj.F(i,j);
                    % update as the first vertex in new F
                    F_new(new_triangle_ind,1) = v;
                    % find which of the edges in the triangle contain v
                    [v_in_t_edges, ~] = find(vertices_in_t_edges == v);
                    % get back the edges indices + the number of vertices
                    new_inds_to_add = t_edges(v_in_t_edges) + obj.numVertices;
                    % adding to the triangle
                    F_new(new_triangle_ind,2:3) = new_inds_to_add ;
                    % update new triangle index
                    new_triangle_ind = new_triangle_ind + 1;
                end
                
                % build the 4th triangle out of the three new vertices from the edges (odd)
                F_new(new_triangle_ind,:) = t_edges + obj.numVertices;
                
                % update new triangle index
                new_triangle_ind = new_triangle_ind+1;
                
            end
            
            % Rearange vertex in faces so that new normals are in same
            % orientation like old normals. Each 4 new triangles need to
            % have same outwards orientation like their creating triangle.
            % Meaning they need to hold the right hand rule like their
            % creator.
            % The same orientation is measured such that dot(N_old, N_i) >
            % 0 for i={1,2,3,4};
            Pij = P_new(F_new(:,2),:) - P_new(F_new(:,1),:);
            Pjk = P_new(F_new(:,3),:) - P_new(F_new(:,2),:);
            normal_new = cross(Pij, Pjk, 2);
            % check if the orientations agree and flip else
            flip_normals_inds = dot(repelem(faces_normals,4,1),normal_new,2) < 0;
            if any(flip_normals_inds)
                F_new(flip_normals_inds,2:3) = F_new(flip_normals_inds, [3,2]);
            end
            
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Subdvide 2-forms vertices (generalized half-box splines):
            % not dealing yet with boundaries
            disp('S2 calculation');
            
            % faces-faces adjacency
            triangles_triangles_adjacency_matrix = abs(d1*d1')-diag(diag(abs(d1*d1')));
            [t_adj_t, ~] = find(triangles_triangles_adjacency_matrix == 1);
            triangles_triangles_adjacency_list = reshape(t_adj_t, 3, [])';
            
            % vertices-faces adjacency
            %             vertices_triangles_adjacency_list = cellfun(@(x) find(x),num2cell(obj.adjacencyVerticesFaces,2),'un',0);
            % for faster computations
            
            
            % Even 2-forms interior rules (inner new triangles)
            % basicaly what we need is to take each row in the faces-faces
            % adjacency matrix in put it with 1/4 wieght in a larger sparse
            % matrix such that rows goes like this: 1->4, 2->8, 3->12,...
            % We can do this by the following construction with find:
            % each triple (starting from 1) in the following is the 3
            % faces adjacent to the corresponding face (1:3 -> 1, 4:6 ->
            % 2,...), this row vectors containing the indices of the
            % adjacent faces to face i should be located with proper weight
            % in the 4,8,12,... location since when buuilding the
            % subdivided mesh they sit in every 4th location. Since this
            % new faces are composed from the new edges that was the middle
            % of every old edge.
            % again assuming no boundaries then every triangle has 3
            % nieghboring triangles. TODO: allow boundaries
            
            % Odd 2-form interior rules (the 3 remaining new triangles)
            % run over all triangles and build the i,j,v vectors for
            % creating the sparse matrix S_2_odd
            % again assuming no boundaries then every triangle has 3
            % nieghboring triangles. TODO: allow boundaries
            % The flow is as follows: for each triangle get it's
            % adjacent triangles t_adj. run over the triangle 3 vertices
            % (i,j,k), and for each get the neighboring triangles t_adj_v.
            % the intersection between t_adj and t_adj_v will give the two
            % relevant triangles with same weight for the i'th vertex
            % calculation. remembering that the i'th vertex corresponds to
            % the specific new triangle composed of i and two other new
            % vertices from the divided edges i-j, i-k.
            
            % Overall there are 4 new triangles per triangle. 1 to give the
            % even rules and the other 3 with odd rules, so the
            % accmulation of S_2_odd & S_2_even is duable with a single for
            % loop over and triangles and keeping track of indices you can
            % build the vectors to create sparse S_2 comletley
            
            % The counting for number indices to prepare for the sparse
            % matrix creation is: each triangle is splitted to 4: the
            % middle one takes weights (even rules) from 4 triangles, the
            % other 3 take weights from its sorrounding triangles. The
            % calculation gives: 4*|F| + sigma_v
            % num_adjacent_triangles(v)^2. The num_adjacent_triangles
            % should be cut off for not more the 5 adjacent triangles per
            % vertex as can be seen in the supplemntary paper.
            
            % calculate the number of faces that adjacent each vertex
            num_adjacent_faces_per_vertex = sum(obj.adjacencyVerticesFaces,2);
            % cutting off to no more the 5 as can been seen in the
            % supplemntary paper
            num_adjacent_faces_per_vertex_cut_off = (num_adjacent_faces_per_vertex>5)*5 + (num_adjacent_faces_per_vertex<=5).*num_adjacent_faces_per_vertex;
            % effective adjacency for indices initiation
            num_adj_f_v_eff = num_adjacent_faces_per_vertex.*num_adjacent_faces_per_vertex_cut_off;
            % initiating the vectors for sparse S_2 building
            i_inds = zeros(sum(num_adj_f_v_eff) + obj.numFaces*4,1);
            j_inds = zeros(sum(num_adj_f_v_eff) + obj.numFaces*4,1);
            vals = zeros(sum(num_adj_f_v_eff) + obj.numFaces*4,1);
            % initiating tracking inds
            running_ind = 1;
            new_triangle_ind = 1;
            % calculate beta
            beta = 1/4 - 1/(16*sin(2*pi/5)^2);
            
            % run over triangles
            for i=1:obj.numFaces
                
                % adjacent triangles to the triangle
                t_adj = triangles_triangles_adjacency_list(i,:);
                
                % odd rules: each triangle is divided to 4, the first 3 is
                % from the (i,j,k) vertices + 2 new from the divided edges.
                % assuming no boundaries, TODO: allow boundaries
                for j=1:3
                    % get the vertex
                    v = obj.F(i,j);
                    % get the adjacent triangles to this vertex
                    t_adj_v = find(obj.adjacencyVerticesFaces(v,:) == 1);
                    %                     t_adj_v = vertices_triangles_adjacency_list{v}; %
                    %                     faster
                    % number of adjacent tirangles to the vertex after
                    % cut-off
                    n = num_adjacent_faces_per_vertex(v);
                    n_cutt_off = num_adjacent_faces_per_vertex_cut_off(v);
                    % get the two closet relevant adjacent triangles
                    t_closet = intersect(t_adj_v, t_adj);
                    % get the other relevamt adjacent triangles (2 at most)
                    t_others = setdiff(intersect(t_adj_v, reshape(triangles_triangles_adjacency_list(t_closet,:),6,1)), [i, t_closet]);
                    % assign indices of weights for j'th new triangle
                    i_inds(running_ind:running_ind + n_cutt_off - 1) = new_triangle_ind;
                    % assign indice and weight from the main triangle
                    j_inds(running_ind) = i;
                    vals(running_ind) = ((n==3)*(2/3) +...
                        (n==4)*(5/8) +...
                        (n==5)*(3/4 - beta)+...
                        (n>=6)*(1/2))*(1/4);
                    
                    % assign indices and weights for the 2 closet triangles
                    for m = 1:2
                        j_inds(running_ind+m) = t_closet(m);
                        vals(running_ind+m) = ((n==3)*(1/6) +...
                            (n>=4)*(1/8))*(1/4);
                    end
                    
                    % assign indices and weight for other farther triangles
                    for m=1:numel(t_others)
                        j_inds(running_ind+2+m) = t_others(m);
                        vals(running_ind+2+m) = ((n==4)*(1/8) +...
                            (n==5)*(beta/2)+...
                            (n>=6)*(1/8))*(1/4);
                    end
                    
                    % update running_ind
                    running_ind = running_ind + n_cutt_off;
                    
                    % update new triangle index
                    new_triangle_ind = new_triangle_ind + 1;
                end
                
                % even rules: assign indices and valse to the middle
                % triangle contribution . this are weights from the main
                % triangle and its 3 neighboring triangles, as assuming no
                % boundaries. TODO: allow boundaries
                i_inds(running_ind:running_ind + 3) = new_triangle_ind;
                j_inds(running_ind:running_ind + 3) = [i, t_adj];
                vals(running_ind:running_ind + 3) = 1/16;
                
                % update running_ind
                running_ind = running_ind + 4;
                
                % update new triangle index
                new_triangle_ind = new_triangle_ind+1;
                
            end
            
            % construct S_2 sparse matrix
            S_2 = sparse(i_inds, j_inds, vals, obj.numFaces*4, obj.numFaces);
            
            
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % subdivide 1-forms (edges)
            disp('S1 calculation');
            
            % edges-faces adjacency list
            
            % get old edges
            E_old = obj.edges(false);
            
            % create the new edge-list by my convention that the
            % oreintation is from smaller index to higher index, PAY
            % ATTENTION TO THE ORIENTATION COMPARED TO THE OLD EDGE WHEN
            % BUILDING THIS WAY
            E_new = unique(sort([F_new(:,[1,2]); F_new(:,[2,3]); F_new(:,[3,1])], 2), 'rows');
            
            % initiating the vectors for sparse S_2 building
            % In the Even rules for each vertex we look at all outoing
            % edges and the cut off number of the ring of edges, for each
            % edge the number of outoing vertices : sigma_i_in_V {(n_i +
            % n_i_cutoff)n_i}
            % Also in the cutoff when there is a triangle with 3 vertices,
            % gamma_0 = 0 and should not be accounted.
            % In the odd rules, each complete new edge is supported by two
            % other triangles (3 including it creator), over all 7 values
            num_adjacent_edges_per_vertex = full(sum(abs(d0)));
            num_adjacent_edges_per_vertex_cut_off = (num_adjacent_edges_per_vertex>=4)*4 + (num_adjacent_edges_per_vertex==3).*(num_adjacent_edges_per_vertex-1);
            edges_num_contrib_even_rules = sum((num_adjacent_edges_per_vertex + num_adjacent_edges_per_vertex_cut_off).*num_adjacent_edges_per_vertex);
            edges_num_contrib_odd_rules  = obj.numFaces*3*7;
            i_inds = zeros(edges_num_contrib_even_rules + edges_num_contrib_odd_rules,1);
            j_inds = zeros(edges_num_contrib_even_rules + edges_num_contrib_odd_rules,1);
            vals   = zeros(edges_num_contrib_even_rules + edges_num_contrib_odd_rules,1);
            
            % run over the new edges and calculate values of Even rules
            % because the new edge list is sorted like the convention,
            % small vertex and then larger, and all new vertices were
            % inserted in V_new after the old vertices, to run over Even
            % rules it means to run on new edges that contain an old
            % vertex. And in the sorted new edge list its to run from 1 to
            % num_edges_old*2.
            running_ind = 1;
            for i = 1:obj.numEdges*2
                
                % get the edge
                e = E_new(i,:);
                
                % old vertex, this is the beggining of edge e (according to
                % the orientation(
                v_old = e(1);
                
                % new vertex
                v_new = e(2);
                
                % get father edge of this new edge
                e_father_ind = v_new-obj.numVertices; % father edge index
                
                e_father = E_old(e_father_ind,:);
                
                % calculate orientation according to the father edge
                sigma_0_orientation = (v_old == e_father(1))*1 + (v_old ~= e_father(1))*(-1);
                
                % get the father edge ind which is where the new edges points to,  according to its orientation to the
                % father end
                if sigma_0_orientation == 1
                    e_towards_v = e_father(2);
                else
                    e_towards_v = e_father(1);
                end
                
                % get all outgoing edges indices from v_old
                [edges_outgoing_inds, ~] = find(abs(d0(:, v_old)) == 1);
                % get rid of the father edges index
                edges_outgoing_inds = setdiff(edges_outgoing_inds, e_father_ind);
                
                % calculate the number of outgoind edges, add 1 for father
                % edge - this is n
                n = numel(edges_outgoing_inds) + 1;
                
                % calculate alpha and beta according to n
                switch n
                    case 3
                        alpha = 3/16;
                        beta  = 1/12;
                    case 4
                        alpha = 3/(8*n);
                        beta  = 1/8;
                    case 5
                        alpha = 3/(8*n);
                        beta  = 1/4 - 1/(16*sin(2*pi/5)^2);
                    otherwise
                        alpha = 3/(8*n);
                        beta  = 1/4;
                end
                
                % update the value for the inds and vals for the sparse matrix
                i_inds(running_ind) = i;
                j_inds(running_ind) = e_father_ind;
                vals(running_ind)   = (3/8 -alpha -beta/4)*sigma_0_orientation;
                running_ind = running_ind + 1;
                
                % get the edge two faces (the edge we work on)
                edge_faces = find(abs(d1(:, e_father_ind)) == 1);
                
                % get the edges on this two faces
                edges_faces_all_edges = triangle_edges(edge_faces, :);
                
                % reduce father edges ind
                edges_faces_all_edges(edges_faces_all_edges == e_father_ind) = [];
                
                % get sigma_1 and sigma_n-1 indices, edges and orientation
                sigma_1_and_n_minus_1_inds = intersect(edges_faces_all_edges, edges_outgoing_inds);
                sigma_1_and_n_minus_1_edges = E_old(sigma_1_and_n_minus_1_inds,:);
                % orientation - +1 if it is outgoing, -1 if it is ingoing
                sigma_1_and_n_minus_1_orientation = (v_old == sigma_1_and_n_minus_1_edges(:,1))*1 + (v_old ~= sigma_1_and_n_minus_1_edges(:,1))*(-1);
                
                % get the rest of sigmas inds
                sigma_rest_inds = setdiff(edges_outgoing_inds, sigma_1_and_n_minus_1_inds);
                
                % get gamma_0 and gamma_n-1 indices, edges and orientation
                gamma_0_and_n_minus_1_inds = setdiff(edges_faces_all_edges, sigma_1_and_n_minus_1_inds);
                gamma_0_and_n_minus_1_edges = E_old(gamma_0_and_n_minus_1_inds,:);
                % orientation - +1 if the egdge end is towards sigma_0 -1
                % else
                edges_gamma_0_and_n_minus_1_orientation = (e_towards_v == gamma_0_and_n_minus_1_edges(:,2))*1 + (e_towards_v ~= gamma_0_and_n_minus_1_edges(:,2))*(-1);
                
                
                % if n==3 we are done, can set this edge sparse inds and
                % vals
                if n == 3
                    
                    % update sigma 1 and sigma n-1 in inds and vals for the
                    % sparse matrix
                    i_inds(running_ind:running_ind+1) = i;
                    j_inds(running_ind:running_ind+1) = sigma_1_and_n_minus_1_inds;
                    vals(running_ind:running_ind+1)   = (1/8 -alpha + beta/8)*sigma_1_and_n_minus_1_orientation;
                    running_ind = running_ind + 2;
                    
                    % update gamma 0 and gamma n-1 in inds and vals for the
                    % sparse matrix
                    i_inds(running_ind:running_ind+1) = i;
                    j_inds(running_ind:running_ind+1) = gamma_0_and_n_minus_1_inds;
                    vals(running_ind:running_ind+1)   = (beta/8)*edges_gamma_0_and_n_minus_1_orientation;
                    running_ind = running_ind + 2;
                    
                    % continure to the next new edge
                    continue;
                    
                else % n==4 or n>4
                    
                    % update sigma 1 and sigma n-1 in inds and vals for the
                    % sparse matrix
                    i_inds(running_ind:running_ind+1) = i;
                    j_inds(running_ind:running_ind+1) = sigma_1_and_n_minus_1_inds;
                    vals(running_ind:running_ind+1)   = (1/8 -alpha)*sigma_1_and_n_minus_1_orientation;
                    running_ind = running_ind + 2;
                    
                    % update gamma 0 and gamma n-1 in inds and vals for the
                    % sparse matrix
                    i_inds(running_ind:running_ind+1) = i;
                    j_inds(running_ind:running_ind+1) = gamma_0_and_n_minus_1_inds;
                    vals(running_ind:running_ind+1)   = (beta/8)*edges_gamma_0_and_n_minus_1_orientation;
                    running_ind = running_ind + 2;
                    
                    %                     if n>=4 %(==4)
                    %                         % when n = 4 we only have sigma_2
                    %
                    %                         % get sigma_2 ind, edge and orientation. the
                    %                         % orientation is +1 if its outgoing edge
                    %                         sigma_2_ind         = sigma_rest_inds;
                    %                         sigma_2_edge        = E_old(sigma_2_ind, :);
                    %                         sigma_2_orientation = (v_old == sigma_2_edge(1))*1 + (v_old ~= sigma_2_edge(1))*(-1);
                    %
                    %                         % update sigma 2 in inds and vals for the sparse
                    %                         % matrix
                    %                         i_inds(running_ind) = i;
                    %                         j_inds(running_ind) = sigma_2_ind;
                    %                         vals(running_ind)   = (-alpha + beta/8)*sigma_2_orientation;
                    %                         vals(running_ind)   = (-alpha + beta/4)*sigma_2_orientation; %TODO: Check why this fix was needed
                    %                         running_ind = running_ind + 1;
                    %
                    %                         % get gamma_1 and gamma_2, inds, edges, orientations.
                    %                         % the orientation is +1 if the end of gamma1/2 edge is
                    %                         % on the end of sigma_1 and sigma_3 (n-1) end
                    %                         sigma_1_and_n_minus_1_ends = setdiff(sigma_1_and_n_minus_1_edges, v_old);
                    %                         gamma_1_2_edges = sort([repmat(setdiff(sigma_2_edge, v_old),2,1), sigma_1_and_n_minus_1_ends], 2);
                    %                         [~, gamma_1_2_inds] = ismember(gamma_1_2_edges(:,:), E_old, 'rows');
                    %                         gamma_1_2_orientation = (sigma_1_and_n_minus_1_ends == gamma_1_2_edges(:,2))*1 + (sigma_1_and_n_minus_1_ends ~= gamma_1_2_edges(:,2))*(-1);
                    %
                    %                         % update gamma_1 and gamma_2 in inds and vals for the sparse
                    %                         % matrix
                    %                         i_inds(running_ind:running_ind+1) = i;
                    %                         j_inds(running_ind:running_ind+1) = gamma_1_2_inds;
                    %                         vals(running_ind:running_ind+1)   = (beta/8)*gamma_1_2_orientation;
                    %                         running_ind = running_ind + 2;
                    %
                    
                    %(else) % n > 4
                    
                    % get the ends of sigma 1 and sigma n-1
                    sigma_1_and_n_minus_1_ends = setdiff(sigma_1_and_n_minus_1_edges,v_old);
                    
                    % get sigma_2 and gamma_1 ind, edge, orientation
                    sigma_1_ind = sigma_1_and_n_minus_1_inds(1);
                    sigma_1_end = sigma_1_and_n_minus_1_ends(1);
                    [simga_1_faces, ~] = find(abs(d1(:,sigma_1_ind)) == 1);
                    sigma_1_2_gamma_1_face = setdiff(simga_1_faces, edge_faces);
                    sigma_1_2_gamma_1_edges_inds = triangle_edges(sigma_1_2_gamma_1_face, :);
                    sigma_2_ind = intersect(sigma_rest_inds,sigma_1_2_gamma_1_edges_inds);
                    sigma_2_edge = E_old(sigma_2_ind,:);
                    sigma_2_orientation = (v_old == sigma_2_edge(1))*1 + (v_old ~= sigma_2_edge(1))*(-1);
                    gamma_1_ind = setdiff(sigma_1_2_gamma_1_edges_inds, [sigma_1_ind, sigma_2_ind]);
                    gamma_1_edge = E_old(gamma_1_ind,:);
                    gamma_1_orientation = (sigma_1_end == gamma_1_edge(2))*1 + (sigma_1_end ~= gamma_1_edge(2))*(-1);
                    
                    % get sigma_n-2 and gamma_n-2 ind, edge, orientation
                    sigma_n_minus_1_ind = sigma_1_and_n_minus_1_inds(2);
                    sigma_n_minus_1_end = sigma_1_and_n_minus_1_ends(2);
                    [simga_n_minus_1_faces, ~] = find(abs(d1(:,sigma_n_minus_1_ind)) == 1);
                    sigma_n_minus_1_n_minus_2_gamma_n_minus_2_face = setdiff(simga_n_minus_1_faces, edge_faces);
                    sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds = triangle_edges(sigma_n_minus_1_n_minus_2_gamma_n_minus_2_face, :);
                    sigma_n_minus_2_ind = intersect(sigma_rest_inds,sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds);
                    sigma_n_minus_2_edge = E_old(sigma_n_minus_2_ind, :);
                    sigma_n_minus_2_orientation = (v_old == sigma_n_minus_2_edge(1))*1 + (v_old ~= sigma_n_minus_2_edge(1))*(-1);
                    gamma_n_minus_2_ind = setdiff(sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds, [sigma_n_minus_1_ind, sigma_n_minus_2_ind]);
                    gamma_n_minus_2_edge = E_old(gamma_n_minus_2_ind,:);
                    gamma_n_minus_2_orientation = (sigma_n_minus_1_end == gamma_n_minus_2_edge(2))*1 + (sigma_n_minus_1_end ~= gamma_n_minus_2_edge(2))*(-1);
                    
                    if n==4
                        
                        % update sigma 2 in inds and vals for the sparse
                        % matrix
                        i_inds(running_ind) = i;
                        j_inds(running_ind) = sigma_2_ind;
                        vals(running_ind)   = (-alpha + beta/8)*sigma_2_orientation;
                        vals(running_ind)   = (-alpha + beta/4)*sigma_2_orientation; %TODO: Check why this fix was needed
                        running_ind = running_ind + 1;
                        
                        % update the gamma_1, gamma_n-2 inds and vals for
                        % the sparse matrix
                        i_inds(running_ind:running_ind+1) = i;
                        j_inds(running_ind:running_ind+1) = [gamma_1_ind, gamma_n_minus_2_ind];
                        vals(running_ind:running_ind+1)   = (beta/8)*[gamma_1_orientation, gamma_n_minus_2_orientation];
                        running_ind = running_ind + 2;
                        continue
                        
                    else %n>=5
                        
                        % update the sigma_2, sigma_n-2 inds and vals for
                        % the sparse matrix
                        i_inds(running_ind:running_ind+1) = i;
                        j_inds(running_ind:running_ind+1) = [sigma_2_ind, sigma_n_minus_2_ind];
                        vals(running_ind:running_ind+1)   = (-alpha + beta/8)*[sigma_2_orientation, sigma_n_minus_2_orientation];
                        running_ind = running_ind + 2;
                        
                        % update the gamma_1, gamma_n-2 inds and vals for
                        % the sparse matrix
                        i_inds(running_ind:running_ind+1) = i;
                        j_inds(running_ind:running_ind+1) = [gamma_1_ind, gamma_n_minus_2_ind];
                        vals(running_ind:running_ind+1)   = (beta/8)*[gamma_1_orientation, gamma_n_minus_2_orientation];
                        running_ind = running_ind + 2;
                    end
                    
                    if n>=6
                        % get the rest of sigmas inds, edges, orientations.
                        rest_of_sigma_rest_inds = setdiff(sigma_rest_inds, [sigma_2_ind, sigma_n_minus_2_ind]);
                        num_rest_rest_sigmas = numel(rest_of_sigma_rest_inds);
                        sigma_rest_rest_edge = E_old(rest_of_sigma_rest_inds, :);
                        sigma_rest_rest_orientation = (v_old == sigma_rest_rest_edge(:,1))*1 + (v_old ~= sigma_rest_rest_edge(:,1))*-1;
                        
                        % update the rest of sigmas inds and vals for
                        % the sparse matrix
                        i_inds(running_ind:running_ind+num_rest_rest_sigmas-1) = i;
                        j_inds(running_ind:running_ind+num_rest_rest_sigmas-1) = rest_of_sigma_rest_inds;
                        vals(running_ind:running_ind+num_rest_rest_sigmas-1)   = (-alpha)*sigma_rest_rest_orientation;
                        running_ind = running_ind + num_rest_rest_sigmas;
                    end
                    
                    
                    
                    %                     end
                end
            end
            
            % run over the rest of the edges, which are the complete new
            % edges, this are the Odd rules
            for i = obj.numEdges*2+1: obj.numEdges*2+obj.numFaces*3
                
                % get the edge
                e = E_new(i,:);
                
                % get father edges of this new edge (an edge between edges)
                e_father_begin_ind = e(1) - obj.numVertices;
                e_father_end_ind = e(2) - obj.numVertices;
                
                % get the vertices of the father edges
                e_father_begin_edge = E_old(e_father_begin_ind,:);
                e_father_end_edge   = E_old(e_father_end_ind, :);
                e_fathers_common_vert = intersect(e_father_begin_edge, e_father_end_edge);
                e_fathers_common_vert2 = e_father_begin_edge(ismembc(e_father_begin_edge, e_father_end_edge));
                assert(isequal(e_fathers_common_vert, e_fathers_common_vert2));
                [all_edges_inds_from_fathers_common_vert, ~] = find(abs(d0(:, e_fathers_common_vert)) == 1 );
                
                % get the new edge orientation
                e_orientation_verts = setxor(e_father_begin_edge, e_father_end_edge, 'stable');
                e_orientation_verts2 = [e_father_begin_edge(~ismembc(e_father_begin_edge,e_father_end_edge)), e_father_end_edge(~ismembc(e_father_end_edge,e_father_begin_edge))];
                assert(isequal(e_orientation_verts, e_orientation_verts2));
                
                % get the fathers triangles
                [e_father_begin_faces, ~] = find(abs(d1(:, e_father_begin_ind)) == 1);
                [e_father_end_faces, ~] = find(abs(d1(:, e_father_end_ind)) == 1);
                
                % get the main face
                e_main_face = intersect(e_father_begin_faces, e_father_end_faces);
                e_main_face2 = e_father_begin_faces(ismembc(e_father_begin_faces, e_father_end_faces));
                assert(isequal(e_main_face, e_main_face2));
                e_main_face_edges_inds = triangle_edges(e_main_face,:);
                
                % get the third edge on the main face
                e_main_parll_ind = setdiff(e_main_face_edges_inds, [e_father_begin_ind, e_father_end_ind]);
                e_main_parll_ind2 = e_main_face_edges_inds(~ismembc(e_main_face_edges_inds, sort([e_father_begin_ind, e_father_end_ind])));
                assert(isequal(e_main_parll_ind, e_main_parll_ind2));
                e_main_parll_edge = E_old(e_main_parll_ind,:);
                e_main_parll_orientation = all(e_orientation_verts == e_main_parll_edge)*1 + any(e_orientation_verts ~= e_main_parll_edge)*(-1);
                
                % update the main parrallel edge in inds and vals for
                % sparse matrix
                i_inds(running_ind) = i;
                j_inds(running_ind) = e_main_parll_ind;
                vals(running_ind)   = (6/32)*e_main_parll_orientation;
                running_ind = running_ind + 1;
                
                % get the orientation of the father edges
                e_father_begin_orientation = (e_father_begin_edge(2) == e_orientation_verts(2) || e_father_begin_edge(1) == e_orientation_verts(1))*1 + (e_father_begin_edge(2) ~= e_orientation_verts(2) && e_father_begin_edge(1) ~= e_orientation_verts(1))*(-1);
                e_father_end_orientation = (e_father_end_edge(2) == e_orientation_verts(2) || e_father_end_edge(1) == e_orientation_verts(1))*1 + (e_father_end_edge(2) ~= e_orientation_verts(2) && e_father_end_edge(1) ~= e_orientation_verts(1))*(-1);
                
                % update the fathers begin and end edges in inds and vals for
                % sparse matrix
                i_inds(running_ind: running_ind + 1) = i;
                j_inds(running_ind: running_ind + 1) = [e_father_begin_ind, e_father_end_ind];
                vals(running_ind: running_ind + 1)   = (3/32)*[e_father_begin_orientation, e_father_end_orientation];
                running_ind = running_ind + 2;
                
                % get the two other edges in the face of father begin
                e_father_begin_other_face = setdiff(e_father_begin_faces, e_main_face);
                e_father_begin_other_face2 = e_father_begin_faces(~ismembc(e_father_begin_faces, e_main_face));
                assert(isequal(e_father_begin_other_face, e_father_begin_other_face2));
                e_father_begin_other_face_edges_inds = triangle_edges(e_father_begin_other_face,:);
                e_father_begin_other_face_edges_inds = setdiff(e_father_begin_other_face_edges_inds, e_father_begin_ind);
                e_father_begin_other_face_edges_inds2 = e_father_begin_other_face_edges_inds(~ismembc(e_father_begin_other_face_edges_inds, e_father_begin_ind));
                assert(isequal(e_father_begin_other_face_edges_inds, e_father_begin_other_face_edges_inds2));
                e_father_begin_upper_ind = intersect(e_father_begin_other_face_edges_inds, all_edges_inds_from_fathers_common_vert);
                e_father_begin_upper_ind2 = e_father_begin_other_face_edges_inds(ismembc(e_father_begin_other_face_edges_inds, all_edges_inds_from_fathers_common_vert));
                assert(isequal(e_father_begin_upper_ind, e_father_begin_upper_ind2));
                e_father_begin_upper_edge = E_old(e_father_begin_upper_ind, :);
                e_father_begin_upper_orientation = (e_father_begin_upper_edge(2) == e_father_begin_edge(2) || e_father_begin_upper_edge(1) == e_father_begin_edge(1))*e_father_begin_orientation +  (e_father_begin_upper_edge(2) ~= e_father_begin_edge(2) && e_father_begin_upper_edge(1) ~= e_father_begin_edge(1))*(-e_father_begin_orientation);
                e_father_begin_lower_ind = setdiff(e_father_begin_other_face_edges_inds, e_father_begin_upper_ind);
                e_father_begin_lower_ind2 = e_father_begin_other_face_edges_inds(~ismembc(e_father_begin_other_face_edges_inds, e_father_begin_upper_ind));
                assert(isequal(e_father_begin_lower_ind, e_father_begin_lower_ind2));
                e_father_begin_lower_edge = E_old(e_father_begin_lower_ind, :);
                e_father_begin_lower_orientation = (e_father_begin_lower_edge(1) == e_main_parll_edge(1) || e_father_begin_lower_edge(2) == e_main_parll_edge(2))*(-e_main_parll_orientation) + (e_father_begin_lower_edge(1) ~= e_main_parll_edge(1) && e_father_begin_lower_edge(2) ~= e_main_parll_edge(2))*(e_main_parll_orientation);
                
                % update the two other edges in the face of father begin in
                % inds and vals for sparse matrix
                i_inds(running_ind: running_ind + 1) = i;
                j_inds(running_ind: running_ind + 1) = [e_father_begin_upper_ind, e_father_begin_lower_ind];
                vals(running_ind: running_ind + 1)   = [(3/32)*e_father_begin_upper_orientation, (1/32)*e_father_begin_lower_orientation];
                running_ind = running_ind + 2;
                
                % get the two other edges in the face of father end
                e_father_end_other_face = setdiff(e_father_end_faces, e_main_face);
                e_father_end_other_face2 = e_father_end_faces(~ismembc(e_father_end_faces, e_main_face));
                assert(isequal(e_father_end_other_face, e_father_end_other_face2));
                e_father_end_other_face_edges_inds = triangle_edges(e_father_end_other_face,:);
                e_father_end_other_face_edges_inds = setdiff(e_father_end_other_face_edges_inds, e_father_end_ind);
                e_father_end_other_face_edges_inds2 = e_father_end_other_face_edges_inds(~ismembc(e_father_end_other_face_edges_inds, e_father_end_ind));
                assert(isequal(e_father_end_other_face_edges_inds, e_father_end_other_face_edges_inds2));
                e_father_end_upper_ind = intersect(e_father_end_other_face_edges_inds, all_edges_inds_from_fathers_common_vert);
                e_father_end_upper_ind2 = e_father_end_upper_ind(ismembc(e_father_end_upper_ind, all_edges_inds_from_fathers_common_vert));
                assert(isequal(e_father_end_upper_ind, e_father_end_upper_ind2));
                e_father_end_upper_edge = E_old(e_father_end_upper_ind, :);
                e_father_end_upper_orientation = (e_father_end_upper_edge(2) == e_father_end_edge(2) || e_father_end_upper_edge(1) == e_father_end_edge(1))*e_father_end_orientation +  (e_father_end_upper_edge(2) ~= e_father_end_edge(2) && e_father_end_upper_edge(1) ~= e_father_end_edge(1))*(-e_father_end_orientation);
                e_father_end_lower_ind = setdiff(e_father_end_other_face_edges_inds, e_father_end_upper_ind);
                e_father_end_lower_ind2 = e_father_end_lower_ind(~ismembc(e_father_end_lower_ind, e_father_end_upper_ind));
                assert(isequal(e_father_end_lower_ind , e_father_end_lower_ind2));
                e_father_end_lower_edge = E_old(e_father_end_lower_ind, :);
                e_father_end_lower_orientation = (e_father_end_lower_edge(1) == e_main_parll_edge(1) || e_father_end_lower_edge(2) == e_main_parll_edge(2))*(-e_main_parll_orientation) + (e_father_end_lower_edge(1) ~= e_main_parll_edge(1) && e_father_end_lower_edge(2) ~= e_main_parll_edge(2))*(e_main_parll_orientation);
                
                % update the two other edges in the face of father end in
                % inds and vals for sparse matrix
                i_inds(running_ind: running_ind + 1) = i;
                j_inds(running_ind: running_ind + 1) = [e_father_end_upper_ind, e_father_end_lower_ind];
                vals(running_ind: running_ind + 1)   = [(3/32)*e_father_end_upper_orientation, (1/32)*e_father_end_lower_orientation];
                running_ind = running_ind + 2;
                
                
            end
            
            % construct S_1 sparse matrix
            S_1 = sparse(i_inds, j_inds, vals, obj.numEdges*2 + 3*obj.numFaces, obj.numEdges);
            
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % return the new mesh
            subdividedMesh = myMeshDEC(P_new, F_new);
        end
        function [subdividedMesh, S_0, S_1, S_2] = subdivideMeshLoop(obj) % Loop subdiviosn
            % subdivideLoop  Perform Loop subdivision for the triangular
            % mesh.
            % Following notations from: "Subdivision Exterior Calculus for
            % Geometry Processing Supplemental Material: Subdivision Rules"
            % Warning: coded not efficiently.
            %
            % Output:
            %   1. subdivdedMesh - the subdivided mesh
            %   2. S_0 - Subdivision matrix for 0-forms
            %   3. S_1 - Subdivision matrix for 1-forms
            %   4. S_2 - Subdivision matrix for 2-forms
            
            % TODO: COMPLETE THE IMPLEMENTATION FOR SURFACE WITH BOUNDARIES
            assert(isempty(obj.boundaryEdges), 'The implementation does not deal yet with surface with a boundary');
            
            % Assert the input is a manifold
            assert(size(obj.edges(1),1)/2 == size(obj.edges(0),1), 'The input is non-manifold');
            
            % Subdvide the vertices (0-form Loop scheme):
            disp('S0 calculation');
            
            % Even vertices (the old vertices new location)
            n = full(sum(obj.adjacencyMatrix,2)); % number of neighbours for each vertex
            alpha = (n>3).*(3./(8*n)) + (n==3)*(3/16); % alpha
            % fast version
            S_0_even = sparse(1:obj.numVertices, 1:obj.numVertices, alpha, obj.numVertices, obj.numVertices)*obj.adjacencyMatrix;
            S_0_even = S_0_even + sparse(1:obj.numVertices, 1:obj.numVertices, (1-alpha.*n), obj.numVertices, obj.numVertices);
            % slow version
            % S_0_even = obj.adjacencyMatrix.*alpha;
            % S_0_even = S_0_even + speye(obj.numVertices).*(1-alpha.*n);
            
            % Odd vertices (new vertices indtroduced by the edges)
            % say the edge is of vertices j,k and on the 2 triangles F1, F2
            % that hold the edge there are two more vertices i,l
            % vertices on the edge: j,k
            d0 = obj.d0_matrix;
            d1 = obj.d1_matrix;
            S_0_odd_edge_verts = abs(d0)*(3/8);
            % 2 triangles of the edge: F1, F2
            [row, ~] = find(abs(d1) == 1);
            % all vertices of the 2 triangles of the edge: i,j(x2),k(x2),l
            all_verts_around_edge = reshape(obj.F(row,:)',6,[])';
            % getting vertices i,j, using: https://stackoverflow.com/questions/32494331/setdiff-row-by-row-without-using-loops-in-matlab
            mask = all(bsxfun(@ne,all_verts_around_edge,permute(obj.E,[1 3 2])),3);
            At = all_verts_around_edge';
            oppsoite_vertices = reshape(At(mask.'),[],size(all_verts_around_edge,1));
            % insert the oppsosite vertices with the right weight
            S_0_odd_opposite_verts = sparse(reshape([1:obj.numEdges;1:obj.numEdges],[],1),oppsoite_vertices(:),1/8,obj.numEdges,obj.numVertices);
            % build S_odd
            S_0_odd = S_0_odd_edge_verts + S_0_odd_opposite_verts;
            
            % S is the combination of S_even (old vertices new wieghts) and
            % S_odd (the new vertices weights)
            S_0 = [S_0_even; S_0_odd];
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % New Triangulation of subdivded mesh
            disp('New triangulation');
            
            % create the new vertices positions
            P_new = S_0*obj.P;
            
            % edges in each triangle
            [triangle_edges, ~] = find(abs(d1') == 1);
            triangle_edges = reshape(triangle_edges,3,[])';
            
            % break triangles and create new ones
            % each triangle gives 4 new ones - not dealing yet with
            % surfaces with boundaries
            F_new = zeros(obj.numFaces*4,3);
            new_triangle_ind = 1;
            for i=1:obj.numFaces
                
                % get the edges of the triangle
                t_edges = triangle_edges(i,:);
                
                % get the list of vertices in the face's three edges
                vertices_in_t_edges = obj.E(t_edges(:),:);
                
                % build the 3 new triangles that include the old vertices
                % that just changed (even) location and two new edge ones
                % (odd). There are numEdges new vertices in the subdivided
                % mesh and they are sorted according to obj.E list.
                % obj.E(1)->the first new vertex etc. Meaning that in the new
                % vertex list the first numVertices vertices are the old
                % ones (even) that changed location (will show itself in P
                % list of xyz). The new vertices will count from
                % numVertices+1 to numVertices + numEdges, since each edge
                % is giving a new vertex.
                for j=1:3
                    
                    % get the vertex
                    v = obj.F(i,j);
                    % update as the first vertex in new F
                    F_new(new_triangle_ind,1) = v;
                    % find which of the edges in the triangle contain v
                    [v_in_t_edges, ~] = find(vertices_in_t_edges == v);
                    vertices_in_t_edges_v = vertices_in_t_edges(v_in_t_edges, :);
                    % get the vertices with the face orientation
                    edge1 = obj.F(i, [j, mod(j+1-1, 3)+1]);
                    % check if the edge we are about to put in place 1
                    % is in the correct way else flip the order. THIS WAY
                    % KEEPING THE TRIANGLE ORIENTATION OF THE COARSE MESH
                    if ~isequal(vertices_in_t_edges_v(1, :), sort(edge1))
                        v_in_t_edges = [v_in_t_edges(2), v_in_t_edges(1)];
                    end
                    % get back the edges indices + the number of vertices
                    new_inds_to_add = t_edges(v_in_t_edges) + obj.numVertices;
                    % adding to the triangle
                    F_new(new_triangle_ind,2:3) = new_inds_to_add ;
                    % update new triangle index
                    new_triangle_ind = new_triangle_ind + 1;
                end
                
                % build the 4th triangle out of the three new vertices from the edges (odd)
                % constructing from previous 3 triangles constructions,
                % since there we chose the right hand rule preserving new
                % edges. because this is the central triangle we need
                % exatcly the opposite set of vertices in order to keep the
                % right hand rule.
                F_new_4th_first_edge = F_new(new_triangle_ind-3, 2:3);
                F_new_4th_second_edge_sorted = sort(F_new(new_triangle_ind-2, 2:3));
                F_new_4th = F_new_4th_second_edge_sorted(~ismembc(F_new_4th_second_edge_sorted, sort(F_new_4th_first_edge)));
                F_new(new_triangle_ind, :) = [F_new_4th_first_edge(1), F_new_4th, F_new_4th_first_edge(2)];
                
                % update new triangle index
                new_triangle_ind = new_triangle_ind+1;
                
            end
            
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Subdvide 2-forms vertices (generalized half-box splines):
            % not dealing yet with boundaries
            disp('S2 calculation');
            
            % faces-faces adjacency
            triangles_triangles_adjacency_matrix = abs(d1*d1')-diag(diag(abs(d1*d1')));
            [t_adj_t, ~] = find(triangles_triangles_adjacency_matrix == 1);
            triangles_triangles_adjacency_list = reshape(t_adj_t, 3, [])';
            
            % vertices-faces adjacency
            triangles_adj_vertices = obj.adjacencyVerticesFaces';
            
            
            % Even 2-forms interior rules (inner new triangles)
            % basicaly what we need is to take each row in the faces-faces
            % adjacency matrix in put it with 1/4 wieght in a larger sparse
            % matrix such that rows goes like this: 1->4, 2->8, 3->12,...
            % We can do this by the following construction with find:
            % each triple (starting from 1) in the following is the 3
            % faces adjacent to the corresponding face (1:3 -> 1, 4:6 ->
            % 2,...), this row vectors containing the indices of the
            % adjacent faces to face i should be located with proper weight
            % in the 4,8,12,... location since when buuilding the
            % subdivided mesh they sit in every 4th location. Since this
            % new faces are composed from the new edges that was the middle
            % of every old edge.
            % again assuming no boundaries then every triangle has 3
            % nieghboring triangles. TODO: allow boundaries
            
            % Odd 2-form interior rules (the 3 remaining new triangles)
            % run over all triangles and build the i,j,v vectors for
            % creating the sparse matrix S_2_odd
            % again assuming no boundaries then every triangle has 3
            % nieghboring triangles. TODO: allow boundaries
            % The flow is as follows: for each triangle get it's
            % adjacent triangles t_adj. run over the triangle 3 vertices
            % (i,j,k), and for each get the neighboring triangles t_adj_v.
            % the intersection between t_adj and t_adj_v will give the two
            % relevant triangles with same weight for the i'th vertex
            % calculation. remembering that the i'th vertex corresponds to
            % the specific new triangle composed of i and two other new
            % vertices from the divided edges i-j, i-k.
            
            % Overall there are 4 new triangles per triangle. 1 to give the
            % even rules and the other 3 with odd rules, so the
            % accmulation of S_2_odd & S_2_even is duable with a single for
            % loop over and triangles and keeping track of indices you can
            % build the vectors to create sparse S_2 comletley
            
            % The counting for number indices to prepare for the sparse
            % matrix creation is: each triangle is splitted to 4: the
            % middle one takes weights (even rules) from 4 triangles, the
            % other 3 take weights from its sorrounding triangles. The
            % calculation gives: 4*|F| + sigma_v
            % num_adjacent_triangles(v)^2. The num_adjacent_triangles
            % should be cut off for not more the 5 adjacent triangles per
            % vertex as can be seen in the supplemntary paper.
            
            % calculate the number of faces that adjacent each vertex
            num_adjacent_faces_per_vertex = sum(obj.adjacencyVerticesFaces,2);
            % cutting off to no more the 5 as can been seen in the
            % supplemntary paper
            num_adjacent_faces_per_vertex_cut_off = (num_adjacent_faces_per_vertex>5)*5 + (num_adjacent_faces_per_vertex<=5).*num_adjacent_faces_per_vertex;
            % effective adjacency for indices initiation
            num_adj_f_v_eff = num_adjacent_faces_per_vertex.*num_adjacent_faces_per_vertex_cut_off;
            % initiating the vectors for sparse S_2 building
            i_inds = zeros(sum(num_adj_f_v_eff) + obj.numFaces*4,1);
            j_inds = zeros(sum(num_adj_f_v_eff) + obj.numFaces*4,1);
            vals = zeros(sum(num_adj_f_v_eff) + obj.numFaces*4,1);
            % initiating tracking inds
            running_ind = 1;
            new_triangle_ind = 1;
            % calculate beta
            beta = 1/4 - 1/(16*sin(2*pi/5)^2);
            
            % run over triangles
            for i=1:obj.numFaces
                
                % adjacent triangles to the triangle
                t_adj = triangles_triangles_adjacency_list(i,:);
                
                % odd rules: each triangle is divided to 4, the first 3 is
                % from the (i,j,k) vertices + 2 new from the divided edges.
                % assuming no boundaries, TODO: allow boundaries
                for j=1:3
                    % get the vertex
                    v = obj.F(i,j);
                    % get the adjacent triangles to this vertex
                    t_adj_v = find(triangles_adj_vertices(:,v) == 1);
                    % number of adjacent tirangles to the vertex after
                    % cut-off
                    n = num_adjacent_faces_per_vertex(v);
                    n_cutt_off = num_adjacent_faces_per_vertex_cut_off(v);
                    % get the two closet relevant adjacent triangles
                    % t_closet = intersect(t_adj_v, t_adj); - slower
                    t_closet = t_adj_v(ismembc(t_adj_v, t_adj));
                    % get the other relevant adjacent triangles (2 at most)
                    % t_others = setdiff(intersect(t_adj_v,reshape(triangles_triangles_adjacency_list(t_closet,:),6,1)),[i; t_closet]); %- slower
                    traingles_near_triangles = reshape(triangles_triangles_adjacency_list(t_closet,:),6,1);
                    % temp = t_adj_v(ismember(t_adj_v, traingles_near_triangles)); %- faster but slower then next temp
                    traingles_near_triangles_sorted = sort(traingles_near_triangles);
                    temp = t_adj_v(ismembc(t_adj_v, traingles_near_triangles_sorted([true; diff(traingles_near_triangles_sorted)~= 0])));
                    t_others = temp(~ismembc(temp, sort([i; t_closet])));
                    % assign indices of weights for j'th new triangle
                    i_inds(running_ind:running_ind + n_cutt_off - 1) = new_triangle_ind;
                    % assign indice and weight from the main triangle
                    j_inds(running_ind) = i;
                    switch n
                        case 3
                            vals(running_ind) = 1/6;
                        case 4
                            vals(running_ind) = 5/32;
                        case 5
                            vals(running_ind) = (3/4 - beta)/4;
                        otherwise %>=6
                            vals(running_ind) = 1/8;
                    end
                    
                    % assign indices and weights for the 2 closet triangles
                    for m = 1:2
                        j_inds(running_ind+m) = t_closet(m);
                        if n == 3
                            vals(running_ind+m) = 1/24;
                        else % n>=4
                            vals(running_ind+m) = 1/32;
                        end
                    end
                    
                    % assign indices and weight for other farther triangles
                    for m=1:numel(t_others)
                        j_inds(running_ind+2+m) = t_others(m);
                        switch n
                            case 4
                                vals(running_ind+2+m) = 1/32;
                            case 5
                                vals(running_ind+2+m) = beta/8;
                            otherwise % >=6
                                vals(running_ind+2+m) = 1/32;
                        end
                    end
                    
                    % update running_ind
                    running_ind = running_ind + n_cutt_off;
                    
                    % update new triangle index
                    new_triangle_ind = new_triangle_ind + 1;
                end
                
                % even rules: assign indices and valse to the middle
                % triangle contribution . this are weights from the main
                % triangle and its 3 neighboring triangles, as assuming no
                % boundaries. TODO: allow boundaries
                i_inds(running_ind:running_ind + 3) = new_triangle_ind;
                j_inds(running_ind:running_ind + 3) = [i, t_adj];
                vals(running_ind:running_ind + 3) = 1/16;
                
                % update running_ind
                running_ind = running_ind + 4;
                
                % update new triangle index
                new_triangle_ind = new_triangle_ind+1;
                
            end
            
            % construct S_2 sparse matrix
            S_2 = sparse(i_inds, j_inds, vals, obj.numFaces*4, obj.numFaces);
            
            
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % subdivide 1-forms (edges)
            disp('S1 calculation');
            
            % edges-faces adjacency list
            
            % get old edges
            E_old = obj.edges(false);
            
            % create the new edge-list by my convention that the
            % oreintation is from smaller index to higher index, PAY
            % ATTENTION TO THE ORIENTATION COMPARED TO THE OLD EDGE WHEN
            % BUILDING THIS WAY
            E_new = unique(sort([F_new(:,[1,2]); F_new(:,[2,3]); F_new(:,[3,1])], 2), 'rows');
            
            % initiating the vectors for sparse S_2 building
            % In the Even rules for each vertex we look at all outoing
            % edges and the cut off number of the ring of edges, for each
            % edge the number of outoing vertices : sigma_i_in_V {(n_i +
            % n_i_cutoff)n_i}
            % Also in the cutoff when there is a triangle with 3 vertices,
            % gamma_0 = 0 and should not be accounted.
            % In the odd rules, each complete new edge is supported by two
            % other triangles (3 including it creator), over all 7 values
            num_adjacent_edges_per_vertex = full(sum(abs(d0)));
            num_adjacent_edges_per_vertex_cut_off = (num_adjacent_edges_per_vertex>=4)*4 + (num_adjacent_edges_per_vertex==3).*(num_adjacent_edges_per_vertex-1);
            edges_num_contrib_even_rules = sum((num_adjacent_edges_per_vertex + num_adjacent_edges_per_vertex_cut_off).*num_adjacent_edges_per_vertex);
            edges_num_contrib_odd_rules  = obj.numFaces*3*7;
            i_inds = zeros(edges_num_contrib_even_rules + edges_num_contrib_odd_rules,1);
            j_inds = zeros(edges_num_contrib_even_rules + edges_num_contrib_odd_rules,1);
            vals   = zeros(edges_num_contrib_even_rules + edges_num_contrib_odd_rules,1);
            
            % run over the new edges and calculate values of Even rules
            % because the new edge list is sorted like the convention,
            % small vertex and then larger, and all new vertices were
            % inserted in V_new after the old vertices, to run over Even
            % rules it means to run on new edges that contain an old
            % vertex. And in the sorted new edge list its to run from 1 to
            % num_edges_old*2.
            running_ind = 1;
            for i = 1:obj.numEdges*2
                
                % get the edge
                e = E_new(i,:);
                
                % old vertex, this is the beggining of edge e (according to
                % the orientation(
                v_old = e(1);
                
                % new vertex
                v_new = e(2);
                
                % get father edge of this new edge
                e_father_ind = v_new-obj.numVertices; % father edge index
                
                e_father = E_old(e_father_ind,:);
                
                % calculate orientation according to the father edge
                sigma_0_orientation = (v_old == e_father(1))*1 + (v_old ~= e_father(1))*(-1);
                
                % get the father edge ind which is where the new edges points to,  according to its orientation to the
                % father end
                if sigma_0_orientation == 1
                    e_towards_v = e_father(2);
                else
                    e_towards_v = e_father(1);
                end
                
                % get all outgoing edges indices from v_old
                edges_outgoing_inds = find(abs(d0(:, v_old)) == 1);
                % get rid of the father edges index
                % edges_outgoing_inds = setdiff(edges_outgoing_inds, e_father_ind); - slower
                edges_outgoing_inds = edges_outgoing_inds(~(edges_outgoing_inds==e_father_ind));
                
                % calculate the number of outgoind edges, add 1 for father
                % edge - this is n
                n = numel(edges_outgoing_inds) + 1;
                
                % calculate alpha and beta according to n
                switch n
                    case 3
                        alpha = 3/16;
                        beta  = 1/12;
                    case 4
                        alpha = 3/(8*n);
                        beta  = 1/8;
                    case 5
                        alpha = 3/(8*n);
                        beta  = 1/4 - 1/(16*sin(2*pi/5)^2);
                    otherwise
                        alpha = 3/(8*n);
                        beta  = 1/4;
                end
                
                % update the value for the inds and vals for the sparse matrix
                i_inds(running_ind) = i;
                j_inds(running_ind) = e_father_ind;
                vals(running_ind)   = (3/8 -alpha -beta/4)*sigma_0_orientation;
                running_ind = running_ind + 1;
                
                % get the edge two faces (the edge we work on)
                edge_faces = find(abs(d1(:, e_father_ind)) == 1);
                
                % get the edges on this two faces
                edges_faces_all_edges = triangle_edges(edge_faces, :);
                
                % reduce father edges ind
                % edges_faces_all_edges(edges_faces_all_edges == e_father_ind) = []; - slower
                edges_faces_all_edges = sort(edges_faces_all_edges(~(edges_faces_all_edges == e_father_ind)));
                
                % get sigma_1 and sigma_n-1 indices, edges and orientation
                % sigma_1_and_n_minus_1_inds = intersect(edges_faces_all_edges, edges_outgoing_inds); - slower
                sigma_1_and_n_minus_1_inds = edges_faces_all_edges(ismembc(edges_faces_all_edges, edges_outgoing_inds));
                sigma_1_and_n_minus_1_edges = E_old(sigma_1_and_n_minus_1_inds,:);
                % orientation - +1 if it is outgoing, -1 if it is ingoing
                sigma_1_and_n_minus_1_orientation = (v_old == sigma_1_and_n_minus_1_edges(:,1))*1 + (v_old ~= sigma_1_and_n_minus_1_edges(:,1))*(-1);
                
                % get the rest of sigmas inds
                % sigma_rest_inds = setdiff(edges_outgoing_inds, sigma_1_and_n_minus_1_inds); - slower
                sigma_rest_inds = edges_outgoing_inds(~ismembc(edges_outgoing_inds, sigma_1_and_n_minus_1_inds));
                
                % get gamma_0 and gamma_n-1 indices, edges and orientation
                % gamma_0_and_n_minus_1_inds = setdiff(edges_faces_all_edges, sigma_1_and_n_minus_1_inds); -slower
                gamma_0_and_n_minus_1_inds = edges_faces_all_edges(~ismembc(edges_faces_all_edges, sigma_1_and_n_minus_1_inds));
                gamma_0_and_n_minus_1_edges = E_old(gamma_0_and_n_minus_1_inds,:);
                % orientation - +1 if the egdge end is towards sigma_0 -1
                % else
                edges_gamma_0_and_n_minus_1_orientation = (e_towards_v == gamma_0_and_n_minus_1_edges(:,2))*1 + (e_towards_v ~= gamma_0_and_n_minus_1_edges(:,2))*(-1);
                
                
                % if n==3 we are done, can set this edge sparse inds and
                % vals
                if n == 3
                    
                    % update sigma 1 and sigma n-1 in inds and vals for the
                    % sparse matrix
                    i_inds(running_ind:running_ind+1) = i;
                    j_inds(running_ind:running_ind+1) = sigma_1_and_n_minus_1_inds;
                    vals(running_ind:running_ind+1)   = (1/8 -alpha + beta/8)*sigma_1_and_n_minus_1_orientation;
                    running_ind = running_ind + 2;
                    
                    % update gamma 0 and gamma n-1 in inds and vals for the
                    % sparse matrix
                    i_inds(running_ind:running_ind+1) = i;
                    j_inds(running_ind:running_ind+1) = gamma_0_and_n_minus_1_inds;
                    vals(running_ind:running_ind+1)   = (beta/8)*edges_gamma_0_and_n_minus_1_orientation;
                    running_ind = running_ind + 2;
                    
                    % continure to the next new edge
                    continue;
                    
                else % n==4 or n>4
                    
                    % update sigma 1 and sigma n-1 in inds and vals for the
                    % sparse matrix
                    i_inds(running_ind:running_ind+1) = i;
                    j_inds(running_ind:running_ind+1) = sigma_1_and_n_minus_1_inds;
                    vals(running_ind:running_ind+1)   = (1/8 -alpha)*sigma_1_and_n_minus_1_orientation;
                    running_ind = running_ind + 2;
                    
                    % update gamma 0 and gamma n-1 in inds and vals for the
                    % sparse matrix
                    i_inds(running_ind:running_ind+1) = i;
                    j_inds(running_ind:running_ind+1) = gamma_0_and_n_minus_1_inds;
                    vals(running_ind:running_ind+1)   = (beta/8)*edges_gamma_0_and_n_minus_1_orientation;
                    running_ind = running_ind + 2;
                    
                    % get the ends of sigma 1 and sigma n-1
                    % sigma_1_and_n_minus_1_ends = setdiff(sigma_1_and_n_minus_1_edges,v_old); - slower
                    sigma_1_end = sigma_1_and_n_minus_1_edges(1,~ismembc(sigma_1_and_n_minus_1_edges(1,:), v_old));
                    sigma_n_minus_1_end = sigma_1_and_n_minus_1_edges(2,~ismembc(sigma_1_and_n_minus_1_edges(2,:), v_old));
                    
                    % get sigma_2 and gamma_1 ind, edge, orientation
                    sigma_1_ind = sigma_1_and_n_minus_1_inds(1);
                    sigma_1_faces = find(abs(d1(:,sigma_1_ind)) == 1);
                    % sigma_1_2_gamma_1_face = setdiff(simga_1_faces, edge_faces); - slower
                    sigma_1_2_gamma_1_face = sigma_1_faces(~ismembc(sigma_1_faces, edge_faces));
                    sigma_1_2_gamma_1_edges_inds = triangle_edges(sigma_1_2_gamma_1_face, :);
                    % sigma_2_ind = intersect(sigma_rest_inds,sigma_1_2_gamma_1_edges_inds); - slower
                    sigma_2_ind = sigma_rest_inds(ismembc(sigma_rest_inds, sigma_1_2_gamma_1_edges_inds));
                    sigma_2_edge = E_old(sigma_2_ind,:);
                    sigma_2_orientation = (v_old == sigma_2_edge(1))*1 + (v_old ~= sigma_2_edge(1))*(-1);
                    % gamma_1_ind = setdiff(sigma_1_2_gamma_1_edges_inds, [sigma_1_ind, sigma_2_ind]); - slower
                    gamma_1_ind = sigma_1_2_gamma_1_edges_inds(~ismembc(sigma_1_2_gamma_1_edges_inds, sort([sigma_1_ind, sigma_2_ind])));
                    gamma_1_edge = E_old(gamma_1_ind,:);
                    gamma_1_orientation = (sigma_1_end == gamma_1_edge(2))*1 + (sigma_1_end ~= gamma_1_edge(2))*(-1);
                    
                    % get sigma_n-2 and gamma_n-2 ind, edge, orientation
                    sigma_n_minus_1_ind = sigma_1_and_n_minus_1_inds(2);
                    sigma_n_minus_1_faces = find(abs(d1(:,sigma_n_minus_1_ind)) == 1);
                    % sigma_n_minus_1_n_minus_2_gamma_n_minus_2_face = setdiff(simga_n_minus_1_faces, edge_faces); - slower
                    sigma_n_minus_1_n_minus_2_gamma_n_minus_2_face = sigma_n_minus_1_faces(~ismembc(sigma_n_minus_1_faces, edge_faces));
                    sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds = triangle_edges(sigma_n_minus_1_n_minus_2_gamma_n_minus_2_face, :);
                    % sigma_n_minus_2_ind = intersect(sigma_rest_inds,sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds); - slower
                    sigma_n_minus_2_ind = sigma_rest_inds(ismembc(sigma_rest_inds, sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds));
                    sigma_n_minus_2_edge = E_old(sigma_n_minus_2_ind, :);
                    sigma_n_minus_2_orientation = (v_old == sigma_n_minus_2_edge(1))*1 + (v_old ~= sigma_n_minus_2_edge(1))*(-1);
                    % gamma_n_minus_2_ind = setdiff(sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds, [sigma_n_minus_1_ind, sigma_n_minus_2_ind]); - slower
                    gamma_n_minus_2_ind = sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds(~ismembc(sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds, sort([sigma_n_minus_1_ind, sigma_n_minus_2_ind])));
                    gamma_n_minus_2_edge = E_old(gamma_n_minus_2_ind,:);
                    gamma_n_minus_2_orientation = (sigma_n_minus_1_end == gamma_n_minus_2_edge(2))*1 + (sigma_n_minus_1_end ~= gamma_n_minus_2_edge(2))*(-1);
                    
                    if n==4
                        
                        % update sigma 2 in inds and vals for the sparse
                        % matrix
                        i_inds(running_ind) = i;
                        j_inds(running_ind) = sigma_2_ind;
                        vals(running_ind)   = (-alpha + beta/8)*sigma_2_orientation;
                        vals(running_ind)   = (-alpha + beta/4)*sigma_2_orientation; %TODO: Check why this fix was needed
                        running_ind = running_ind + 1;
                        
                        % update the gamma_1, gamma_n-2 inds and vals for
                        % the sparse matrix
                        i_inds(running_ind:running_ind+1) = i;
                        j_inds(running_ind:running_ind+1) = [gamma_1_ind, gamma_n_minus_2_ind];
                        vals(running_ind:running_ind+1)   = (beta/8)*[gamma_1_orientation, gamma_n_minus_2_orientation];
                        running_ind = running_ind + 2;
                        continue
                        
                    else %n>=5
                        
                        % update the sigma_2, sigma_n-2 inds and vals for
                        % the sparse matrix
                        i_inds(running_ind:running_ind+1) = i;
                        j_inds(running_ind:running_ind+1) = [sigma_2_ind, sigma_n_minus_2_ind];
                        vals(running_ind:running_ind+1)   = (-alpha + beta/8)*[sigma_2_orientation, sigma_n_minus_2_orientation];
                        running_ind = running_ind + 2;
                        
                        % update the gamma_1, gamma_n-2 inds and vals for
                        % the sparse matrix
                        i_inds(running_ind:running_ind+1) = i;
                        j_inds(running_ind:running_ind+1) = [gamma_1_ind, gamma_n_minus_2_ind];
                        vals(running_ind:running_ind+1)   = (beta/8)*[gamma_1_orientation, gamma_n_minus_2_orientation];
                        running_ind = running_ind + 2;
                    end
                    
                    if n>=6
                        % get the rest of sigmas inds, edges, orientations.
                        % rest_of_sigma_rest_inds = setdiff(sigma_rest_inds, [sigma_2_ind, sigma_n_minus_2_ind]); - slower
                        rest_of_sigma_rest_inds = sigma_rest_inds(~ismembc(sigma_rest_inds, sort([sigma_2_ind, sigma_n_minus_2_ind])));
                        num_rest_rest_sigmas = numel(rest_of_sigma_rest_inds);
                        sigma_rest_rest_edge = E_old(rest_of_sigma_rest_inds, :);
                        sigma_rest_rest_orientation = (v_old == sigma_rest_rest_edge(:,1))*1 + (v_old ~= sigma_rest_rest_edge(:,1))*-1;
                        
                        % update the rest of sigmas inds and vals for
                        % the sparse matrix
                        i_inds(running_ind:running_ind+num_rest_rest_sigmas-1) = i;
                        j_inds(running_ind:running_ind+num_rest_rest_sigmas-1) = rest_of_sigma_rest_inds;
                        vals(running_ind:running_ind+num_rest_rest_sigmas-1)   = (-alpha)*sigma_rest_rest_orientation;
                        running_ind = running_ind + num_rest_rest_sigmas;
                    end
                end
            end
            
            % run over the rest of the edges, which are the complete new
            % edges, this are the Odd rules
            for i = obj.numEdges*2+1: obj.numEdges*2+obj.numFaces*3
                
                % get the edge
                e = E_new(i,:);
                
                % get father edges of this new edge (an edge between edges)
                e_father_begin_ind = e(1) - obj.numVertices;
                e_father_end_ind = e(2) - obj.numVertices;
                
                % get the vertices of the father edges
                e_father_begin_edge = E_old(e_father_begin_ind,:);
                e_father_end_edge   = E_old(e_father_end_ind, :);
                % e_fathers_common_vert = intersect(e_father_begin_edge, e_father_end_edge); - slower
                e_fathers_common_vert = e_father_begin_edge(ismembc(e_father_begin_edge, e_father_end_edge));
                all_edges_inds_from_fathers_common_vert = find(abs(d0(:, e_fathers_common_vert)) == 1 );
                
                % get the new edge orientation
                % e_orientation_verts = setxor(e_father_begin_edge, e_father_end_edge, 'stable'); - slower
                e_orientation_verts = [e_father_begin_edge(~ismembc(e_father_begin_edge,e_father_end_edge)), e_father_end_edge(~ismembc(e_father_end_edge,e_father_begin_edge))];
                
                % get the fathers triangles
                e_father_begin_faces = find(abs(d1(:, e_father_begin_ind)) == 1);
                e_father_end_faces = find(abs(d1(:, e_father_end_ind)) == 1);
                
                % get the main face
                % e_main_face = intersect(e_father_begin_faces, e_father_end_faces); - slower
                e_main_face = e_father_begin_faces(ismembc(e_father_begin_faces, e_father_end_faces));
                e_main_face_edges_inds = triangle_edges(e_main_face,:);
                
                % get the third edge on the main face
                % e_main_parll_ind = setdiff(e_main_face_edges_inds, [e_father_begin_ind, e_father_end_ind]); - slower
                e_main_parll_ind = e_main_face_edges_inds(~ismembc(e_main_face_edges_inds, sort([e_father_begin_ind, e_father_end_ind])));
                e_main_parll_edge = E_old(e_main_parll_ind,:);
                e_main_parll_orientation = all(e_orientation_verts == e_main_parll_edge)*1 + any(e_orientation_verts ~= e_main_parll_edge)*(-1);
                
                % update the main parrallel edge in inds and vals for
                % sparse matrix
                i_inds(running_ind) = i;
                j_inds(running_ind) = e_main_parll_ind;
                vals(running_ind)   = (6/32)*e_main_parll_orientation;
                running_ind = running_ind + 1;
                
                % get the orientation of the father edges
                e_father_begin_orientation = (e_father_begin_edge(2) == e_orientation_verts(2) || e_father_begin_edge(1) == e_orientation_verts(1))*1 + (e_father_begin_edge(2) ~= e_orientation_verts(2) && e_father_begin_edge(1) ~= e_orientation_verts(1))*(-1);
                e_father_end_orientation = (e_father_end_edge(2) == e_orientation_verts(2) || e_father_end_edge(1) == e_orientation_verts(1))*1 + (e_father_end_edge(2) ~= e_orientation_verts(2) && e_father_end_edge(1) ~= e_orientation_verts(1))*(-1);
                
                % update the fathers begin and end edges in inds and vals for
                % sparse matrix
                i_inds(running_ind: running_ind + 1) = i;
                j_inds(running_ind: running_ind + 1) = [e_father_begin_ind, e_father_end_ind];
                vals(running_ind: running_ind + 1)   = (3/32)*[e_father_begin_orientation, e_father_end_orientation];
                running_ind = running_ind + 2;
                
                % get the two other edges in the face of father begin
                % e_father_begin_other_face = setdiff(e_father_begin_faces, e_main_face); - slower
                e_father_begin_other_face = e_father_begin_faces(~ismembc(e_father_begin_faces, e_main_face));
                e_father_begin_other_face_edges_inds = triangle_edges(e_father_begin_other_face,:);
                % e_father_begin_other_face_edges_inds = setdiff(e_father_begin_other_face_edges_inds, e_father_begin_ind); - slower
                e_father_begin_other_face_edges_inds = e_father_begin_other_face_edges_inds(~ismembc(e_father_begin_other_face_edges_inds, e_father_begin_ind));
                % e_father_begin_upper_ind = intersect(e_father_begin_other_face_edges_inds, all_edges_inds_from_fathers_common_vert); - slower
                e_father_begin_upper_ind = e_father_begin_other_face_edges_inds(ismembc(e_father_begin_other_face_edges_inds, all_edges_inds_from_fathers_common_vert));
                e_father_begin_upper_edge = E_old(e_father_begin_upper_ind, :);
                e_father_begin_upper_orientation = (e_father_begin_upper_edge(2) == e_father_begin_edge(2) || e_father_begin_upper_edge(1) == e_father_begin_edge(1))*e_father_begin_orientation +  (e_father_begin_upper_edge(2) ~= e_father_begin_edge(2) && e_father_begin_upper_edge(1) ~= e_father_begin_edge(1))*(-e_father_begin_orientation);
                % e_father_begin_lower_ind = setdiff(e_father_begin_other_face_edges_inds, e_father_begin_upper_ind); - slower
                e_father_begin_lower_ind = e_father_begin_other_face_edges_inds(~ismembc(e_father_begin_other_face_edges_inds, e_father_begin_upper_ind));
                e_father_begin_lower_edge = E_old(e_father_begin_lower_ind, :);
                e_father_begin_lower_orientation = (e_father_begin_lower_edge(1) == e_main_parll_edge(1) || e_father_begin_lower_edge(2) == e_main_parll_edge(2))*(-e_main_parll_orientation) + (e_father_begin_lower_edge(1) ~= e_main_parll_edge(1) && e_father_begin_lower_edge(2) ~= e_main_parll_edge(2))*(e_main_parll_orientation);
                
                % update the two other edges in the face of father begin in
                % inds and vals for sparse matrix
                i_inds(running_ind: running_ind + 1) = i;
                j_inds(running_ind: running_ind + 1) = [e_father_begin_upper_ind, e_father_begin_lower_ind];
                vals(running_ind: running_ind + 1)   = [(3/32)*e_father_begin_upper_orientation, (1/32)*e_father_begin_lower_orientation];
                running_ind = running_ind + 2;
                
                % get the two other edges in the face of father end
                % e_father_end_other_face = setdiff(e_father_end_faces, e_main_face); - slower
                e_father_end_other_face = e_father_end_faces(~ismembc(e_father_end_faces, e_main_face));
                e_father_end_other_face_edges_inds = triangle_edges(e_father_end_other_face,:);
                % e_father_end_other_face_edges_inds = setdiff(e_father_end_other_face_edges_inds, e_father_end_ind); - slower
                e_father_end_other_face_edges_inds = e_father_end_other_face_edges_inds(~ismembc(e_father_end_other_face_edges_inds, e_father_end_ind));
                % e_father_end_upper_ind = intersect(e_father_end_other_face_edges_inds, all_edges_inds_from_fathers_common_vert); - slower
                e_father_end_upper_ind = e_father_end_other_face_edges_inds(ismembc(e_father_end_other_face_edges_inds, all_edges_inds_from_fathers_common_vert));
                e_father_end_upper_edge = E_old(e_father_end_upper_ind, :);
                e_father_end_upper_orientation = (e_father_end_upper_edge(2) == e_father_end_edge(2) || e_father_end_upper_edge(1) == e_father_end_edge(1))*e_father_end_orientation +  (e_father_end_upper_edge(2) ~= e_father_end_edge(2) && e_father_end_upper_edge(1) ~= e_father_end_edge(1))*(-e_father_end_orientation);
                % e_father_end_lower_ind = setdiff(e_father_end_other_face_edges_inds, e_father_end_upper_ind); - slower
                e_father_end_lower_ind = e_father_end_other_face_edges_inds(~ismembc(e_father_end_other_face_edges_inds, e_father_end_upper_ind));
                e_father_end_lower_edge = E_old(e_father_end_lower_ind, :);
                e_father_end_lower_orientation = (e_father_end_lower_edge(1) == e_main_parll_edge(1) || e_father_end_lower_edge(2) == e_main_parll_edge(2))*(-e_main_parll_orientation) + (e_father_end_lower_edge(1) ~= e_main_parll_edge(1) && e_father_end_lower_edge(2) ~= e_main_parll_edge(2))*(e_main_parll_orientation);
                
                % update the two other edges in the face of father end in
                % inds and vals for sparse matrix
                i_inds(running_ind: running_ind + 1) = i;
                j_inds(running_ind: running_ind + 1) = [e_father_end_upper_ind, e_father_end_lower_ind];
                vals(running_ind: running_ind + 1)   = [(3/32)*e_father_end_upper_orientation, (1/32)*e_father_end_lower_orientation];
                running_ind = running_ind + 2;
                
            end
            
            % construct S_1 sparse matrix
            S_1 = sparse(i_inds, j_inds, vals, obj.numEdges*2 + 3*obj.numFaces, obj.numEdges);
            
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % return the new mesh
            subdividedMesh = myMeshDEC(P_new, F_new);
        end
        function [subdividedMesh, S_0, S_1, S_2] = subdivideMeshLoop_parfor(obj) % Loop subdiviosn
            % subdivideLoop  Perform Loop subdivision for the triangular
            % mesh.
            % Following notations from: "Subdivision Exterior Calculus for
            % Geometry Processing Supplemental Material: Subdivision Rules"
            % Warning: coded not efficiently.
            %
            % Output:
            %   1. subdivdedMesh - the subdivided mesh
            %   2. S_0 - Subdivision matrix for 0-forms
            %   3. S_1 - Subdivision matrix for 1-forms
            %   4. S_2 - Subdivision matrix for 2-forms
            
            % TODO: COMPLETE THE IMPLEMENTATION FOR SURFACE WITH BOUNDARIES
            assert(isempty(obj.boundaryEdges), 'The implementation does not deal yet with surface with a boundary');
            
            % Assert the input is a manifold
            assert(size(obj.edges(1),1)/2 == size(obj.edges(0),1), 'The input is non-manifold');
            
            % Subdvide the vertices (0-form Loop scheme):
            %disp('S0 calculation');
            
            % Even vertices (the old vertices new location)
            n = full(sum(obj.adjacencyMatrix,2)); % number of neighbours for each vertex
            alpha = (n>3).*(3./(8*n)) + (n==3)*(3/16); % alpha
            % fast version
            S_0_even = sparse(1:obj.numVertices, 1:obj.numVertices, alpha, obj.numVertices, obj.numVertices)*obj.adjacencyMatrix;
            S_0_even = S_0_even + sparse(1:obj.numVertices, 1:obj.numVertices, (1-alpha.*n), obj.numVertices, obj.numVertices);
            % slow version
            % S_0_even = obj.adjacencyMatrix.*alpha;
            % S_0_even = S_0_even + speye(obj.numVertices).*(1-alpha.*n);
            
            % Odd vertices (new vertices indtroduced by the edges)
            % say the edge is of vertices j,k and on the 2 triangles F1, F2
            % that hold the edge there are two more vertices i,l
            % vertices on the edge: j,k
            d0 = obj.d0_matrix;
            d1 = obj.d1_matrix;
            S_0_odd_edge_verts = abs(d0)*(3/8);
            % 2 triangles of the edge: F1, F2
            [row, ~] = find(abs(d1) == 1);
            % all vertices of the 2 triangles of the edge: i,j(x2),k(x2),l
            all_verts_around_edge = reshape(obj.F(row,:)',6,[])';
            % getting vertices i,j, using: https://stackoverflow.com/questions/32494331/setdiff-row-by-row-without-using-loops-in-matlab
            mask = all(bsxfun(@ne,all_verts_around_edge,permute(obj.E,[1 3 2])),3);
            At = all_verts_around_edge';
            oppsoite_vertices = reshape(At(mask.'),[],size(all_verts_around_edge,1));
            % insert the oppsosite vertices with the right weight
            S_0_odd_opposite_verts = sparse(reshape([1:obj.numEdges;1:obj.numEdges],[],1),oppsoite_vertices(:),1/8,obj.numEdges,obj.numVertices);
            % build S_odd
            S_0_odd = S_0_odd_edge_verts + S_0_odd_opposite_verts;
            
            % S is the combination of S_even (old vertices new wieghts) and
            % S_odd (the new vertices weights)
            S_0 = [S_0_even; S_0_odd];
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % New Triangulation of subdivded mesh
            %disp('New triangulation');
            
            % create the new vertices positions
            P_new = S_0*obj.P;
            
            % edges in each triangle
            [triangle_edges, ~] = find(abs(d1') == 1);
            triangle_edges = reshape(triangle_edges,3,[])';
            
            % break triangles and create new ones
            % each triangle gives 4 new ones - not dealing yet with
            % surfaces with boundaries
            %             F_new = zeros(obj.numFaces*4,3);
            
            F_new_cell_array = cell(obj.numFaces, 1);
            vertices_in_t_edges_all_faces = mat2cell(obj.E(reshape(triangle_edges',[],1), :), repmat(3, 1, obj.numFaces));
            
            %             new_triangle_ind = 1;
            parfor i=1:obj.numFaces
                
                % calculate new triangle ind for this face
                %                 new_triangle_ind = (i-1)*4+1;
                
                % initialize new 4 faces
                new_4_faces = nan(4,3);
                new_face_ind = 1;
                
                % get the edges of the triangle
                t_edges = triangle_edges(i,:);
                
                % get the list of vertices in the face's three edges
                %                 vertices_in_t_edges = obj.E(t_edges(:),:);
                vertices_in_t_edges = vertices_in_t_edges_all_faces{i};
                
                
                % build the 3 new triangles that include the old vertices
                % that just changed (even) location and two new edge ones
                % (odd). There are numEdges new vertices in the subdivided
                % mesh and they are sorted according to obj.E list.
                % obj.E(1)->the first new vertex etc. Meaning that in the new
                % vertex list the first numVertices vertices are the old
                % ones (even) that changed location (will show itself in P
                % list of xyz). The new vertices will count from
                % numVertices+1 to numVertices + numEdges, since each edge
                % is giving a new vertex.
                for j=1:3
                    
                    % get the vertex
                    v = obj.F(i,j);
                    % update as the first vertex in new F
                    new_4_faces(new_face_ind, 1) = v;
                    %                     F_new(new_triangle_ind,1) = v;
                    % find which of the edges in the triangle contain v
                    [v_in_t_edges, ~] = find(vertices_in_t_edges == v);
                    vertices_in_t_edges_v = vertices_in_t_edges(v_in_t_edges, :);
                    % get the vertices with the face orientation
                    edge1 = obj.F(i, [j, mod(j+1-1, 3)+1]);
                    % check if the edge we are about to put in place 1
                    % is in the correct way else flip the order. THIS WAY
                    % KEEPING THE TRIANGLE ORIENTATION OF THE COARSE MESH
                    if ~isequal(vertices_in_t_edges_v(1, :), sort(edge1))
                        v_in_t_edges = [v_in_t_edges(2), v_in_t_edges(1)];
                    end
                    % get back the edges indices + the number of vertices
                    new_inds_to_add = t_edges(v_in_t_edges) + obj.numVertices;
                    % adding to the triangle
                    new_4_faces(new_face_ind, 2:3) = new_inds_to_add;
                    %                     F_new(new_triangle_ind,2:3) = new_inds_to_add ;
                    % update new triangle index
                    new_face_ind = new_face_ind + 1;
                    %                     new_triangle_ind = new_triangle_ind + 1;
                end
                
                % build the 4th triangle out of the three new vertices from the edges (odd)
                % constructing from previous 3 triangles constructions,
                % since there we chose the right hand rule preserving new
                % edges. because this is the central triangle we need
                % exatcly the opposite set of vertices in order to keep the
                % right hand rule.
                F_new_4th_first_edge = new_4_faces(new_face_ind-3, 2:3);
                F_new_4th_second_edge_sorted = sort(new_4_faces(new_face_ind-2, 2:3));
                %                 F_new_4th_first_edge = F_new(new_triangle_ind-3, 2:3);
                %                 F_new_4th_second_edge_sorted = sort(F_new(new_triangle_ind-2, 2:3));
                F_new_4th = F_new_4th_second_edge_sorted(~ismembc(F_new_4th_second_edge_sorted, sort(F_new_4th_first_edge)));
                new_4_faces(new_face_ind, :) = [F_new_4th_first_edge(1), F_new_4th, F_new_4th_first_edge(2)];
                %                 F_new(new_triangle_ind, :) = [F_new_4th_first_edge(1), F_new_4th, F_new_4th_first_edge(2)];
                
                % update new triangle index
                %                 new_triangle_ind = new_triangle_ind+1;
                
                % update the 4 new faces
                F_new_cell_array{i} = new_4_faces;
                
            end
            F_new = cell2mat(F_new_cell_array);
            
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Subdvide 2-forms vertices (generalized half-box splines):
            % not dealing yet with boundaries
            %disp('S2 calculation');
            
            % faces-faces adjacency
            triangles_triangles_adjacency_matrix = abs(d1*d1')-diag(diag(abs(d1*d1')));
            [t_adj_t, ~] = find(triangles_triangles_adjacency_matrix == 1);
            triangles_triangles_adjacency_list = reshape(t_adj_t, 3, [])';
            
            % vertices-faces adjacency
            triangles_adj_vertices = obj.adjacencyVerticesFaces';
            
            
            % Even 2-forms interior rules (inner new triangles)
            % basicaly what we need is to take each row in the faces-faces
            % adjacency matrix in put it with 1/4 wieght in a larger sparse
            % matrix such that rows goes like this: 1->4, 2->8, 3->12,...
            % We can do this by the following construction with find:
            % each triple (starting from 1) in the following is the 3
            % faces adjacent to the corresponding face (1:3 -> 1, 4:6 ->
            % 2,...), this row vectors containing the indices of the
            % adjacent faces to face i should be located with proper weight
            % in the 4,8,12,... location since when buuilding the
            % subdivided mesh they sit in every 4th location. Since this
            % new faces are composed from the new edges that was the middle
            % of every old edge.
            % again assuming no boundaries then every triangle has 3
            % nieghboring triangles. TODO: allow boundaries
            
            % Odd 2-form interior rules (the 3 remaining new triangles)
            % run over all triangles and build the i,j,v vectors for
            % creating the sparse matrix S_2_odd
            % again assuming no boundaries then every triangle has 3
            % nieghboring triangles. TODO: allow boundaries
            % The flow is as follows: for each triangle get it's
            % adjacent triangles t_adj. run over the triangle 3 vertices
            % (i,j,k), and for each get the neighboring triangles t_adj_v.
            % the intersection between t_adj and t_adj_v will give the two
            % relevant triangles with same weight for the i'th vertex
            % calculation. remembering that the i'th vertex corresponds to
            % the specific new triangle composed of i and two other new
            % vertices from the divided edges i-j, i-k.
            
            % Overall there are 4 new triangles per triangle. 1 to give the
            % even rules and the other 3 with odd rules, so the
            % accmulation of S_2_odd & S_2_even is duable with a single for
            % loop over and triangles and keeping track of indices you can
            % build the vectors to create sparse S_2 comletley
            
            % The counting for number indices to prepare for the sparse
            % matrix creation is: each triangle is splitted to 4: the
            % middle one takes weights (even rules) from 4 triangles, the
            % other 3 take weights from its sorrounding triangles. The
            % calculation gives: 4*|F| + sigma_v
            % num_adjacent_triangles(v)^2. The num_adjacent_triangles
            % should be cut off for not more the 5 adjacent triangles per
            % vertex as can be seen in the supplemntary paper.
            
            % calculate the number of faces that adjacent each vertex
            num_adjacent_faces_per_vertex = sum(obj.adjacencyVerticesFaces,2);
            % cutting off to no more the 5 as can been seen in the
            % supplemntary paper
            num_adjacent_faces_per_vertex_cut_off = (num_adjacent_faces_per_vertex>5)*5 + (num_adjacent_faces_per_vertex<=5).*num_adjacent_faces_per_vertex;
            % effective adjacency for indices initiation
            %             num_adj_f_v_eff = num_adjacent_faces_per_vertex.*num_adjacent_faces_per_vertex_cut_off;
            
            % initiating the vectors for sparse S_2 building
            %             i_inds = zeros(sum(num_adj_f_v_eff) + obj.numFaces*4,1);
            %             j_inds = zeros(sum(num_adj_f_v_eff) + obj.numFaces*4,1);
            %             vals = zeros(sum(num_adj_f_v_eff) + obj.numFaces*4,1);
            
            % initiating tracking inds
            %             running_inds = zeros(obj.numFaces,1);
            %             for i = 1:obj.numFaces
            %                 running_inds(i) = sum(num_adjacent_faces_per_vertex_cut_off(obj.faces(i,:)))+4;
            %             end
            %             running_inds = [0; cumsum(running_inds)];
            
            %             running_ind = 1;
            %             new_triangle_ind = 1;
            
            % calculate beta
            beta = 1/4 - 1/(16*sin(2*pi/5)^2);
            
            % initiate cells to fill for each face
            i_inds_cell_array  = cell(1, obj.numFaces);
            j_inds_cell_array  = cell(1, obj.numFaces);
            vals_cell_array    = cell(1, obj.numFaces);
            
            faces = obj.F;
            n_faces = full(num_adjacent_faces_per_vertex(obj.F));
            n_cut_off_faces = full(num_adjacent_faces_per_vertex_cut_off(obj.F));
            t_adj_v_faces = cell(obj.numFaces,1 );
            for i =1:obj.numFaces
                t_adj_v_faces{i}{1} = find(triangles_adj_vertices(:,faces(i,1)) == 1);
                t_adj_v_faces{i}{2} = find(triangles_adj_vertices(:,faces(i,2)) == 1);
                t_adj_v_faces{i}{3} = find(triangles_adj_vertices(:,faces(i,3)) == 1);
            end
            triangles_triangles_adjacency_list_faces = cell(obj.numFaces,1 );
            for i=1:obj.numFaces
                triangles_triangles_adjacency_list_faces{i}{1} = triangles_triangles_adjacency_list(t_adj_v_faces{i}{1}, :);
                triangles_triangles_adjacency_list_faces{i}{2} = triangles_triangles_adjacency_list(t_adj_v_faces{i}{2}, :);
                triangles_triangles_adjacency_list_faces{i}{3} = triangles_triangles_adjacency_list(t_adj_v_faces{i}{3}, :);
            end
            
            
            % run over triangles
            parfor i=1:obj.numFaces
                
                % prepare running inds
                %                 running_ind = running_inds(i) + 1;
                new_triangle_ind = (i-1)*4+1;
                
                % adjacent triangles to the triangle
                t_adj = triangles_triangles_adjacency_list(i,:);
                
                n_face = n_faces(i, :);
                n_cut_off_face = n_cut_off_faces(i, :);
                t_adj_v_face = t_adj_v_faces{i};
                triangles_triangles_adjacency_list_face = triangles_triangles_adjacency_list_faces{i};
                
                
                % odd rules: each triangle is divided to 4, the first 3 is
                % from the (i,j,k) vertices + 2 new from the divided edges.
                % assuming no boundaries, TODO: allow boundaries
                for j=1:3
                    % get the vertex
                    %                     v = faces(i,j);
                    % get the adjacent triangles to this vertex
                    %                     t_adj_v = find(triangles_adj_vertices(:,v) == 1);
                    t_adj_v = t_adj_v_face{j};
                    % number of adjacent tirangles to the vertex after
                    % cut-off
                    %n = num_adjacent_faces_per_vertex(v);
                    n = n_face(j);
                    %assert(isequal(n,n2));
                    %n_cutt_off = num_adjacent_faces_per_vertex_cut_off(v);
                    n_cutt_off = n_cut_off_face(j);
                    %assert(isequal(n_cutt_off, n_cutt_off2));
                    % get the two closet relevant adjacent triangles
                    % t_closet = intersect(t_adj_v, t_adj); - slower
                    t_closet = t_adj_v(ismembc(t_adj_v, t_adj));
                    % get the other relevant adjacent triangles (2 at most)
                    % t_others = setdiff(intersect(t_adj_v,reshape(triangles_triangles_adjacency_list(t_closet,:),6,1)),[i; t_closet]); %- slower
                    %                     traingles_near_triangles = reshape(triangles_triangles_adjacency_list(t_closet,:),6,1);
                    traingles_near_triangles = reshape(triangles_triangles_adjacency_list_face{j}(ismembc(t_adj_v, t_adj), :),6,1);
                    %                     assert(isequal(traingles_near_triangles, traingles_near_triangles2));
                    % temp = t_adj_v(ismember(t_adj_v, traingles_near_triangles)); %- faster but slower then next temp
                    traingles_near_triangles_sorted = sort(traingles_near_triangles);
                    temp = t_adj_v(ismembc(t_adj_v, traingles_near_triangles_sorted([true; diff(traingles_near_triangles_sorted)~= 0])));
                    t_others = temp(~ismembc(temp, sort([i; t_closet])));
                    % assign indices of weights for j'th new triangle
                    i_inds_cell_array{i} = [i_inds_cell_array{i}, repmat(new_triangle_ind, 1, n_cutt_off)];
                    %                     i_inds(running_ind:running_ind + n_cutt_off - 1) = new_triangle_ind;
                    % assign indice and weight from the main triangle
                    j_inds_cell_array{i} = [j_inds_cell_array{i}, i];
                    %                     j_inds(running_ind) = i;
                    switch n
                        case 3
                            vals_cell_array{i} = [vals_cell_array{i}, 1/6];
                            %                             vals(running_ind) = 1/6;
                        case 4
                            vals_cell_array{i} = [vals_cell_array{i}, 5/32];
                            %                             vals(running_ind) = 5/32;
                        case 5
                            vals_cell_array{i} = [vals_cell_array{i}, (3/4 - beta)/4];
                            %                             vals(running_ind) = (3/4 - beta)/4;
                        otherwise %>=6
                            vals_cell_array{i} = [vals_cell_array{i}, 1/8];
                            %                             vals(running_ind) = 1/8;
                    end
                    
                    % assign indices and weights for the 2 closet triangles
                    for m = 1:2
                        j_inds_cell_array{i} = [j_inds_cell_array{i}, t_closet(m)];
                        %                         j_inds(running_ind+m) = t_closet(m);
                        if n == 3
                            vals_cell_array{i} = [vals_cell_array{i}, 1/24];
                            %                             vals(running_ind+m) = 1/24;
                        else % n>=4
                            vals_cell_array{i} = [vals_cell_array{i}, 1/32];
                            %                             vals(running_ind+m) = 1/32;
                        end
                    end
                    
                    % assign indices and weight for other farther triangles
                    for m=1:numel(t_others)
                        j_inds_cell_array{i} = [j_inds_cell_array{i}, t_others(m)];
                        %                         j_inds(running_ind+2+m) = t_others(m);
                        switch n
                            case 4
                                vals_cell_array{i} = [vals_cell_array{i}, 1/32];
                                %                                 vals(running_ind+2+m) = 1/32;
                            case 5
                                vals_cell_array{i} = [vals_cell_array{i}, beta/8];
                                %                                 vals(running_ind+2+m) = beta/8;
                            otherwise % >=6
                                vals_cell_array{i} = [vals_cell_array{i}, 1/32];
                                %                                 vals(running_ind+2+m) = 1/32;
                        end
                    end
                    
                    % update running_ind
                    %                     running_ind = running_ind + n_cutt_off;
                    
                    % update new triangle index
                    new_triangle_ind = new_triangle_ind + 1;
                end
                
                % even rules: assign indices and valse to the middle
                % triangle contribution . this are weights from the main
                % triangle and its 3 neighboring triangles, as assuming no
                % boundaries. TODO: allow boundaries
                i_inds_cell_array{i} = [i_inds_cell_array{i}, repmat(new_triangle_ind, 1, 4)];
                %                 i_inds(running_ind:running_ind + 3) = new_triangle_ind;
                j_inds_cell_array{i} = [j_inds_cell_array{i}, [i, t_adj]];
                %                 j_inds(running_ind:running_ind + 3) = [i, t_adj];
                vals_cell_array{i} = [vals_cell_array{i}, repmat(1/16, 1, 4)];
                %                 vals(running_ind:running_ind + 3) = 1/16;
                
                %update running_ind
                %                 running_ind = running_ind + 4;
                
                % update new triangle index
                %                 new_triangle_ind = new_triangle_ind+1;
                
            end
            % construct S_2 sparse matrix
            i_inds = cell2mat(i_inds_cell_array);
            j_inds = cell2mat(j_inds_cell_array);
            vals   = cell2mat(vals_cell_array);
            S_2 = sparse(i_inds, j_inds, vals, obj.numFaces*4, obj.numFaces);
            
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % subdivide 1-forms (edges)
            %disp('S1 calculation');
            
            % edges-faces adjacency list
            
            % get old edges
            E_old = obj.edges(false);
            
            % create the new edge-list by my convention that the
            % oreintation is from smaller index to higher index, PAY
            % ATTENTION TO THE ORIENTATION COMPARED TO THE OLD EDGE WHEN
            % BUILDING THIS WAY
            E_new = unique(sort([F_new(:,[1,2]); F_new(:,[2,3]); F_new(:,[3,1])], 2), 'rows');
            
            % initiating the vectors for sparse S_2 building
            % In the Even rules for each vertex we look at all outoing
            % edges and the cut off number of the ring of edges, for each
            % edge the number of outoing vertices : sigma_i_in_V {(n_i +
            % n_i_cutoff)n_i}
            % Also in the cutoff when there is a triangle with 3 vertices,
            % gamma_0 = 0 and should not be accounted.
            % In the odd rules, each complete new edge is supported by two
            % other triangles (3 including it creator), over all 7 values
            num_adjacent_edges_per_vertex = full(sum(abs(d0)));
            num_adjacent_edges_per_vertex_cut_off = (num_adjacent_edges_per_vertex>=4)*4 + (num_adjacent_edges_per_vertex==3).*(num_adjacent_edges_per_vertex-1);
            edges_num_contrib_even_rules = sum((num_adjacent_edges_per_vertex + num_adjacent_edges_per_vertex_cut_off).*num_adjacent_edges_per_vertex);
            edges_num_contrib_odd_rules  = obj.numFaces*3*7;
            %             i_inds = zeros(edges_num_contrib_even_rules + edges_num_contrib_odd_rules,1);
            %             j_inds = zeros(edges_num_contrib_even_rules + edges_num_contrib_odd_rules,1);
            %             vals   = zeros(edges_num_contrib_even_rules + edges_num_contrib_odd_rules,1);
            i_inds_cell_array = cell(1, edges_num_contrib_even_rules + edges_num_contrib_odd_rules);
            j_inds_cell_array = cell(1, edges_num_contrib_even_rules + edges_num_contrib_odd_rules);
            vals_cell_array   = cell(1, edges_num_contrib_even_rules + edges_num_contrib_odd_rules);
            
            % run over the new edges and calculate values of Even rules
            % because the new edge list is sorted like the convention,
            % small vertex and then larger, and all new vertices were
            % inserted in V_new after the old vertices, to run over Even
            % rules it means to run on new edges that contain an old
            % vertex. And in the sorted new edge list its to run from 1 to
            % num_edges_old*2.
            %             running_ind = 1;
            parfor i = 1:obj.numEdges*2
                
                % get the edge
                e = E_new(i,:);
                
                % old vertex, this is the beggining of edge e (according to
                % the orientation(
                v_old = e(1);
                
                % new vertex
                v_new = e(2);
                
                % get father edge of this new edge
                e_father_ind = v_new-obj.numVertices; % father edge index
                
                e_father = E_old(e_father_ind,:);
                
                % calculate orientation according to the father edge
                sigma_0_orientation = (v_old == e_father(1))*1 + (v_old ~= e_father(1))*(-1);
                
                % get the father edge ind which is where the new edges points to,  according to its orientation to the
                % father end
                if sigma_0_orientation == 1
                    e_towards_v = e_father(2);
                else
                    e_towards_v = e_father(1);
                end
                
                % get all outgoing edges indices from v_old
                edges_outgoing_inds = find(abs(d0(:, v_old)) == 1);
                % get rid of the father edges index
                % edges_outgoing_inds = setdiff(edges_outgoing_inds, e_father_ind); - slower
                edges_outgoing_inds = edges_outgoing_inds(~(edges_outgoing_inds==e_father_ind));
                
                % calculate the number of outgoind edges, add 1 for father
                % edge - this is n
                n = numel(edges_outgoing_inds) + 1;
                
                % calculate alpha and beta according to n
                switch n
                    case 3
                        alpha = 3/16;
                        beta  = 1/12;
                    case 4
                        alpha = 3/(8*n);
                        beta  = 1/8;
                    case 5
                        alpha = 3/(8*n);
                        beta  = 1/4 - 1/(16*sin(2*pi/5)^2);
                    otherwise
                        alpha = 3/(8*n);
                        beta  = 1/4;
                end
                
                % update the value for the inds and vals for the sparse matrix
                i_inds_cell_array{i} = [i_inds_cell_array{i}, i];
                %                 i_inds(running_ind) = i;
                j_inds_cell_array{i} = [j_inds_cell_array{i}, e_father_ind];
                %                 j_inds(running_ind) = e_father_ind;
                vals_cell_array{i} = [vals_cell_array{i}, (3/8 -alpha -beta/4)*sigma_0_orientation];
                %                 vals(running_ind)   = (3/8 -alpha -beta/4)*sigma_0_orientation;
                %                 running_ind = running_ind + 1;
                
                % get the edge two faces (the edge we work on)
                edge_faces = find(abs(d1(:, e_father_ind)) == 1);
                
                % get the edges on this two faces
                edges_faces_all_edges = triangle_edges(edge_faces, :);
                
                % reduce father edges ind
                % edges_faces_all_edges(edges_faces_all_edges == e_father_ind) = []; - slower
                edges_faces_all_edges = sort(edges_faces_all_edges(~(edges_faces_all_edges == e_father_ind)));
                
                % get sigma_1 and sigma_n-1 indices, edges and orientation
                % sigma_1_and_n_minus_1_inds = intersect(edges_faces_all_edges, edges_outgoing_inds); - slower
                sigma_1_and_n_minus_1_inds = edges_faces_all_edges(ismembc(edges_faces_all_edges, edges_outgoing_inds));
                sigma_1_and_n_minus_1_edges = E_old(sigma_1_and_n_minus_1_inds,:);
                % orientation - +1 if it is outgoing, -1 if it is ingoing
                sigma_1_and_n_minus_1_orientation = (v_old == sigma_1_and_n_minus_1_edges(:,1))*1 + (v_old ~= sigma_1_and_n_minus_1_edges(:,1))*(-1);
                
                % get the rest of sigmas inds
                % sigma_rest_inds = setdiff(edges_outgoing_inds, sigma_1_and_n_minus_1_inds); - slower
                sigma_rest_inds = edges_outgoing_inds(~ismembc(edges_outgoing_inds, sigma_1_and_n_minus_1_inds));
                
                % get gamma_0 and gamma_n-1 indices, edges and orientation
                % gamma_0_and_n_minus_1_inds = setdiff(edges_faces_all_edges, sigma_1_and_n_minus_1_inds); -slower
                gamma_0_and_n_minus_1_inds = edges_faces_all_edges(~ismembc(edges_faces_all_edges, sigma_1_and_n_minus_1_inds));
                gamma_0_and_n_minus_1_edges = E_old(gamma_0_and_n_minus_1_inds,:);
                % orientation - +1 if the egdge end is towards sigma_0 -1
                % else
                edges_gamma_0_and_n_minus_1_orientation = (e_towards_v == gamma_0_and_n_minus_1_edges(:,2))*1 + (e_towards_v ~= gamma_0_and_n_minus_1_edges(:,2))*(-1);
                
                
                % if n==3 we are done, can set this edge sparse inds and
                % vals
                if n == 3
                    
                    % update sigma 1 and sigma n-1 in inds and vals for the
                    % sparse matrix
                    i_inds_cell_array{i} = [i_inds_cell_array{i}, [i,i]];
                    %                     i_inds(running_ind:running_ind+1) = i;
                    j_inds_cell_array{i} = [j_inds_cell_array{i}, sigma_1_and_n_minus_1_inds'];
                    %                     j_inds(running_ind:running_ind+1) = sigma_1_and_n_minus_1_inds;
                    vals_cell_array{i} = [vals_cell_array{i}, (1/8 -alpha + beta/8)*sigma_1_and_n_minus_1_orientation'];
                    %                     vals(running_ind:running_ind+1)   = (1/8 -alpha + beta/8)*sigma_1_and_n_minus_1_orientation;
                    %                     running_ind = running_ind + 2;
                    
                    % update gamma 0 and gamma n-1 in inds and vals for the
                    % sparse matrix
                    i_inds_cell_array{i} = [i_inds_cell_array{i}, [i,i]];
                    %                     i_inds(running_ind:running_ind+1) = i;
                    j_inds_cell_array{i} = [j_inds_cell_array{i}, gamma_0_and_n_minus_1_inds'];
                    %                     j_inds(running_ind:running_ind+1) = gamma_0_and_n_minus_1_inds;
                    vals_cell_array{i} = [vals_cell_array{i}, (beta/8)*edges_gamma_0_and_n_minus_1_orientation'];
                    %                     vals(running_ind:running_ind+1)   = (beta/8)*edges_gamma_0_and_n_minus_1_orientation;
                    %                     running_ind = running_ind + 2;
                    
                    % continure to the next new edge
                    continue;
                    
                else % n==4 or n>4
                    
                    % update sigma 1 and sigma n-1 in inds and vals for the
                    % sparse matrix
                    i_inds_cell_array{i} = [i_inds_cell_array{i}, [i, i]];
                    %                     i_inds(running_ind:running_ind+1) = i;
                    j_inds_cell_array{i} = [j_inds_cell_array{i}, sigma_1_and_n_minus_1_inds'];
                    %                     j_inds(running_ind:running_ind+1) = sigma_1_and_n_minus_1_inds;
                    vals_cell_array{i} = [vals_cell_array{i}, (1/8 -alpha)*sigma_1_and_n_minus_1_orientation'];
                    %                     vals(running_ind:running_ind+1)   = (1/8 -alpha)*sigma_1_and_n_minus_1_orientation;
                    %                     running_ind = running_ind + 2;
                    
                    % update gamma 0 and gamma n-1 in inds and vals for the
                    % sparse matrix
                    i_inds_cell_array{i} = [i_inds_cell_array{i}, [i, i]];
                    %                     i_inds(running_ind:running_ind+1) = i;
                    j_inds_cell_array{i} = [j_inds_cell_array{i}, gamma_0_and_n_minus_1_inds'];
                    %                     j_inds(running_ind:running_ind+1) = gamma_0_and_n_minus_1_inds;
                    vals_cell_array{i} = [vals_cell_array{i}, (beta/8)*edges_gamma_0_and_n_minus_1_orientation'];
                    %                     vals(running_ind:running_ind+1)   = (beta/8)*edges_gamma_0_and_n_minus_1_orientation;
                    %                     running_ind = running_ind + 2;
                    
                    % get the ends of sigma 1 and sigma n-1
                    % sigma_1_and_n_minus_1_ends = setdiff(sigma_1_and_n_minus_1_edges,v_old); - slower
                    sigma_1_end = sigma_1_and_n_minus_1_edges(1,~ismembc(sigma_1_and_n_minus_1_edges(1,:), v_old));
                    sigma_n_minus_1_end = sigma_1_and_n_minus_1_edges(2,~ismembc(sigma_1_and_n_minus_1_edges(2,:), v_old));
                    
                    % get sigma_2 and gamma_1 ind, edge, orientation
                    sigma_1_ind = sigma_1_and_n_minus_1_inds(1);
                    sigma_1_faces = find(abs(d1(:,sigma_1_ind)) == 1);
                    % sigma_1_2_gamma_1_face = setdiff(simga_1_faces, edge_faces); - slower
                    sigma_1_2_gamma_1_face = sigma_1_faces(~ismembc(sigma_1_faces, edge_faces));
                    sigma_1_2_gamma_1_edges_inds = triangle_edges(sigma_1_2_gamma_1_face, :);
                    % sigma_2_ind = intersect(sigma_rest_inds,sigma_1_2_gamma_1_edges_inds); - slower
                    sigma_2_ind = sigma_rest_inds(ismembc(sigma_rest_inds, sigma_1_2_gamma_1_edges_inds));
                    sigma_2_edge = E_old(sigma_2_ind,:);
                    sigma_2_orientation = (v_old == sigma_2_edge(1))*1 + (v_old ~= sigma_2_edge(1))*(-1);
                    % gamma_1_ind = setdiff(sigma_1_2_gamma_1_edges_inds, [sigma_1_ind, sigma_2_ind]); - slower
                    gamma_1_ind = sigma_1_2_gamma_1_edges_inds(~ismembc(sigma_1_2_gamma_1_edges_inds, sort([sigma_1_ind, sigma_2_ind])));
                    gamma_1_edge = E_old(gamma_1_ind,:);
                    gamma_1_orientation = (sigma_1_end == gamma_1_edge(2))*1 + (sigma_1_end ~= gamma_1_edge(2))*(-1);
                    
                    % get sigma_n-2 and gamma_n-2 ind, edge, orientation
                    sigma_n_minus_1_ind = sigma_1_and_n_minus_1_inds(2);
                    sigma_n_minus_1_faces = find(abs(d1(:,sigma_n_minus_1_ind)) == 1);
                    % sigma_n_minus_1_n_minus_2_gamma_n_minus_2_face = setdiff(simga_n_minus_1_faces, edge_faces); - slower
                    sigma_n_minus_1_n_minus_2_gamma_n_minus_2_face = sigma_n_minus_1_faces(~ismembc(sigma_n_minus_1_faces, edge_faces));
                    sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds = triangle_edges(sigma_n_minus_1_n_minus_2_gamma_n_minus_2_face, :);
                    % sigma_n_minus_2_ind = intersect(sigma_rest_inds,sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds); - slower
                    sigma_n_minus_2_ind = sigma_rest_inds(ismembc(sigma_rest_inds, sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds));
                    sigma_n_minus_2_edge = E_old(sigma_n_minus_2_ind, :);
                    sigma_n_minus_2_orientation = (v_old == sigma_n_minus_2_edge(1))*1 + (v_old ~= sigma_n_minus_2_edge(1))*(-1);
                    % gamma_n_minus_2_ind = setdiff(sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds, [sigma_n_minus_1_ind, sigma_n_minus_2_ind]); - slower
                    gamma_n_minus_2_ind = sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds(~ismembc(sigma_n_minus_1_n_minus_2_gamma_n_minus_2_edges_inds, sort([sigma_n_minus_1_ind, sigma_n_minus_2_ind])));
                    gamma_n_minus_2_edge = E_old(gamma_n_minus_2_ind,:);
                    gamma_n_minus_2_orientation = (sigma_n_minus_1_end == gamma_n_minus_2_edge(2))*1 + (sigma_n_minus_1_end ~= gamma_n_minus_2_edge(2))*(-1);
                    
                    if n==4
                        
                        % update sigma 2 in inds and vals for the sparse
                        % matrix
                        i_inds_cell_array{i} = [i_inds_cell_array{i}, i];
                        %                         i_inds(running_ind) = i;
                        j_inds_cell_array{i} = [j_inds_cell_array{i}, sigma_2_ind];
                        %                         j_inds(running_ind) = sigma_2_ind;
                        %                         vals(running_ind)   = (-alpha + beta/8)*sigma_2_orientation;
                        vals_cell_array{i} = [vals_cell_array{i}, (-alpha + beta/4)*sigma_2_orientation];
                        %                         vals(running_ind)   = (-alpha + beta/4)*sigma_2_orientation; %TODO: Check why this fix was needed
                        %                         running_ind = running_ind + 1;
                        
                        % update the gamma_1, gamma_n-2 inds and vals for
                        % the sparse matrix
                        i_inds_cell_array{i} = [i_inds_cell_array{i}, [i,i]];
                        %                         i_inds(running_ind:running_ind+1) = i;
                        j_inds_cell_array{i} = [j_inds_cell_array{i}, [gamma_1_ind, gamma_n_minus_2_ind]];
                        %                         j_inds(running_ind:running_ind+1) = [gamma_1_ind, gamma_n_minus_2_ind];
                        vals_cell_array{i} = [vals_cell_array{i}, (beta/8)*[gamma_1_orientation, gamma_n_minus_2_orientation]];
                        %                         vals(running_ind:running_ind+1)   = (beta/8)*[gamma_1_orientation, gamma_n_minus_2_orientation];
                        %                         running_ind = running_ind + 2;
                        continue
                        
                    else %n>=5
                        
                        % update the sigma_2, sigma_n-2 inds and vals for
                        % the sparse matrix
                        i_inds_cell_array{i} = [i_inds_cell_array{i}, [i,i]];
                        %                         i_inds(running_ind:running_ind+1) = i;
                        j_inds_cell_array{i} = [j_inds_cell_array{i}, [sigma_2_ind, sigma_n_minus_2_ind]];
                        %                         j_inds(running_ind:running_ind+1) = [sigma_2_ind, sigma_n_minus_2_ind];
                        vals_cell_array{i} = [vals_cell_array{i}, (-alpha + beta/8)*[sigma_2_orientation, sigma_n_minus_2_orientation]];
                        %                         vals(running_ind:running_ind+1)   = (-alpha + beta/8)*[sigma_2_orientation, sigma_n_minus_2_orientation];
                        %                         running_ind = running_ind + 2;
                        
                        % update the gamma_1, gamma_n-2 inds and vals for
                        % the sparse matrix
                        i_inds_cell_array{i} = [i_inds_cell_array{i}, [i,i]];
                        %                         i_inds(running_ind:running_ind+1) = i;
                        j_inds_cell_array{i} = [j_inds_cell_array{i}, [gamma_1_ind, gamma_n_minus_2_ind]];
                        %                         j_inds(running_ind:running_ind+1) = [gamma_1_ind, gamma_n_minus_2_ind];
                        vals_cell_array{i} = [vals_cell_array{i}, (beta/8)*[gamma_1_orientation, gamma_n_minus_2_orientation]];
                        %                         vals(running_ind:running_ind+1)   = (beta/8)*[gamma_1_orientation, gamma_n_minus_2_orientation];
                        %                         running_ind = running_ind + 2;
                    end
                    
                    if n>=6
                        % get the rest of sigmas inds, edges, orientations.
                        % rest_of_sigma_rest_inds = setdiff(sigma_rest_inds, [sigma_2_ind, sigma_n_minus_2_ind]); - slower
                        rest_of_sigma_rest_inds = sigma_rest_inds(~ismembc(sigma_rest_inds, sort([sigma_2_ind, sigma_n_minus_2_ind])));
                        num_rest_rest_sigmas = numel(rest_of_sigma_rest_inds);
                        sigma_rest_rest_edge = E_old(rest_of_sigma_rest_inds, :);
                        sigma_rest_rest_orientation = (v_old == sigma_rest_rest_edge(:,1))*1 + (v_old ~= sigma_rest_rest_edge(:,1))*-1;
                        
                        % update the rest of sigmas inds and vals for
                        % the sparse matrix
                        i_inds_cell_array{i} = [i_inds_cell_array{i}, repmat(i,1,num_rest_rest_sigmas)];
                        %                         i_inds(running_ind:running_ind+num_rest_rest_sigmas-1) = i;
                        j_inds_cell_array{i} = [j_inds_cell_array{i}, rest_of_sigma_rest_inds'];
                        %                         j_inds(running_ind:running_ind+num_rest_rest_sigmas-1) = rest_of_sigma_rest_inds;
                        vals_cell_array{i} = [vals_cell_array{i}, (-alpha)*sigma_rest_rest_orientation'];
                        %                         vals(running_ind:running_ind+num_rest_rest_sigmas-1)   = (-alpha)*sigma_rest_rest_orientation;
                        %                         running_ind = running_ind + num_rest_rest_sigmas;
                    end
                end
            end
            
            % run over the rest of the edges, which are the complete new
            % edges, this are the Odd rules
            parfor i = obj.numEdges*2+1: obj.numEdges*2+obj.numFaces*3
                
                % get the edge
                e = E_new(i,:);
                
                % get father edges of this new edge (an edge between edges)
                e_father_begin_ind = e(1) - obj.numVertices;
                e_father_end_ind = e(2) - obj.numVertices;
                
                % get the vertices of the father edges
                e_father_begin_edge = E_old(e_father_begin_ind,:);
                e_father_end_edge   = E_old(e_father_end_ind, :);
                % e_fathers_common_vert = intersect(e_father_begin_edge, e_father_end_edge); - slower
                e_fathers_common_vert = e_father_begin_edge(ismembc(e_father_begin_edge, e_father_end_edge));
                all_edges_inds_from_fathers_common_vert = find(abs(d0(:, e_fathers_common_vert)) == 1 );
                
                % get the new edge orientation
                % e_orientation_verts = setxor(e_father_begin_edge, e_father_end_edge, 'stable'); - slower
                e_orientation_verts = [e_father_begin_edge(~ismembc(e_father_begin_edge,e_father_end_edge)), e_father_end_edge(~ismembc(e_father_end_edge,e_father_begin_edge))];
                
                % get the fathers triangles
                e_father_begin_faces = find(abs(d1(:, e_father_begin_ind)) == 1);
                e_father_end_faces = find(abs(d1(:, e_father_end_ind)) == 1);
                
                % get the main face
                % e_main_face = intersect(e_father_begin_faces, e_father_end_faces); - slower
                e_main_face = e_father_begin_faces(ismembc(e_father_begin_faces, e_father_end_faces));
                e_main_face_edges_inds = triangle_edges(e_main_face,:);
                
                % get the third edge on the main face
                % e_main_parll_ind = setdiff(e_main_face_edges_inds, [e_father_begin_ind, e_father_end_ind]); - slower
                e_main_parll_ind = e_main_face_edges_inds(~ismembc(e_main_face_edges_inds, sort([e_father_begin_ind, e_father_end_ind])));
                e_main_parll_edge = E_old(e_main_parll_ind,:);
                e_main_parll_orientation = all(e_orientation_verts == e_main_parll_edge)*1 + any(e_orientation_verts ~= e_main_parll_edge)*(-1);
                
                % update the main parrallel edge in inds and vals for
                % sparse matrix
                i_inds_cell_array{i} = [i_inds_cell_array{i}, i];
                %                 i_inds(running_ind) = i;
                j_inds_cell_array{i} = [j_inds_cell_array{i}, e_main_parll_ind];
                %                 j_inds(running_ind) = e_main_parll_ind;
                vals_cell_array{i} = [vals_cell_array{i}, (6/32)*e_main_parll_orientation];
                %                 vals(running_ind)   = (6/32)*e_main_parll_orientation;
                %                 running_ind = running_ind + 1;
                
                % get the orientation of the father edges
                e_father_begin_orientation = (e_father_begin_edge(2) == e_orientation_verts(2) || e_father_begin_edge(1) == e_orientation_verts(1))*1 + (e_father_begin_edge(2) ~= e_orientation_verts(2) && e_father_begin_edge(1) ~= e_orientation_verts(1))*(-1);
                e_father_end_orientation = (e_father_end_edge(2) == e_orientation_verts(2) || e_father_end_edge(1) == e_orientation_verts(1))*1 + (e_father_end_edge(2) ~= e_orientation_verts(2) && e_father_end_edge(1) ~= e_orientation_verts(1))*(-1);
                
                % update the fathers begin and end edges in inds and vals for
                % sparse matrix
                i_inds_cell_array{i} = [i_inds_cell_array{i}, [i,i]];
                %                 i_inds(running_ind: running_ind + 1) = i;
                j_inds_cell_array{i} = [j_inds_cell_array{i}, [e_father_begin_ind, e_father_end_ind]];
                %                 j_inds(running_ind: running_ind + 1) = [e_father_begin_ind, e_father_end_ind];
                vals_cell_array{i} = [vals_cell_array{i}, (3/32)*[e_father_begin_orientation, e_father_end_orientation]];
                %                 vals(running_ind: running_ind + 1)   = (3/32)*[e_father_begin_orientation, e_father_end_orientation];
                %                 running_ind = running_ind + 2;
                
                % get the two other edges in the face of father begin
                % e_father_begin_other_face = setdiff(e_father_begin_faces, e_main_face); - slower
                e_father_begin_other_face = e_father_begin_faces(~ismembc(e_father_begin_faces, e_main_face));
                e_father_begin_other_face_edges_inds = triangle_edges(e_father_begin_other_face,:);
                % e_father_begin_other_face_edges_inds = setdiff(e_father_begin_other_face_edges_inds, e_father_begin_ind); - slower
                e_father_begin_other_face_edges_inds = e_father_begin_other_face_edges_inds(~ismembc(e_father_begin_other_face_edges_inds, e_father_begin_ind));
                % e_father_begin_upper_ind = intersect(e_father_begin_other_face_edges_inds, all_edges_inds_from_fathers_common_vert); - slower
                e_father_begin_upper_ind = e_father_begin_other_face_edges_inds(ismembc(e_father_begin_other_face_edges_inds, all_edges_inds_from_fathers_common_vert));
                e_father_begin_upper_edge = E_old(e_father_begin_upper_ind, :);
                e_father_begin_upper_orientation = (e_father_begin_upper_edge(2) == e_father_begin_edge(2) || e_father_begin_upper_edge(1) == e_father_begin_edge(1))*e_father_begin_orientation +  (e_father_begin_upper_edge(2) ~= e_father_begin_edge(2) && e_father_begin_upper_edge(1) ~= e_father_begin_edge(1))*(-e_father_begin_orientation);
                % e_father_begin_lower_ind = setdiff(e_father_begin_other_face_edges_inds, e_father_begin_upper_ind); - slower
                e_father_begin_lower_ind = e_father_begin_other_face_edges_inds(~ismembc(e_father_begin_other_face_edges_inds, e_father_begin_upper_ind));
                e_father_begin_lower_edge = E_old(e_father_begin_lower_ind, :);
                e_father_begin_lower_orientation = (e_father_begin_lower_edge(1) == e_main_parll_edge(1) || e_father_begin_lower_edge(2) == e_main_parll_edge(2))*(-e_main_parll_orientation) + (e_father_begin_lower_edge(1) ~= e_main_parll_edge(1) && e_father_begin_lower_edge(2) ~= e_main_parll_edge(2))*(e_main_parll_orientation);
                
                % update the two other edges in the face of father begin in
                % inds and vals for sparse matrix
                i_inds_cell_array{i} = [i_inds_cell_array{i}, [i,i]];
                %                 i_inds(running_ind: running_ind + 1) = i;
                j_inds_cell_array{i} = [j_inds_cell_array{i}, [e_father_begin_upper_ind, e_father_begin_lower_ind]];
                %                 j_inds(running_ind: running_ind + 1) = [e_father_begin_upper_ind, e_father_begin_lower_ind];
                vals_cell_array{i} = [vals_cell_array{i}, [(3/32)*e_father_begin_upper_orientation, (1/32)*e_father_begin_lower_orientation]];
                %                 vals(running_ind: running_ind + 1)   = [(3/32)*e_father_begin_upper_orientation, (1/32)*e_father_begin_lower_orientation];
                %                 running_ind = running_ind + 2;
                
                % get the two other edges in the face of father end
                % e_father_end_other_face = setdiff(e_father_end_faces, e_main_face); - slower
                e_father_end_other_face = e_father_end_faces(~ismembc(e_father_end_faces, e_main_face));
                e_father_end_other_face_edges_inds = triangle_edges(e_father_end_other_face,:);
                % e_father_end_other_face_edges_inds = setdiff(e_father_end_other_face_edges_inds, e_father_end_ind); - slower
                e_father_end_other_face_edges_inds = e_father_end_other_face_edges_inds(~ismembc(e_father_end_other_face_edges_inds, e_father_end_ind));
                % e_father_end_upper_ind = intersect(e_father_end_other_face_edges_inds, all_edges_inds_from_fathers_common_vert); - slower
                e_father_end_upper_ind = e_father_end_other_face_edges_inds(ismembc(e_father_end_other_face_edges_inds, all_edges_inds_from_fathers_common_vert));
                e_father_end_upper_edge = E_old(e_father_end_upper_ind, :);
                e_father_end_upper_orientation = (e_father_end_upper_edge(2) == e_father_end_edge(2) || e_father_end_upper_edge(1) == e_father_end_edge(1))*e_father_end_orientation +  (e_father_end_upper_edge(2) ~= e_father_end_edge(2) && e_father_end_upper_edge(1) ~= e_father_end_edge(1))*(-e_father_end_orientation);
                % e_father_end_lower_ind = setdiff(e_father_end_other_face_edges_inds, e_father_end_upper_ind); - slower
                e_father_end_lower_ind = e_father_end_other_face_edges_inds(~ismembc(e_father_end_other_face_edges_inds, e_father_end_upper_ind));
                e_father_end_lower_edge = E_old(e_father_end_lower_ind, :);
                e_father_end_lower_orientation = (e_father_end_lower_edge(1) == e_main_parll_edge(1) || e_father_end_lower_edge(2) == e_main_parll_edge(2))*(-e_main_parll_orientation) + (e_father_end_lower_edge(1) ~= e_main_parll_edge(1) && e_father_end_lower_edge(2) ~= e_main_parll_edge(2))*(e_main_parll_orientation);
                
                % update the two other edges in the face of father end in
                % inds and vals for sparse matrix
                i_inds_cell_array{i} = [i_inds_cell_array{i}, [i,i]];
                %                 i_inds(running_ind: running_ind + 1) = i;
                j_inds_cell_array{i} = [j_inds_cell_array{i}, [e_father_end_upper_ind, e_father_end_lower_ind]];
                %                 j_inds(running_ind: running_ind + 1) = [e_father_end_upper_ind, e_father_end_lower_ind];
                vals_cell_array{i} = [vals_cell_array{i}, [(3/32)*e_father_end_upper_orientation, (1/32)*e_father_end_lower_orientation]];
                %                 vals(running_ind: running_ind + 1)   = [(3/32)*e_father_end_upper_orientation, (1/32)*e_father_end_lower_orientation];
                %                 running_ind = running_ind + 2;
                
            end
            
            % construct S_1 sparse matrix
            i_inds = cell2mat(i_inds_cell_array);
            j_inds = cell2mat(j_inds_cell_array);
            vals   = cell2mat(vals_cell_array);
            S_1 = sparse(i_inds, j_inds, vals, obj.numEdges*2 + 3*obj.numFaces, obj.numEdges);
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % return the new mesh
            subdividedMesh = myMeshDEC(P_new, F_new);
        end
        function [subdividedMesh, S_0] = subdivideMesh(obj) % Subdivide the mesh
            % subdivideMesh  Subdivide the mesh with Loop rules for 0-forms
            % Output:
            %   1. subdivdedMesh - the subdivided mesh
            
            % TODO: COMPLETE THE IMPLEMENTATION FOR SURFACE WITH BOUNDARIES
            assert(isempty(obj.boundaryEdges), 'The implementation does not deal yet with surface with a boundary');
            
            % Subdvide the vertices (0-form Loop scheme):
            
            % Even vertices (the old vertices new location)
            n = full(sum(obj.adjacencyMatrix,2)); % number of neighbours for each vertex
            alpha = (n>3).*(3./(8*n)) + (n==3)*(3/16); % alpha
            % fast version
            S_0_even = sparse(1:obj.numVertices, 1:obj.numVertices, alpha, obj.numVertices, obj.numVertices)*obj.adjacencyMatrix;
            S_0_even = S_0_even + sparse(1:obj.numVertices, 1:obj.numVertices, (1-alpha.*n), obj.numVertices, obj.numVertices);
            % slow version
            % S_0_even = obj.adjacencyMatrix.*alpha;
            % S_0_even = S_0_even + speye(obj.numVertices).*(1-alpha.*n);
            
            
            % Odd vertices (new vertices indtroduced by the edges)
            % say the edge is of vertices j,k and on the 2 triangles F1, F2
            % that hold the edge there are two more vertices i,l
            % vertices on the edge: j,k
            d0 = obj.d0_matrix;
            d1 = obj.d1_matrix;
            S_0_odd_edge_verts = abs(d0)*(3/8);
            % 2 triangles of the edge: F1, F2
            [row, ~] = find(abs(d1) == 1);
            % all vertices of the 2 triangles of the edge: i,j(x2),k(x2),l
            all_verts_around_edge = reshape(obj.F(row,:)',6,[])';
            % getting vertices i,j, using: https://stackoverflow.com/questions/32494331/setdiff-row-by-row-without-using-loops-in-matlab
            mask = all(bsxfun(@ne,all_verts_around_edge,permute(obj.E,[1 3 2])),3);
            At = all_verts_around_edge';
            oppsoite_vertices = reshape(At(mask.'),[],size(all_verts_around_edge,1));
            % insert the oppsosite vertices with the right weight
            S_0_odd_opposite_verts = sparse(reshape([1:obj.numEdges;1:obj.numEdges],[],1),oppsoite_vertices(:),1/8,obj.numEdges,obj.numVertices);
            % build S_odd
            S_0_odd = S_0_odd_edge_verts + S_0_odd_opposite_verts;
            
            % S is the combination of S_even (old vertices new wieghts) and
            % S_odd (the new vertices weights)
            S_0 = [S_0_even; S_0_odd];
            
            % create the new vertices positions
            P_new = S_0*obj.P;
            
            % edges in each triangle
            [triangle_edges, ~] = find(abs(d1') == 1);
            triangle_edges = reshape(triangle_edges,3,[])';
            
            % break triangles and create new ones
            % each triangle gives 4 new ones - not dealing yet with
            % surfaces with boundaries
            F_new = zeros(obj.numFaces*4,3);
            new_triangle_ind = 1;
            for i=1:obj.numFaces
                
                % get the edges of the triangle
                t_edges = triangle_edges(i,:);
                
                % get the list of vertices in the face's three edges
                vertices_in_t_edges = obj.E(t_edges(:),:);
                
                % build the 3 new triangles that include the old vertices
                % that just changed (even) location and two new edge ones
                % (odd). There are numEdges new vertices in the subdivided
                % mesh and they are sorted according to obj.E list.
                % obj.E(1)->the first new vertex etc. Meaning that in the new
                % vertex list the first numVertices vertices are the old
                % ones (even) that changed location (will show itself in P
                % list of xyz). The new vertices will count from
                % numVertices+1 to numVertices + numEdges, since each edge
                % is giving a new vertex.
                for j=1:3
                    
                    % get the vertex
                    v = obj.F(i,j);
                    % update as the first vertex in new F
                    F_new(new_triangle_ind,1) = v;
                    % find which of the edges in the triangle contain v
                    [v_in_t_edges, ~] = find(vertices_in_t_edges == v);
                    vertices_in_t_edges_v = vertices_in_t_edges(v_in_t_edges, :);
                    % get the vertices with the face orientation
                    edge1 = obj.F(i, [j, mod(j+1-1, 3)+1]);
                    % check if the edge we are about to put in place 1
                    % is in the correct way else flip the order. THIS WAY
                    % KEEPING THE TRIANGLE ORIENTATION OF THE COARSE MESH
                    if ~isequal(vertices_in_t_edges_v(1, :), sort(edge1))
                        v_in_t_edges = [v_in_t_edges(2), v_in_t_edges(1)];
                    end
                    % get back the edges indices + the number of vertices
                    new_inds_to_add = t_edges(v_in_t_edges) + obj.numVertices;
                    % adding to the triangle
                    F_new(new_triangle_ind,2:3) = new_inds_to_add ;
                    % update new triangle index
                    new_triangle_ind = new_triangle_ind + 1;
                end
                
                % build the 4th triangle out of the three new vertices from the edges (odd)
                % constructing from previous 3 triangles constructions,
                % since there we chose the right hand rule preserving new
                % edges. because this is the central triangle we need
                % exatcly the opposite set of vertices in order to keep the
                % right hand rule.
                F_new_4th_first_edge = F_new(new_triangle_ind-3, 2:3);
                F_new_4th_second_edge_sorted = sort(F_new(new_triangle_ind-2, 2:3));
                F_new_4th = F_new_4th_second_edge_sorted(~ismembc(F_new_4th_second_edge_sorted, sort(F_new_4th_first_edge)));
                F_new(new_triangle_ind, :) = [F_new_4th_first_edge(1), F_new_4th, F_new_4th_first_edge(2)];
                
                % update new triangle index
                new_triangle_ind = new_triangle_ind+1;
                
            end
            
            % return the new mesh
            subdividedMesh = myMeshDEC(P_new, F_new);
        end
        function [subdividedMesh, S_0] = subdivideMeshLoop_parfor_S0(obj) % Loop subdiviosn
            % subdivideLoop  Perform Loop subdivision for the triangular
            % mesh.
            % Following notations from: "Subdivision Exterior Calculus for
            % Geometry Processing Supplemental Material: Subdivision Rules"
            % Warning: coded not efficiently.
            %
            % Output:
            %   1. subdivdedMesh - the subdivided mesh
            %   2. S_0 - Subdivision matrix for 0-forms
            %   3. S_1 - Subdivision matrix for 1-forms
            %   4. S_2 - Subdivision matrix for 2-forms
            
            % TODO: COMPLETE THE IMPLEMENTATION FOR SURFACE WITH BOUNDARIES
            assert(isempty(obj.boundaryEdges), 'The implementation does not deal yet with surface with a boundary');
            
            % Assert the input is a manifold
            assert(size(obj.edges(1),1)/2 == size(obj.edges(0),1), 'The input is non-manifold');
            
            % Subdvide the vertices (0-form Loop scheme):
            %disp('S0 calculation');
            
            % Even vertices (the old vertices new location)
            n = full(sum(obj.adjacencyMatrix,2)); % number of neighbours for each vertex
            alpha = (n>3).*(3./(8*n)) + (n==3)*(3/16); % alpha
            % fast version
            S_0_even = sparse(1:obj.numVertices, 1:obj.numVertices, alpha, obj.numVertices, obj.numVertices)*obj.adjacencyMatrix;
            S_0_even = S_0_even + sparse(1:obj.numVertices, 1:obj.numVertices, (1-alpha.*n), obj.numVertices, obj.numVertices);
            % slow version
            % S_0_even = obj.adjacencyMatrix.*alpha;
            % S_0_even = S_0_even + speye(obj.numVertices).*(1-alpha.*n);
            
            % Odd vertices (new vertices indtroduced by the edges)
            % say the edge is of vertices j,k and on the 2 triangles F1, F2
            % that hold the edge there are two more vertices i,l
            % vertices on the edge: j,k
            d0 = obj.d0_matrix;
            d1 = obj.d1_matrix;
            S_0_odd_edge_verts = abs(d0)*(3/8);
            % 2 triangles of the edge: F1, F2
            [row, ~] = find(abs(d1) == 1);
            % all vertices of the 2 triangles of the edge: i,j(x2),k(x2),l
            all_verts_around_edge = reshape(obj.F(row,:)',6,[])';
            % getting vertices i,j, using: https://stackoverflow.com/questions/32494331/setdiff-row-by-row-without-using-loops-in-matlab
            mask = all(bsxfun(@ne,all_verts_around_edge,permute(obj.E,[1 3 2])),3);
            At = all_verts_around_edge';
            oppsoite_vertices = reshape(At(mask.'),[],size(all_verts_around_edge,1));
            % insert the oppsosite vertices with the right weight
            S_0_odd_opposite_verts = sparse(reshape([1:obj.numEdges;1:obj.numEdges],[],1),oppsoite_vertices(:),1/8,obj.numEdges,obj.numVertices);
            % build S_odd
            S_0_odd = S_0_odd_edge_verts + S_0_odd_opposite_verts;
            
            % S is the combination of S_even (old vertices new wieghts) and
            % S_odd (the new vertices weights)
            S_0 = [S_0_even; S_0_odd];
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % New Triangulation of subdivded mesh
            %disp('New triangulation');
            
            % create the new vertices positions
            P_new = S_0*obj.P;
            
            % edges in each triangle
            [triangle_edges, ~] = find(abs(d1') == 1);
            triangle_edges = reshape(triangle_edges,3,[])';
            
            % break triangles and create new ones
            % each triangle gives 4 new ones - not dealing yet with
            % surfaces with boundaries
            %             F_new = zeros(obj.numFaces*4,3);
            
            F_new_cell_array = cell(obj.numFaces, 1);
            vertices_in_t_edges_all_faces = mat2cell(obj.E(reshape(triangle_edges',[],1), :), repmat(3, 1, obj.numFaces));
            
            %             new_triangle_ind = 1;
            parfor i=1:obj.numFaces
                
                % calculate new triangle ind for this face
                %                 new_triangle_ind = (i-1)*4+1;
                
                % initialize new 4 faces
                new_4_faces = nan(4,3);
                new_face_ind = 1;
                
                % get the edges of the triangle
                t_edges = triangle_edges(i,:);
                
                % get the list of vertices in the face's three edges
                %                 vertices_in_t_edges = obj.E(t_edges(:),:);
                vertices_in_t_edges = vertices_in_t_edges_all_faces{i};
                
                
                % build the 3 new triangles that include the old vertices
                % that just changed (even) location and two new edge ones
                % (odd). There are numEdges new vertices in the subdivided
                % mesh and they are sorted according to obj.E list.
                % obj.E(1)->the first new vertex etc. Meaning that in the new
                % vertex list the first numVertices vertices are the old
                % ones (even) that changed location (will show itself in P
                % list of xyz). The new vertices will count from
                % numVertices+1 to numVertices + numEdges, since each edge
                % is giving a new vertex.
                for j=1:3
                    
                    % get the vertex
                    v = obj.F(i,j);
                    % update as the first vertex in new F
                    new_4_faces(new_face_ind, 1) = v;
                    %                     F_new(new_triangle_ind,1) = v;
                    % find which of the edges in the triangle contain v
                    [v_in_t_edges, ~] = find(vertices_in_t_edges == v);
                    vertices_in_t_edges_v = vertices_in_t_edges(v_in_t_edges, :);
                    % get the vertices with the face orientation
                    edge1 = obj.F(i, [j, mod(j+1-1, 3)+1]);
                    % check if the edge we are about to put in place 1
                    % is in the correct way else flip the order. THIS WAY
                    % KEEPING THE TRIANGLE ORIENTATION OF THE COARSE MESH
                    if ~isequal(vertices_in_t_edges_v(1, :), sort(edge1))
                        v_in_t_edges = [v_in_t_edges(2), v_in_t_edges(1)];
                    end
                    % get back the edges indices + the number of vertices
                    new_inds_to_add = t_edges(v_in_t_edges) + obj.numVertices;
                    % adding to the triangle
                    new_4_faces(new_face_ind, 2:3) = new_inds_to_add;
                    %                     F_new(new_triangle_ind,2:3) = new_inds_to_add ;
                    % update new triangle index
                    new_face_ind = new_face_ind + 1;
                    %                     new_triangle_ind = new_triangle_ind + 1;
                end
                
                % build the 4th triangle out of the three new vertices from the edges (odd)
                % constructing from previous 3 triangles constructions,
                % since there we chose the right hand rule preserving new
                % edges. because this is the central triangle we need
                % exatcly the opposite set of vertices in order to keep the
                % right hand rule.
                F_new_4th_first_edge = new_4_faces(new_face_ind-3, 2:3);
                F_new_4th_second_edge_sorted = sort(new_4_faces(new_face_ind-2, 2:3));
                %                 F_new_4th_first_edge = F_new(new_triangle_ind-3, 2:3);
                %                 F_new_4th_second_edge_sorted = sort(F_new(new_triangle_ind-2, 2:3));
                F_new_4th = F_new_4th_second_edge_sorted(~ismembc(F_new_4th_second_edge_sorted, sort(F_new_4th_first_edge)));
                new_4_faces(new_face_ind, :) = [F_new_4th_first_edge(1), F_new_4th, F_new_4th_first_edge(2)];
                %                 F_new(new_triangle_ind, :) = [F_new_4th_first_edge(1), F_new_4th, F_new_4th_first_edge(2)];
                
                % update new triangle index
                %                 new_triangle_ind = new_triangle_ind+1;
                
                % update the 4 new faces
                F_new_cell_array{i} = new_4_faces;
                
            end
            F_new = cell2mat(F_new_cell_array);
            
            
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % return the new mesh
            subdividedMesh = myMeshDEC(P_new, F_new);
        end
        function [subdividedMesh, S_0] = subdivideMesh_trivial(obj) % Subdivide the mesh
            % subdivideMesh  Subdivide the mesh with trivial rules for 0-forms
            % Output:
            %   1. subdivdedMesh - the subdivided mesh
            
            % TODO: COMPLETE THE IMPLEMENTATION FOR SURFACE WITH BOUNDARIES
            assert(isempty(obj.boundaryEdges), 'The implementation does not deal yet with surface with a boundary');
            
            % Subdvide the vertices (0-form Loop scheme):
            
            % Even vertices (the old vertices new location)
            S_0_even = sparse(1:obj.numVertices, 1:obj.numVertices, 1, obj.numVertices, obj.numVertices);
            
            % Odd vertices (new vertices indtroduced by the edges)
            % say the edge is of vertices j,k and on the 2 triangles F1, F2
            % that hold the edge there are two more vertices i,l
            % vertices on the edge: j,k
            d0 = obj.d0_matrix;
            d1 = obj.d1_matrix;
            S_0_odd = abs(d0)*(1/2);
            % S is the combination of S_even (old vertices new wieghts) and
            % S_odd (the new vertices weights)
            S_0 = [S_0_even; S_0_odd];
            
            % create the new vertices positions
            P_new = S_0*obj.P;
            
            % edges in each triangle
            [triangle_edges, ~] = find(abs(d1') == 1);
            triangle_edges = reshape(triangle_edges,3,[])';
            
            % break triangles and create new ones
            % each triangle gives 4 new ones - not dealing yet with
            % surfaces with boundaries
            F_new = zeros(obj.numFaces*4,3);
            new_triangle_ind = 1;
            for i=1:obj.numFaces
                
                % get the edges of the triangle
                t_edges = triangle_edges(i,:);
                
                % get the list of vertices in the face's three edges
                vertices_in_t_edges = obj.E(t_edges(:),:);
                
                % build the 3 new triangles that include the old vertices
                % that just changed (even) location and two new edge ones
                % (odd). There are numEdges new vertices in the subdivided
                % mesh and they are sorted according to obj.E list.
                % obj.E(1)->the first new vertex etc. Meaning that in the new
                % vertex list the first numVertices vertices are the old
                % ones (even) that changed location (will show itself in P
                % list of xyz). The new vertices will count from
                % numVertices+1 to numVertices + numEdges, since each edge
                % is giving a new vertex.
                for j=1:3
                    
                    % get the vertex
                    v = obj.F(i,j);
                    % update as the first vertex in new F
                    F_new(new_triangle_ind,1) = v;
                    % find which of the edges in the triangle contain v
                    [v_in_t_edges, ~] = find(vertices_in_t_edges == v);
                    vertices_in_t_edges_v = vertices_in_t_edges(v_in_t_edges, :);
                    % get the vertices with the face orientation
                    edge1 = obj.F(i, [j, mod(j+1-1, 3)+1]);
                    % check if the edge we are about to put in place 1
                    % is in the correct way else flip the order. THIS WAY
                    % KEEPING THE TRIANGLE ORIENTATION OF THE COARSE MESH
                    if ~isequal(vertices_in_t_edges_v(1, :), sort(edge1))
                        v_in_t_edges = [v_in_t_edges(2), v_in_t_edges(1)];
                    end
                    % get back the edges indices + the number of vertices
                    new_inds_to_add = t_edges(v_in_t_edges) + obj.numVertices;
                    % adding to the triangle
                    F_new(new_triangle_ind,2:3) = new_inds_to_add ;
                    % update new triangle index
                    new_triangle_ind = new_triangle_ind + 1;
                end
                
                % build the 4th triangle out of the three new vertices from the edges (odd)
                % constructing from previous 3 triangles constructions,
                % since there we chose the right hand rule preserving new
                % edges. because this is the central triangle we need
                % exatcly the opposite set of vertices in order to keep the
                % right hand rule.
                F_new_4th_first_edge = F_new(new_triangle_ind-3, 2:3);
                F_new_4th_second_edge_sorted = sort(F_new(new_triangle_ind-2, 2:3));
                F_new_4th = F_new_4th_second_edge_sorted(~ismembc(F_new_4th_second_edge_sorted, sort(F_new_4th_first_edge)));
                F_new(new_triangle_ind, :) = [F_new_4th_first_edge(1), F_new_4th, F_new_4th_first_edge(2)];
                
                % update new triangle index
                new_triangle_ind = new_triangle_ind+1;
                
            end
            
            % return the new mesh
            subdividedMesh = myMeshDEC(P_new, F_new);
        end
        function phi = heatDistance(obj, source_inds) % compute heat distance
            % heatDistance  Computing heat distance
            % This method computes the heat distance according to Crane et
            % al. (2013) "The heat method for distance computation".
            %
            % Input:
            %   1. sourceInds - the source points
            %   2. W - the conformal laplacian (cotan weight, d0^T*star1*d0 in DEC)
            %   3. M - the mass matrix for vertics (star0 in DEC)
            %   4. delta - star0^-1 * d0^T * star1 from DEC
            %   5. t - the time parameter for integration (if t=0 take the
            %          default = h^2, where h = mean edge length)
            %
            % Output:
            %   1. phi - the heat distance to each point
            %
            % Algorithm steps:
            %   0. If t=0, choose t = h^2
            %   1. Integreate heat flow for a fixed time t by:
            %      solving (M - tW)u = u0
            %      where u0 is indicator function on the source points
            %   2. Compute the normalized vector field X = -grad(u)/||grad(u)||
            %   3. Solve the Poisson equation L*phi = div*X
            
            % DEC operators
            d0 = obj.d0_matrix;
            star1 = obj.star1_matrix;
            star0 = obj.star0_matrix;
            
            % Operators
            W = d0'*star1*d0;
            M = star0;
            delta = spdiags(1./diag(star0), 0, size(star0,1), size(star0,2)) * d0' * star1;
            
            % check if t is given else t = h^2
            t = max(obj.edgesLength)^2;
            
            % prepare u0
            u0 = sparse(source_inds, ones(numel(source_inds),1), ones(numel(source_inds),1), obj.numVertices, 1);
            
            % Integrate heat flow for time t
            u = (M+t*W) \ u0;
            u = full(u);
            
            % Compute the normalized vector field X
            X = reshape(obj.grad_matrix*u, [] ,3);
            X = -X./sqrt(sum(X.^2,2));
            
            % Prepre div*X (integrated) leaving de-Rham complex
            % d2 = -obj.div_matrix_integrated*X(:);
            
            % Prepare div*X in de-Rham complex
            % Turn X back to 1-forms on the edges by:
            % c1 = <e1,X1>/2 + <e1,X2>/2 = <e1, X1 + X2>/2
            % where X1 and X2 are to two vectors
            % from X that are on the adjacent faces to e1.
            % Remebering that my convention is that the orientation of the
            % edge is from low ind to high ind;
            e = obj.P(obj.E(:,2),:) - obj.P(obj.E(:,1),:);
            X1_plus_X2 = abs(obj.d1_matrix')*X;
            c = dot(e,X1_plus_X2,2)/2;
            % div using DEC delta
            d = delta*c;
            
            % make rhs with zero mean
            d = d-mean(d);
            
            % Solve the Poisson equation
            phi = W \ d;
            
            % since the solution to is determined only up to an additive
            % constant, final values are shifted such that the smallest
            % distance is zero
            shift = min(phi);
            phi = phi - shift;
            
        end
        function [M0, M1, M2, A_0, A_1, A_2, shape_subdiv, M0_full, S_0_cell, S_1_cell, shape_subdiv_cell] = SDEC_operators(obj, subdivision_levels) % compute SDEC operators
            % SDEC_operators  compute SDEC operators to l = subdivision_levels
            
            % copy for subdivision
            shape_subdiv = copy(obj);
            
            % initialize accumulation matrices
            A_0 = speye(obj.numVertices);
            A_1 = speye(obj.numEdges);
            A_2 = speye(obj.numFaces);
            S_0_cell = cell(1,subdivision_levels);
            S_1_cell = cell(1,subdivision_levels);
            shape_subdiv_cell = cell(1,subdivision_levels);
            
            % subdivide
            for l=1:subdivision_levels
                
                % subdivide and get SDEC matrices
                disp(['subdividing to level ', num2str(l), '/', num2str(subdivision_levels)]);
                %[shape_subdiv, S_0, S_1, S_2] = shape_subdiv.subdivideMeshLoop;
                [shape_subdiv, S_0, S_1, S_2] = shape_subdiv.subdivideMeshLoop_parfor;
                
                % update accumulation matrices
                A_0 = S_0*A_0;
                A_1 = S_1*A_1;
                A_2 = S_2*A_2;
                
                % update S_0 cell array and shape_subdiv cell array
                S_0_cell{l} = S_0;
                S_1_cell{l} = S_1;
                shape_subdiv_cell{l} = shape_subdiv;
                
            end
            
            % comptute SDEC operators
            M0 = A_0'*shape_subdiv.star0_matrix*A_0;
            M0_full = A_0'*shape_subdiv.star0_matrix_FEM*A_0;
            M1 = A_1'*shape_subdiv.star1_matrix*A_1;
            M2 = A_2'*shape_subdiv.star2_matrix*A_2;
            
            % assert comutation
            assert(norm(shape_subdiv.d0_matrix*A_0 - A_1*obj.d0_matrix, 'fro') < 1e-12, 'SEC Commutation Error');
            assert(norm(shape_subdiv.d1_matrix*A_1 - A_2*obj.d1_matrix, 'fro') < 1e-12, 'SEC Commutation Error');
            
        end
        function [S_0_cell, shape_subdiv_cell] = SDEC_operators_S0(obj, subdivision_levels) % compute SDEC operators
            % SDEC_operators  compute SDEC operators to l = subdivision_levels
            
            % copy for subdivision
            shape_subdiv = copy(obj);
            
            % initialize accumulation matrices
            S_0_cell = cell(1,subdivision_levels);            
            shape_subdiv_cell = cell(1,subdivision_levels);
            
            % subdivide
            for l=1:subdivision_levels
                
                % subdivide and get SDEC matrices
                disp(['subdividing to level ', num2str(l), '/', num2str(subdivision_levels)]);
                [shape_subdiv, S_0] = shape_subdiv.subdivideMeshLoop_parfor_S0;
                
                % update S_0 cell array and shape_subdiv cell array
                S_0_cell{l} = S_0;
                shape_subdiv_cell{l} = shape_subdiv;
                
            end
            
        end
        function [M0, M1, M2, A_0, A_1, A_2, shape_subdiv, M0_full] = SDEC_operators_trivial(obj, subdivision_levels) % compute SDEC operators
            % SDEC_operators_trivial  compute SDEC operators to l = subdivision_levels
            % The mass taken from a trivial subdivision scheme and not loop
            
            % copy for subdivision
            shape_subdiv = copy(obj);
            
            % initialize accumulation matrices
            A_0 = speye(obj.numVertices);
            A_1 = speye(obj.numEdges);
            A_2 = speye(obj.numFaces);
            
            % subdivide
            for l=1:subdivision_levels
                
                % subdivide and get SDEC matrices
                disp(['subdividing to level ', num2str(l), '/', num2str(subdivision_levels)]);
                %[shape_subdiv, S_0, S_1, S_2] = shape_subdiv.subdivideMeshLoop;
                [~, S_0, S_1, S_2] = shape_subdiv.subdivideMeshLoop_parfor;
                shape_subdiv = shape_subdiv.subdivideMesh_trivial;
                
                % update accumulation matrices
                A_0 = S_0*A_0;
                A_1 = S_1*A_1;
                A_2 = S_2*A_2;
                
            end
            
            % comptute SDEC operators
            M0 = A_0'*shape_subdiv.star0_matrix*A_0;
            M0_full = A_0'*shape_subdiv.star0_matrix_FEM*A_0;
            M1 = A_1'*shape_subdiv.star1_matrix*A_1;
            M2 = A_2'*shape_subdiv.star2_matrix*A_2;
            
            % assert comutation
            assert(norm(shape_subdiv.d0_matrix*A_0 - A_1*obj.d0_matrix, 'fro') < 1e-13, 'SEC Commutation Error');
            assert(norm(shape_subdiv.d1_matrix*A_1 - A_2*obj.d1_matrix, 'fro') < 1e-13, 'SEC Commutation Error');
            
        end
        function basis = lbo_basis_DEC(obj, num_eigen) % compute DEC lbo eigen basis
            % lbo_basis_DEC  Compute the num_eigen sub basis of the LBO
            
            % get DEC operators
            d0 = obj.d0_matrix;
            star0 = obj.star0_matrix;
            star1 = obj.star1_matrix;
            W_DEC = d0'*star1*d0;
            
            % compute DEC eigenfunctions
            [V_DEC, D_DEC] = eigs(W_DEC, star0, num_eigen, -1e-5, 'IsSymmetricDefinite', 1, 'Display', 0);
            
            % verify A*V = B*V*D, calculate frobenius norm
            solution_norm = norm(W_DEC*V_DEC - star0*V_DEC*D_DEC, 'fro');
            %assert(solution_norm < 1e-5, 'Error: A*V != B*V*D');
            
            % verify othornormality with respect to mass matrix
            assert(norm(V_DEC'*star0*V_DEC - speye(num_eigen)) < 1e-11, 'Error: not orthonormal with respect to mass');
            
            % sort ascend
            [vals, inds] = sort(diag(D_DEC));
            V_DEC = V_DEC(:, inds);
            
            % check if there are complex eigen values
            assert(all(imag(vals) == 0), 'Error: complex eigenvalues');
            
            % return the lbo eigen functions/values, W and mass matrix
            % return in struct
            basis.eigenfunctions = V_DEC;
            basis.eigenvalues = vals;
            basis.W = W_DEC;
            basis.A = star0;
            basis.A_full = obj.star0_matrix_FEM; % not diagonal
            
            
        end
        function [basis, shape_subdiv] = lbo_basis_SDEC(obj, num_eigen, subdivision_levels, flag_full_M0) % compute SDEC lbo eigen basis
            % lbo_basis_SDEC  Compute the num_eigen sub basis of the LBO
            
            % get SDEC operators
            [M0, M1, ~, A_0, ~, ~, shape_subdiv, M0_full] = obj.SDEC_operators(subdivision_levels);
            if flag_full_M0 == 1 % use full (not lumped) M0 of the subdivided mesh
                M0 = M0_full;
            end
            
            % comptute SDEC LBO operator
            W_SDEC = obj.d0_matrix'*M1*obj.d0_matrix;
            
            % compute SDEC eigenfunctions
            disp('Compute SDEC eigenfucntions');
            [V_SDEC, D_SDEC] = eigs((W_SDEC+W_SDEC')/2, (M0 + M0')/2, num_eigen, -1e-5, 'IsSymmetricDefinite', 1);
            
            % verify A*V = B*V*D, calculate frobenius norm
            solution_norm = norm(W_SDEC*V_SDEC - M0*V_SDEC*D_SDEC, 'fro');
            %             assert(solution_norm < sqrt(1e-11), 'Error: A*V != B*V*D');
            
            % verify othornormality with respect to mass matrix
            assert(norm(V_SDEC'*(M0 + M0')/2*V_SDEC - speye(num_eigen) ,'fro') < 1e-11, 'Error: not orthonormal with respect to mass');
            
            % sort ascend
            [vals, inds] = sort(diag(D_SDEC));
            V_SDEC = V_SDEC(:, inds);
            
            % check if there are complex eigen values
            assert(all(imag(vals) == 0), 'Error: complex eigenvalues');
            
            % return the lbo eigen functions/values, W, mass matrix,
            % subdivision operator for 0-forms (vertices)
            % return in struct
            basis.eigenfunctions = V_SDEC;
            basis.eigenvalues = vals;
            basis.W = (W_SDEC+W_SDEC')/2;
            basis.A = (M0 + M0')/2;
            basis.A_0 = A_0;
            
            
        end
        function basis_hierarchical = lbo_basis_hierarchical_SDEC(obj, num_eigen_vec, subdivision_levels) % compute SDEC lbo eigen basis hierarchical
            % lbo_basis_hierarchical_SDEC  Compute the num_eigen_vec sub basis of the LBO
            
            % check required dimensions
            assert(numel(num_eigen_vec) == subdivision_levels+1, 'Err number of subdivision levels is not equal to size of eigen vecs required');
            
            % initialize
            basis_hierarchical = cell(subdivision_levels+1, 1);
            
            
            %% get SDEC operators
            tic;
            [~, ~, ~, ~, ~, ~, shape_subdiv_finest, ~, S_0_cell, S_1_cell, shape_subdiv_cell] = obj.SDEC_operators(subdivision_levels);
            toc;
            
            %% Get M0 and M1 of the hierarchy
            % initiate cell arrays to contain all M0 and M1 in the
            % hierarchy.
            M0_cell = cell(1, subdivision_levels+1);
            M1_cell = cell(1, subdivision_levels+1);
            % M0 and M1 SEC of the finest level
            M0_cell{subdivision_levels+1} = shape_subdiv_finest.star0_matrix;
            M1_cell{subdivision_levels+1} = shape_subdiv_finest.star1_matrix;
            % run over levels and build M0 and M1. From fine to coarse M0 =
            % S' * ... * S' * M0_DEC_fine * S *...*S
            for i = subdivision_levels:-1:1
                
                % get current level S_0 and S_1
                S_0 = S_0_cell{i};
                S_1 = S_1_cell{i};
                
                % update M0 to this level
                M0_cell{i} = S_0'*M0_cell{i+1}*S_0;
                M1_cell{i} = S_1'*M1_cell{i+1}*S_1;
                
            end
            
            
            %% compute basis for each hierarchy level - starting from coarset level
            tic;
            for i = 1:subdivision_levels+1
                
                fprintf('Compute SEC basis for sbd level %d/%d', i-1, subdivision_levels);
                
                % get current shape
                if i == 1 % coarset shape
                    shape = obj;
                    num_eigen = num_eigen_vec(i);
                    sigma = -1e-5;
                    fprintf(' %d eigens \n', num_eigen);
                else % subdivided shape
                    shape = shape_subdiv_cell{i-1};
                    num_eigen = num_eigen_vec(i)*2;
                    last_largest_eig = basis_hierarchical{i-1}.eigenvalues(end);
                    sigma = last_largest_eig;
                    fprintf(', %d eigens, will use about %d eigens \n', num_eigen, num_eigen_vec(i));
                end
                
                
                % get current operators and matrices
                M0 = M0_cell{i};
                M0 = (M0 + M0')/2; %verify symmetry
                M1 = M1_cell{i};
                
                % comptute SDEC LBO (symmetric) operator
                W_SDEC = shape.d0_matrix'*M1*shape.d0_matrix;
                W_SDEC = (W_SDEC+W_SDEC')/2; %verify symmetry
                
                % compute SDEC eigenfunctions
                [V_SDEC, D_SDEC, eigs_flag] = eigs(W_SDEC, M0, num_eigen, sigma, 'IsSymmetricDefinite', 1);
                
                % verify eigenvalues convergence
                assert(eigs_flag == 0, 'Error: eigenvalues did not converge');
                
                % verify A*V = B*V*D, calculate frobenius norm
                solution_norm = norm(W_SDEC*V_SDEC - M0*V_SDEC*D_SDEC, 'fro');
                assert(solution_norm < sqrt(1e-6), 'Error: A*V != B*V*D');
                
                % verify othornormality with respect to mass matrix
                assert(norm(V_SDEC'*M0*V_SDEC - speye(num_eigen) ,'fro') < 1e-11, 'Error: not orthonormal with respect to mass');
                
                % sort ascend
                [vals, inds] = sort(diag(D_SDEC));
                V_SDEC = V_SDEC(:, inds);
                
                % check if there are complex eigen values
                assert(all(imag(vals) == 0), 'Error: complex eigenvalues');
                
                % return the lbo eigen functions/values, W, mass matrix,
                % subdivision operator for 0-forms (vertices)
                % return in struct
                if i == 1 % low pass of the coarset level
                    basis_hierarchical{i}.eigenfunctions = V_SDEC;
                    basis_hierarchical{i}.eigenvalues = vals;
                else % take just bandpass
                    % TODO: think of a better way, this could lead to
                    % double (or missing) eigenfunction, from the previous and current
                    % level
                    [~, ind_bandpass] = min(abs(vals - last_largest_eig));
                    basis_hierarchical{i}.eigenfunctions = V_SDEC(:, ind_bandpass+1:end);
                    basis_hierarchical{i}.eigenvalues = vals(ind_bandpass+1:end);
                end
                
                
                % construct the basis struct for this level
                basis_hierarchical{i}.shape = shape;
                basis_hierarchical{i}.W = W_SDEC;
                basis_hierarchical{i}.A = M0;
                if i < subdivision_levels+1
                    basis_hierarchical{i}.S_0 = S_0_cell{i};
                end
                basis_hierarchical{i}.num_eigs = numel(basis_hierarchical{i}.eigenvalues);
            end
            toc;
            
        end
        function basis_hierarchical = lbo_basis_hierarchical_SDEC_S0(obj, num_eigen_vec, subdivision_levels) % compute SDEC lbo eigen basis hierarchical
            % lbo_basis_hierarchical_SDEC  Compute the num_eigen_vec sub basis of the LBO
            
            % check required dimensions
            assert(numel(num_eigen_vec) == subdivision_levels+1, 'Err number of subdivision levels is not equal to size of eigen vecs required');
            
            % initialize
            basis_hierarchical = cell(subdivision_levels+1, 1);
            
            
            %% get SDEC operators
            tic;
            [S_0_cell, shape_subdiv_cell] = obj.SDEC_operators_S0(subdivision_levels);
            toc;
            
            %% Get M0 and W of the hierarchy
            % initiate cell arrays to contain all M0 and W in the
            % hierarchy.
            M0_cell = cell(1, subdivision_levels+1);
            W_cell  = cell(1, subdivision_levels+1);
            % M0 SEC and W DEC of the finest level
            if subdivision_levels > 0
                M0_cell{subdivision_levels+1} = shape_subdiv_cell{end}.star0_matrix;
                W_cell{subdivision_levels+1} = shape_subdiv_cell{end}.d0_matrix'*shape_subdiv_cell{end}.star1_matrix*shape_subdiv_cell{end}.d0_matrix;
            else
                M0_cell{subdivision_levels+1} = obj.star0_matrix;
                W_cell{subdivision_levels+1} = obj.d0_matrix'*obj.star1_matrix*obj.d0_matrix;
            end
            % run over levels and build M0 and W. From fine to coarse M0 =
            % S' * ... * S' * M0_DEC_fine * S *...*S
            for i = subdivision_levels:-1:1
                
                % get current level S_0 and S_1
                S_0 = S_0_cell{i};
                
                % update M0 to this level
                M0_cell{i} = S_0'*M0_cell{i+1}*S_0;
                
                % update W to this level
                W_cell{i} = S_0'*W_cell{i+1}*S_0;
                
            end
            
            
            %% compute basis for each hierarchy level - starting from coarset level
%             tic;
%             for i = 1:subdivision_levels+1
%                 
%                 fprintf('Compute SEC basis for sbd level %d/%d', i-1, subdivision_levels);
%                 
%                 % get current shape
%                 if i == 1 % coarset shape
%                     shape = obj;
%                     num_eigen = num_eigen_vec(i);
%                     sigma = -1e-5;
%                     fprintf(' %d eigens \n', num_eigen);
%                     prev_starting_eig_ind = 1;
%                     curr_starting_eig_ind = prev_starting_eig_ind + num_eigen;
%                 else % subdivided shape
%                     shape = shape_subdiv_cell{i-1};
% %                     num_eigen = num_eigen_vec(i)*2;
%                     num_eigen = num_eigen_vec(i);
%                     last_largest_eig = basis_hierarchical{i-1}.eigenvalues(end);
%                     fprintf(', wanted %d eigens', num_eigen);
%                     % aprroximate linear line lambda(i) = m*i
%                     lin_fit = polyfit(prev_starting_eig_ind:1:curr_starting_eig_ind-1, basis_hierarchical{i-1}.eigenvalues', 1);
%                     lin_val = polyval(lin_fit, curr_starting_eig_ind-7:1:curr_starting_eig_ind + num_eigen -1); % taking 3 values back to be sure we don't skeep values
%                     sigma = (max(lin_val) + min(lin_val))/2; % take the middle
%                     
%                 end
%                 
%                 % get current operators and matrices
%                 M0 = M0_cell{i};
%                 M0 = (M0 + M0')/2; %verify symmetry
%                 
%                 % comptute SDEC LBO (symmetric) operator
%                 W_SDEC = W_cell{i};
%                 W_SDEC = (W_SDEC+W_SDEC')/2; %verify symmetry
%                 
%                 % compute SDEC eigenfunctions
%                 [V_SDEC, D_SDEC, eigs_flag] = eigs(W_SDEC, M0, num_eigen, sigma, 'IsSymmetricDefinite', 1);
%                 
%                 % verify eigenvalues convergence
%                 assert(eigs_flag == 0, 'Error: eigenvalues did not converge');
%                 
%                 % verify A*V = B*V*D, calculate frobenius norm
%                 solution_norm = norm(W_SDEC*V_SDEC - M0*V_SDEC*D_SDEC, 'fro');
%                 assert(solution_norm < sqrt(1e-6), 'Error: A*V != B*V*D');
%                 
%                 % verify othornormality with respect to mass matrix
%                 assert(norm(V_SDEC'*M0*V_SDEC - speye(num_eigen) ,'fro') < 1e-11, 'Error: not orthonormal with respect to mass');
%                 
%                 % sort ascend
%                 [vals, inds] = sort(diag(D_SDEC));
%                 V_SDEC = V_SDEC(:, inds);
%                 
%                 % check if there are complex eigen values
%                 assert(all(imag(vals) == 0), 'Error: complex eigenvalues');
%                 
%                 % return the lbo eigen functions/values, W, mass matrix,
%                 % subdivision operator for 0-forms (vertices)
%                 % return in struct
%                 if i == 1 % low pass of the coarset level
%                     basis_hierarchical{i}.eigenfunctions = V_SDEC;
%                     basis_hierarchical{i}.eigenvalues = vals;
%                 else % take just bandpass
%                     % TODO: think of a better way, this could lead to
%                     % double (or missing) eigenfunction, from the previous and current
%                     % level
%                     [~, ind_bandpass] = min(abs(vals - last_largest_eig));
%                     basis_hierarchical{i}.eigenfunctions = V_SDEC(:, ind_bandpass+1:end);
%                     basis_hierarchical{i}.eigenvalues = vals(ind_bandpass+1:end);
%                     prev_starting_eig_ind = curr_starting_eig_ind;
%                     curr_starting_eig_ind = curr_starting_eig_ind + num_eigen - ind_bandpass;
%                     fprintf(', computed %d eigens \n', num_eigen - ind_bandpass - 1);
%                 end
%                 
%                 
%                 % construct the basis struct for this level
%                 basis_hierarchical{i}.shape = shape;
%                 basis_hierarchical{i}.W = W_SDEC;
%                 basis_hierarchical{i}.A = M0;
%                 if i < subdivision_levels+1
%                     basis_hierarchical{i}.S_0 = S_0_cell{i};
%                 end
%                 basis_hierarchical{i}.num_eigs = numel(basis_hierarchical{i}.eigenvalues);
%             end
%             toc;
            tic;
            for i = 1:subdivision_levels+1
                
                fprintf('Compute SEC basis for sbd level %d/%d', i-1, subdivision_levels);
                
                % get current shape
                if i == 1 % coarset shape
                    shape = obj;
                    num_eigen = num_eigen_vec(i);
                    eigs_addition = 0;
                    sigma = -1e-5;
                    fprintf(' %d eigens \n', num_eigen);
                    prev_starting_eig_ind = 1;
                    curr_starting_eig_ind = prev_starting_eig_ind + num_eigen;
                else % subdivided shape
                    shape = shape_subdiv_cell{i-1};
                    num_eigen = num_eigen_vec(i);
                    eigs_addition = 10; % computing extra to get exatcly the demanded num of eigs
                    last_largest_eig = basis_hierarchical{i-1}.eigenvalues(end);
                    % aprroximate linear line lambda(i) = m*i
                    lin_fit = polyfit(prev_starting_eig_ind:1:curr_starting_eig_ind-1, basis_hierarchical{i-1}.eigenvalues', 1);
                    lin_val = polyval(lin_fit, curr_starting_eig_ind:1:curr_starting_eig_ind + num_eigen -1); 
                    if i==2 
                        sigma = (lin_val(end-5) + lin_val(1))/2; % take the middle (one value from the end, else could miss eigens)
                    else
                        sigma = (lin_val(end) + lin_val(1))/2; % take the middle (one value from the end, else could miss eigens)
                    end
                    
                end
                
                % get current operators and matrices
                M0_matrix = M0_cell{i};
                M0_matrix = (M0_matrix + M0_matrix')/2; %verify symmetry
                
                % comptute SDEC LBO (symmetric) operator
                W_SEC = W_cell{i};
                W_SEC = (W_SEC+W_SEC')/2; %verify symmetry
                
                % compute SDEC eigenfunctions
                [V_SEC, D_SEC, eigs_flag] = eigs(W_SEC, M0_matrix, num_eigen + eigs_addition, sigma, 'IsSymmetricDefinite', 1);
                
                % verify eigenvalues convergence
                assert(eigs_flag == 0, 'Error: eigenvalues did not converge');
                
                % verify A*V = B*V*D, calculate frobenius norm
                solution_norm = norm(W_SEC*V_SEC - M0_matrix*V_SEC*D_SEC, 'fro');
                assert(solution_norm < sqrt(1e-6), 'Error: A*V != B*V*D');
                
                % verify othornormality with respect to mass matrix
                assert(norm(V_SEC'*M0_matrix*V_SEC - speye(num_eigen+eigs_addition) ,'fro') < 1e-11, 'Error: not orthonormal with respect to mass');
                
                % sort ascend
                [vals, inds] = sort(diag(D_SEC));
                V_SEC = V_SEC(:, inds);
                
                % check if there are complex eigen values
                assert(all(imag(vals) == 0), 'Error: complex eigenvalues');
                
                % return the lbo eigen functions/values, W, mass matrix,
                % subdivision operator for 0-forms (vertices)
                % return in struct
                if i == 1 % low pass of the coarset level
                    basis_hierarchical{i}.eigenfunctions = V_SEC;
                    basis_hierarchical{i}.eigenvalues = vals;
                else % take just bandpass
                    % TODO: think of a better way, this could lead to
                    % double (or missing) eigenfunction, from the previous and current
                    % level
                    [~, ind_bandpass] = min(abs(vals - last_largest_eig));
                    assert(ind_bandpass <= eigs_addition, 'Error: could not set the wanted number of eigs');
                    basis_hierarchical{i}.eigenfunctions = V_SEC(:, ind_bandpass+1:ind_bandpass+num_eigen);
                    basis_hierarchical{i}.eigenvalues = vals(ind_bandpass+1:ind_bandpass+num_eigen);
                    prev_starting_eig_ind = curr_starting_eig_ind;
                    curr_starting_eig_ind = curr_starting_eig_ind + num_eigen;
                    fprintf(', computed %d eigens, using %d \n',num_eigen + eigs_addition, num_eigen);
                end
                
                
                % construct the basis struct for this level
                basis_hierarchical{i}.shape = shape;
                basis_hierarchical{i}.W = W_SEC;
                basis_hierarchical{i}.A = M0_matrix;
                if i < subdivision_levels+1
                    basis_hierarchical{i}.S_0 = S_0_cell{i};
                end
                basis_hierarchical{i}.num_eigs = numel(basis_hierarchical{i}.eigenvalues);
            end
            toc;
            
        end
        function [basis, shape_subdiv] = lbo_basis_SDEC_trivial(obj, num_eigen, subdivision_levels, flag_full_M0) % compute SDEC lbo eigen basis
            % lbo_basis_SDEC  Compute the num_eigen sub basis of the LBO
            % using the trivial subdivision for the fine level mass
            % matrices.
            
            % get SDEC operators
            [M0, M1, ~, A_0, ~, ~, shape_subdiv, M0_full] = obj.SDEC_operators_trivial(subdivision_levels);
            if flag_full_M0 == 1 % use full (not lumped) M0 of the subdivided mesh
                M0 = M0_full;
            end
            
            % comptute SDEC LBO operator
            W_SDEC = obj.d0_matrix'*M1*obj.d0_matrix;
            
            % compute SDEC eigenfunctions
            disp('Compute SDEC eigenfucntions');
            [V_SDEC, D_SDEC] = eigs((W_SDEC+W_SDEC')/2, (M0 + M0')/2, num_eigen, -1e-5, 'IsSymmetricDefinite', 1);
            
            % verify A*V = B*V*D, calculate frobenius norm
            solution_norm = norm(W_SDEC*V_SDEC - M0*V_SDEC*D_SDEC, 'fro');
            %             assert(solution_norm < sqrt(1e-11), 'Error: A*V != B*V*D');
            
            % verify othornormality with respect to mass matrix
            assert(norm(V_SDEC'*(M0 + M0')/2*V_SDEC - speye(num_eigen), 'fro') < 1e-11, 'Error: not orthonormal with respect to mass');
            
            % sort ascend
            [vals, inds] = sort(diag(D_SDEC));
            V_SDEC = V_SDEC(:, inds);
            
            % check if there are complex eigen values
            assert(all(imag(vals) == 0), 'Error: complex eigenvalues');
            
            % return the lbo eigen functions/values, W, mass matrix,
            % subdivision operator for 0-forms (vertices)
            % return in struct
            basis.eigenfunctions = V_SDEC;
            basis.eigenvalues = vals;
            basis.W = (W_SDEC+W_SDEC')/2;
            basis.A = (M0 + M0')/2;
            basis.A_0 = A_0;
            
            
        end
    end
    
    methods(Static)
        function WKS = WKS(lbo_vecs, lbo_eigenvals, num_energies) % compute WKS
            % WKS  Computing the Wave Kernel Signature for the shape
            
            % prepare log diagonal of eigen-energies, setting -inf for log(0);
            log_eigen_vals = [-inf; log(lbo_eigenvals(2:end))];
            
            % energies for wave kernel signature
            delta = (log_eigen_vals(end) - log_eigen_vals(2))/num_energies;
            %             delta = (log_eigen_vals(end) - log_eigen_vals(2))/(num_energies + 28);
            sigma = 7*delta;
            %             e_min = log_eigen_vals(2) + 2*sigma;
            e_min = log_eigen_vals(2);
            %             e_max = log_eigen_vals(end) -2*sigma;
            e_max = log_eigen_vals(end);
            e_vec = linspace(e_min, e_max, num_energies);
            
            % compute exponential weights per energy
            exponential_coefs_matrix = exp(-(e_vec - log_eigen_vals).^2/(2*sigma.^2));
            
            % normalization factors per energy (using broadcast - row vector minus column vector)
            Ce = sum(exponential_coefs_matrix).^-1;
            
            % compute the WKS
            WKS = lbo_vecs.^2*(exponential_coefs_matrix.*Ce);
            
        end
        function WKM = WKM(lbo_vecs, lbo_eigenvals, num_energies, landmarks) % compute WKS
            % WKM  Computing the Wave Kernel Map for the shape with the
            % given landmrks
            
            % prepare log diagonal of eigen-energies, setting -inf for log(0);
            log_eigen_vals = [-inf; log(lbo_eigenvals(2:end))];
            
            % energies for wave kernel signature
            delta = (log_eigen_vals(end) - log_eigen_vals(2))/num_energies;
            %             delta = (log_eigen_vals(end) - log_eigen_vals(2))/(num_energies + 28);
            sigma = 7*delta;
            %             e_min = log_eigen_vals(2) + 2*sigma;
            e_min = log_eigen_vals(2);
            %             e_max = log_eigen_vals(end) -2*sigma;
            e_max = log_eigen_vals(end);
            e_vec = linspace(e_min, e_max, num_energies);
            
            % compute exponential weights per energy
            exponential_coefs_matrix = exp(-(e_vec - log_eigen_vals).^2/(2*sigma.^2));
            
            % normalization factors per energy (using broadcast - row vector minus column vector)
            Ce = sum(exponential_coefs_matrix).^-1;
            
            % compute the WKM
            WKM = zeros(size(lbo_vecs,1), num_energies*numel(landmarks));
            for i = 1:numel(landmarks)
                % get WKM (using broadcasting e.g. 3x3 matrix time 1x3 row
                % vector)
                WKM(:, (i-1)*num_energies + 1: i*num_energies) = ...
                    lbo_vecs.*lbo_vecs(landmarks(i), :) *(exponential_coefs_matrix.*Ce);
            end
        end
        
    end
    
end